@@ -1,6 +1,6 @@
 /* Miscellaneous error functions
 
-   Copyright (C) 2005, 2007 Free Software Foundation, Inc.
+   Copyright (C) 2005, 2007, 2023-2026 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -13,8 +13,7 @@
    Public License for more details.
 
    You should have received a copy of the GNU General Public License along
-   with this program; if not, write to the Free Software Foundation, Inc.,
-   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+   with this program.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include <system.h>
 #include <paxlib.h>
@@ -23,6 +22,53 @@
 
 void (*error_hook) (void);
 
+void
+paxwarn (int errnum, char const *format, ...)
+{
+  if (error_hook)
+    error_hook ();
+  va_list ap;
+  va_start (ap, format);
+  verror (0, errnum, format, ap);
+  va_end (ap);
+}
+
+void
+paxerror (int errnum, char const *format, ...)
+{
+  if (error_hook)
+    error_hook ();
+  va_list ap;
+  va_start (ap, format);
+  verror (0, errnum, format, ap);
+  va_end (ap);
+  exit_status = PAXEXIT_FAILURE;
+}
+
+void
+paxfatal (int errnum, char const *format, ...)
+{
+  if (error_hook)
+    error_hook ();
+  va_list ap;
+  va_start (ap, format);
+  verror (0, errnum, format, ap);
+  va_end (ap);
+  fatal_exit ();
+}
+
+void
+paxusage (char const *format, ...)
+{
+  if (error_hook)
+    error_hook ();
+  va_list ap;
+  va_start (ap, format);
+  verror (0, 0, format, ap);
+  va_end (ap);
+  usage (PAXEXIT_FAILURE);
+}
+
 /* Decode MODE from its binary form in a stat structure, and encode it
    into a 9-byte string STRING, terminated with a NUL.  */
 
@@ -52,12 +98,11 @@ pax_decode_mode (mode_t mode, char *string)
 void
 call_arg_error (char const *call, char const *name)
 {
-  int e = errno;
-  /* TRANSLATORS: %s after `Cannot' is a function name, e.g. `Cannot open'.
+  /* TRANSLATORS: %s after 'Cannot' is a function name, e.g. 'Cannot open'.
      Directly translating this to another language will not work, first because
      %s itself is not translated.
-     Translate it as `%s: Function %s failed'. */
-  ERROR ((0, e, _("%s: Cannot %s"), quotearg_colon (name), call));
+     Translate it as '%s: Function %s failed'. */
+  paxerror (errno, _("%s: Cannot %s"), quotearg_colon (name), call);
 }
 
 /* Report a fatal error associated with the system call CALL and
@@ -65,12 +110,11 @@ call_arg_error (char const *call, char const *name)
 void
 call_arg_fatal (char const *call, char const *name)
 {
-  int e = errno;
-  /* TRANSLATORS: %s after `Cannot' is a function name, e.g. `Cannot open'.
+  /* TRANSLATORS: %s after 'Cannot' is a function name, e.g. 'Cannot open'.
      Directly translating this to another language will not work, first because
      %s itself is not translated.
-     Translate it as `%s: Function %s failed'. */
-  FATAL_ERROR ((0, e, _("%s: Cannot %s"), quotearg_colon (name),  call));
+     Translate it as '%s: Function %s failed'. */
+  paxfatal (errno, _("%s: Cannot %s"), quotearg_colon (name),  call);
 }
 
 /* Report a warning associated with the system call CALL and
@@ -78,30 +122,27 @@ call_arg_fatal (char const *call, char const *name)
 void
 call_arg_warn (char const *call, char const *name)
 {
-  int e = errno;
-  /* TRANSLATORS: %s after `Cannot' is a function name, e.g. `Cannot open'.
+  /* TRANSLATORS: %s after 'Cannot' is a function name, e.g. 'Cannot open'.
      Directly translating this to another language will not work, first because
      %s itself is not translated.
-     Translate it as `%s: Function %s failed'. */
-  WARN ((0, e, _("%s: Warning: Cannot %s"), quotearg_colon (name), call));
+     Translate it as '%s: Function %s failed'. */
+  paxwarn (errno, _("%s: Warning: Cannot %s"), quotearg_colon (name), call);
 }
 
 void
 chmod_error_details (char const *name, mode_t mode)
 {
-  int e = errno;
   char buf[10];
   pax_decode_mode (mode, buf);
-  ERROR ((0, e, _("%s: Cannot change mode to %s"),
-	  quotearg_colon (name), buf));
+  paxerror (errno, _("%s: Cannot change mode to %s"), quotearg_colon (name), buf);
 }
 
 void
 chown_error_details (char const *name, uid_t uid, gid_t gid)
 {
-  int e = errno;
-  ERROR ((0, e, _("%s: Cannot change ownership to uid %lu, gid %lu"),
-	  quotearg_colon (name), (unsigned long) uid, (unsigned long) gid));
+  uintmax_t u = uid, g = gid;
+  paxerror (errno, _("%s: Cannot change ownership to uid %ju, gid %ju"),
+	    quotearg_colon (name), u, g);
 }
 
 void
@@ -125,9 +166,8 @@ exec_fatal (char const *name)
 void
 link_error (char const *target, char const *source)
 {
-  int e = errno;
-  ERROR ((0, e, _("%s: Cannot hard link to %s"),
-	  quotearg_colon (source), quote_n (1, target)));
+  paxerror (errno, _("%s: Cannot hard link to %s"),
+	    quotearg_colon (source), quote_n (1, target));
 }
 
 void
@@ -173,29 +213,27 @@ read_error (char const *name)
 }
 
 void
-read_error_details (char const *name, off_t offset, size_t size)
+read_error_details (char const *name, off_t offset, idx_t size)
 {
-  char buf[UINTMAX_STRSIZE_BOUND];
-  int e = errno;
-  ERROR ((0, e,
-	  ngettext ("%s: Read error at byte %s, while reading %lu byte",
-		    "%s: Read error at byte %s, while reading %lu bytes",
-		    size),
-	  quotearg_colon (name), STRINGIFY_BIGINT (offset, buf),
-	  (unsigned long) size));
+  intmax_t off = offset;
+  paxerror (errno,
+	    ngettext ("%s: Read error at byte %jd, while reading %td byte",
+		      "%s: Read error at byte %jd, while reading %td bytes",
+		      size),
+	    quotearg_colon (name), off, size);
 }
 
 void
-read_warn_details (char const *name, off_t offset, size_t size)
+read_warn_details (char const *name, off_t offset, idx_t size)
 {
-  char buf[UINTMAX_STRSIZE_BOUND];
-  int e = errno;
-  WARN ((0, e,
-	 ngettext ("%s: Warning: Read error at byte %s, while reading %lu byte",
-		   "%s: Warning: Read error at byte %s, while reading %lu bytes",
-		   size),
-	 quotearg_colon (name), STRINGIFY_BIGINT (offset, buf),
-	 (unsigned long) size));
+  intmax_t off = offset;
+  paxwarn (errno,
+	   ngettext (("%s: Warning: Read error at byte %jd,"
+		      " while reading %td byte"),
+		     ("%s: Warning: Read error at byte %jd,"
+		      " while reading %td bytes"),
+		     size),
+	   quotearg_colon (name), off, size);
 }
 
 void
@@ -205,16 +243,14 @@ read_fatal (char const *name)
 }
 
 void
-read_fatal_details (char const *name, off_t offset, size_t size)
+read_fatal_details (char const *name, off_t offset, idx_t size)
 {
-  char buf[UINTMAX_STRSIZE_BOUND];
-  int e = errno;
-  FATAL_ERROR ((0, e,
-		ngettext ("%s: Read error at byte %s, while reading %lu byte",
-			  "%s: Read error at byte %s, while reading %lu bytes",
-			  size),
-		quotearg_colon (name), STRINGIFY_BIGINT (offset, buf),
-		(unsigned long) size));
+  intmax_t off = offset;
+  paxfatal (errno,
+	    ngettext ("%s: Read error at byte %jd, while reading %td byte",
+		      "%s: Read error at byte %jd, while reading %td bytes",
+		      size),
+	    quotearg_colon (name), off, size);
 }
 
 void
@@ -256,11 +292,8 @@ seek_error (char const *name)
 void
 seek_error_details (char const *name, off_t offset)
 {
-  char buf[UINTMAX_STRSIZE_BOUND];
-  int e = errno;
-  ERROR ((0, e, _("%s: Cannot seek to %s"),
-	  quotearg_colon (name),
-	  STRINGIFY_BIGINT (offset, buf)));
+  intmax_t off = offset;
+  paxerror (errno, _("%s: Cannot seek to %jd"), quotearg_colon (name), off);
 }
 
 void
@@ -272,19 +305,16 @@ seek_warn (char const *name)
 void
 seek_warn_details (char const *name, off_t offset)
 {
-  char buf[UINTMAX_STRSIZE_BOUND];
-  int e = errno;
-  WARN ((0, e, _("%s: Warning: Cannot seek to %s"),
-	 quotearg_colon (name),
-	 STRINGIFY_BIGINT (offset, buf)));
+  intmax_t off = offset;
+  paxwarn (errno, _("%s: Warning: Cannot seek to %jd"),
+	   quotearg_colon (name), off);
 }
 
 void
 symlink_error (char const *contents, char const *name)
 {
-  int e = errno;
-  ERROR ((0, e, _("%s: Cannot create symlink to %s"),
-	  quotearg_colon (name), quote_n (1, contents)));
+  paxerror (errno, _("%s: Cannot create symlink to %s"),
+	    quotearg_colon (name), quote_n (1, contents));
 }
 
 void
@@ -342,16 +372,16 @@ write_error (char const *name)
 }
 
 void
-write_error_details (char const *name, size_t status, size_t size)
+write_error_details (char const *name, idx_t status, idx_t size)
 {
   if (status == 0)
     write_error (name);
   else
-    ERROR ((0, 0,
-	    ngettext ("%s: Wrote only %lu of %lu byte",
-		      "%s: Wrote only %lu of %lu bytes",
-		      size),
-	    name, (unsigned long int) status, (unsigned long int) size));
+    paxerror (0,
+	      ngettext ("%s: Wrote only %td of %td byte",
+			"%s: Wrote only %td of %td bytes",
+			size),
+	      name, status, size);
 }
 
 void
