@@ -1,6 +1,6 @@
 /* Functions for communicating with a remote tape drive.
 
-   Copyright 1988-2023 Free Software Foundation, Inc.
+   Copyright 1988-2026 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,8 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software Foundation,
-   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+   along with this program.  If not, see <http://www.gnu.org/licenses/>. */
 
 /* The man page rmt(8) for /etc/rmt documents the remote mag tape protocol
    which rdump and rrestore use.  Unfortunately, the man page is *WRONG*.
@@ -32,53 +31,30 @@
    code, courtesy of Dan Kegel.  */
 
 #include "system.h"
-#include "system-ioctl.h"
 
-#include <safe-read.h>
-#include <full-write.h>
 #include <verify.h>
 
-/* Try hard to get EOPNOTSUPP defined.  486/ISC has it in net/errno.h,
-   3B2/SVR3 has it in sys/inet.h.  Otherwise, like on MSDOS, use EINVAL.  */
-
-#ifndef EOPNOTSUPP
-# if HAVE_NET_ERRNO_H
-#  include <net/errno.h>
-# endif
-# if HAVE_SYS_INET_H
-#  include <sys/inet.h>
-# endif
-# ifndef EOPNOTSUPP
-#  define EOPNOTSUPP EINVAL
-# endif
-#endif
-
 #include <signal.h>
 
+#if HAVE_SYS_MTIO_H
+# include <sys/mtio.h>
+#endif
+
 #if HAVE_NETDB_H
 # include <netdb.h>
 #endif
 
-#include <rmt.h>
-#include <rmt-command.h>
-
 /* Exit status if exec errors.  */
-#define EXIT_ON_EXEC_ERROR 128
+enum { EXIT_ON_EXEC_ERROR = 128 };
 
 /* FIXME: Size of buffers for reading and writing commands to rmt.  */
-#define COMMAND_BUFFER_SIZE 64
+enum { COMMAND_BUFFER_SIZE = 64 };
 
 /* FIXME: Maximum number of simultaneous remote tape connections.  */
-#define MAXUNIT	4
+enum { MAXUNIT = 4 };
 
-#define	PREAD 0			/* read  file descriptor from pipe() */
-#define	PWRITE 1		/* write file descriptor from pipe() */
-
-/* Return the parent's read side of remote tape connection Fd.  */
-#define READ_SIDE(Fd) (from_remote[Fd][PREAD])
-
-/* Return the parent's write side of remote tape connection Fd.  */
-#define WRITE_SIDE(Fd) (to_remote[Fd][PWRITE])
+/* Read and write file descriptors from pipe () calls.  */
+enum { PREAD, PWRITE };		/* read  file descriptor from pipe() */
 
 /* The pipes for receiving data from remote tape drives.  */
 static int from_remote[MAXUNIT][2] = {{-1, -1}, {-1, -1}, {-1, -1}, {-1, -1}};
@@ -86,39 +62,68 @@ static int from_remote[MAXUNIT][2] = {{-1, -1}, {-1, -1}, {-1, -1}, {-1, -1}};
 /* The pipes for sending data to remote tape drives.  */
 static int to_remote[MAXUNIT][2] = {{-1, -1}, {-1, -1}, {-1, -1}, {-1, -1}};
 
-char const *rmt_command = DEFAULT_RMT_COMMAND;
+/* The parent's read side of remote tape connection Fd.  */
+static int
+read_side (int handle)
+{
+  return from_remote[handle][PREAD];
+}
 
-/* Temporary variable used by macros in rmt.h.  */
-char const *rmt_dev_name__;
+/* The parent's write side of remote tape connection Fd.  */
+static int
+write_side (int handle)
+{
+  return to_remote[handle][PWRITE];
+}
+
+/* ../paxlib/rtape.c #defines PAXLIB_RTAPE.  */
+#ifdef PAXLIB_RTAPE
+# include <localedir.h> /* for DEFAULT_RMT_COMMAND */
+#else
+# define RMT_INLINE _GL_EXTERN_INLINE
+# include <rmt.h>
+# include <rmt-command.h>
+
+char const *rmt_command = DEFAULT_RMT_COMMAND;
 
 /* If true, always consider file names to be local, even if they contain
    colons */
 bool force_local_option;
 
+# define _rmt_rexec (host, user, rmt_command) _rmt_rexec (host, user)
+# define rmt_open(file_name, oflags, bias, remote_shell, rmt_command) \
+    rmt_open__ (file_name, oflags, bias, remote_shell)
+# define rmt_close(handle) rmt_close__ (handle)
+# define rmt_read(handle, buffer, length) rmt_read__ (handle, buffer, length)
+# define rmt_write(handle, buffer, length) rmt_write__ (handle, buffer, length)
+# define rmt_lseek(handle, offset, whence) rmt_lseek__ (handle, offset, whence)
+# define rmt_ioctl(handle, operation, argument) \
+    rmt_ioctl__ (handle, operation, argument)
+#endif
+
 
 
 /* Close remote tape connection HANDLE, and reset errno to ERRNO_VALUE.  */
 static void
 _rmt_shutdown (int handle, int errno_value)
 {
-  close (READ_SIDE (handle));
-  close (WRITE_SIDE (handle));
-  READ_SIDE (handle) = -1;
-  WRITE_SIDE (handle) = -1;
+  close (read_side (handle));
+  close (write_side (handle));
+  from_remote[handle][PREAD] = -1;
+  to_remote[handle][PWRITE] = -1;
   errno = errno_value;
 }
 
-/* Attempt to perform the remote tape command specified in BUFFER on
-   remote tape connection HANDLE.  Return 0 if successful, -1 on
-   error.  */
+/* Using remote tape connection HANDLE, do the remote tape command
+   specified in BUFFER with LENGTH bytes.
+   Return 0 if successful, -1 on error.  */
 static int
-do_command (int handle, const char *buffer)
+do_command (int handle, const char *buffer, idx_t length)
 {
   /* Save the current pipe handler and try to make the request.  */
 
-  size_t length = strlen (buffer);
   void (*pipe_handler) (int) = signal (SIGPIPE, SIG_IGN);
-  ssize_t written = full_write (WRITE_SIDE (handle), buffer, length);
+  idx_t written = full_write (write_side (handle), buffer, length);
   signal (SIGPIPE, pipe_handler);
 
   if (written == length)
@@ -130,41 +135,61 @@ do_command (int handle, const char *buffer)
   return -1;
 }
 
+/* Return the integer represented by the array of bytes at the start of S.
+   The bytes use the usual unsigned ASCII decimal representation,
+   and are terminated by a non ASCII digit.  Return -1 if S does not
+   start with a decimal string, or if the integer exceeds SMAX.  */
+static intmax_t
+dectointmax (char const *s, intmax_t smax)
+{
+  /* Do not use strtoimax, as it allows leading whitespace, '-' and
+     '+', and requires the input to be null-terminated.  */
+  if ('0' <= *s && *s <= '9')
+    {
+      bool overflow = false;
+      intmax_t result = *s - '0';
+      while ('0' <= *++s && *s <= '9')
+	{
+	  overflow |= ckd_mul (&result, result, 10);
+	  overflow |= ckd_add (&result, result, *s - '0');
+	}
+      overflow |= smax < result;
+      if (!overflow)
+	{
+	  assume (0 <= result);
+	  return result;
+	}
+    }
+  return -1;
+}
+
+/* Get from HANDLE a response string into COMMAND_BUFFER, terminated by '\n'.
+   If the response is a successful command, starting with 'A',
+   return the address of the byte following the 'A'.
+   Otherwise, set errno and return a null pointer.  */
 static char *
-get_status_string (int handle, char *command_buffer)
+get_status_string (int handle, char command_buffer[COMMAND_BUFFER_SIZE])
 {
   char *cursor;
-  int counter;
 
   /* Read the reply command line.  */
 
-  for (counter = 0, cursor = command_buffer;
-       counter < COMMAND_BUFFER_SIZE;
-       counter++, cursor++)
+  for (cursor = command_buffer; ; cursor++)
     {
-      if (safe_read (READ_SIDE (handle), cursor, 1) != 1)
+      if (cursor == command_buffer + COMMAND_BUFFER_SIZE
+	  || safe_read (read_side (handle), cursor, 1) != 1)
 	{
 	  _rmt_shutdown (handle, EIO);
-	  return 0;
+	  return nullptr;
 	}
       if (*cursor == '\n')
-	{
-	  *cursor = '\0';
-	  break;
-	}
-    }
-
-  if (counter == COMMAND_BUFFER_SIZE)
-    {
-      _rmt_shutdown (handle, EIO);
-      return 0;
+	break;
     }
 
   /* Check the return status.  */
 
-  for (cursor = command_buffer; *cursor; cursor++)
-    if (*cursor != ' ')
-      break;
+  for (cursor = command_buffer; *cursor == ' '; cursor++)
+    continue;
 
   if (*cursor == 'E' || *cursor == 'F')
     {
@@ -173,20 +198,27 @@ get_status_string (int handle, char *command_buffer)
       /* FIXME: there is better to do than merely ignoring error messages
 	 coming from the remote end.  Translate them, too...  */
 
-      {
-	char character;
+      char character;
+      do
+	{
+	  if (safe_read (read_side (handle), &character, 1) != 1)
+	    {
+	      _rmt_shutdown (handle, EIO);
+	      return nullptr;
+	    }
+	}
+      while (character == '\n');
 
-	while (safe_read (READ_SIDE (handle), &character, 1) == 1)
-	  if (character == '\n')
-	    break;
-      }
-
-      errno = atoi (cursor + 1);
+      /* This assumes remote errno values are the same as local,
+	 which is wrong in general, but does work in common cases
+	 and perhaps is the best we can do here.  */
+      int err = dectointmax (cursor + 1, INT_MAX);
+      errno = err <= 0 ? EIO : err;
 
       if (*cursor == 'F')
 	_rmt_shutdown (handle, errno);
 
-      return 0;
+      return nullptr;
     }
 
   /* Check for mis-synced pipes.  */
@@ -194,24 +226,25 @@ get_status_string (int handle, char *command_buffer)
   if (*cursor != 'A')
     {
       _rmt_shutdown (handle, EIO);
-      return 0;
+      return nullptr;
     }
 
-  /* Got an `A' (success) response.  */
+  /* Got an 'A' (success) response.  */
 
   return cursor + 1;
 }
 
-/* Read and return the status from remote tape connection HANDLE.  If
+/* Read and return the status from remote tape connection HANDLE.
+   The status must be in the range 0..STATUS_MAX.  If
    an error occurred, return -1 and set errno.  */
-static long int
-get_status (int handle)
+static intmax_t
+get_status (int handle, intmax_t status_max)
 {
   char command_buffer[COMMAND_BUFFER_SIZE];
   const char *status = get_status_string (handle, command_buffer);
   if (status)
     {
-      long int result = atol (status);
+      intmax_t result = dectointmax (status, status_max);
       if (0 <= result)
 	return result;
       errno = EIO;
@@ -219,48 +252,6 @@ get_status (int handle)
   return -1;
 }
 
-static off_t
-get_status_off (int handle)
-{
-  char command_buffer[COMMAND_BUFFER_SIZE];
-  const char *status = get_status_string (handle, command_buffer);
-
-  if (! status)
-    return -1;
-  else
-    {
-      /* Parse status, taking care to check for overflow.
-	 We can't use standard functions,
-	 since off_t might be longer than long.  */
-
-      off_t count = 0;
-      int negative;
-
-      for (;  *status == ' ' || *status == '\t';  status++)
-	continue;
-
-      negative = *status == '-';
-      status += negative || *status == '+';
-
-      for (;;)
-	{
-	  int digit = *status++ - '0';
-	  if (9 < (unsigned) digit)
-	    break;
-	  else
-	    {
-	      off_t c10 = 10 * count;
-	      off_t nc = negative ? c10 - digit : c10 + digit;
-	      if (c10 / 10 != count || (negative ? c10 < nc : nc < c10))
-		return -1;
-	      count = nc;
-	    }
-	}
-
-      return count;
-    }
-}
-
 #if WITH_REXEC
 
 /* Execute /etc/rmt as user USER on remote system HOST using rexec.
@@ -273,12 +264,24 @@ get_status_off (int handle)
    password info.  This may be unacceptable, and .rhosts files for use
    with rsh are much more common on BSD systems.  */
 static int
-_rmt_rexec (char *host, char *user)
+_rmt_rexec (char *host, char const *user, char const *rmt_command)
 {
+  struct servent *rexecserv = getservbyname ("exec", "tcp");
+  if (!rexecserv)
+    error (EXIT_ON_EXEC_ERROR, 0, _("exec/tcp: Service not available"));
+
   int saved_stdin = dup (STDIN_FILENO);
+  if ((saved_stdin < 0 && errno != EBADF)
+      || (saved_stdin == STDOUT_FILENO
+	  && ((saved_stdin = dup (STDOUT_FILENO)) < 0
+	      || close (STDOUT_FILENO) < 0)))
+    error (EXIT_ON_EXEC_ERROR, errno, _("stdin"));
   int saved_stdout = dup (STDOUT_FILENO);
-  struct servent *rexecserv;
-  int result;
+  if ((saved_stdout < 0 && errno != EBADF)
+      || (saved_stdout == STDIN_FILENO
+	  && ((saved_stdout = dup (STDIN_FILENO)) < 0
+	      || close (STDIN_FILENO) < 0)))
+    error (EXIT_ON_EXEC_ERROR, errno, _("stdout"));
 
   /* When using cpio -o < filename, stdin is no longer the tty.  But the
      rexec subroutine reads the login and the passwd on stdin, to allow
@@ -286,21 +289,20 @@ _rmt_rexec (char *host, char *user)
      /dev/tty before the rexec and give them back their original value
      after.  */
 
-  if (! freopen ("/dev/tty", "r", stdin))
-    freopen ("/dev/null", "r", stdin);
-  if (! freopen ("/dev/tty", "w", stdout))
-    freopen ("/dev/null", "w", stdout);
+  if ((0 <= saved_stdin && close (STDIN_FILENO) < 0)
+      || (open ("/dev/tty", O_RDONLY) < 0 && open ("/dev/null", O_RDONLY) < 0)
+      || (0 <= saved_stdout && close (STDOUT_FILENO) < 0)
+      || (open ("/dev/tty", O_WRONLY) < 0 && open ("/dev/null", O_WRONLY) < 0))
+    error (EXIT_ON_EXEC_ERROR, errno, "/dev/null");
 
-  if (rexecserv = getservbyname ("exec", "tcp"), !rexecserv)
-    error (EXIT_ON_EXEC_ERROR, 0, _("exec/tcp: Service not available"));
-
-  result = rexec (&host, rexecserv->s_port, user, 0, rmt_command, 0);
-  if (fclose (stdin) == EOF)
+  int result = rexec (&host, rexecserv->s_port,
+		      user, nullptr, rmt_command, nullptr);
+  if (0 <= saved_stdin
+      && (dup2 (saved_stdin, STDIN_FILENO) < 0 || close (saved_stdin) < 0))
     error (0, errno, _("stdin"));
-  fdopen (saved_stdin, "r");
-  if (fclose (stdout) == EOF)
+  if (0 <= saved_stdout
+      && (dup2 (saved_stdout, STDOUT_FILENO) < 0 || close (saved_stdout) < 0))
     error (0, errno, _("stdout"));
-  fdopen (saved_stdout, "w");
 
   return result;
 }
@@ -314,49 +316,51 @@ _rmt_rexec (char *host, char *user)
 
 #endif /* WITH_REXEC */
 
-/* Place into BUF a string representing OFLAG, which must be suitable
-   as argument 2 of `open'.  BUF must be large enough to hold the
-   result.  This function should generate a string that decode_oflag
-   can parse.  */
-static void
-encode_oflag (char *buf, int oflag)
+/* Place into BUF a string representing OFLAGS, which must be suitable
+   as argument 2 of 'open'.  BUF must be large enough to hold the
+   result.  This function should generate a string that decode_oflags
+   can parse.  Return a pointer to the trailing null byte of the string
+   if successful, a null pointer otherwise.  */
+static char *
+encode_oflags (char *buf, int oflags)
 {
-  sprintf (buf, "%d ", oflag);
+  char *p = buf + sprintf (buf, "%d ", oflags);
 
-  switch (oflag & O_ACCMODE)
+  switch (oflags & O_ACCMODE)
     {
-    case O_RDONLY: strcat (buf, "O_RDONLY"); break;
-    case O_RDWR: strcat (buf, "O_RDWR"); break;
-    case O_WRONLY: strcat (buf, "O_WRONLY"); break;
-    default: abort ();
+    case O_RDONLY: p = stpcpy (p, "O_RDONLY"); break;
+    case O_RDWR: p = stpcpy (p, "O_RDWR"); break;
+    case O_WRONLY: p = stpcpy (p, "O_WRONLY"); break;
+    default: return nullptr;
     }
 
 #ifdef O_APPEND
-  if (oflag & O_APPEND) strcat (buf, "|O_APPEND");
+  if (oflags & O_APPEND) p = stpcpy (p, "|O_APPEND");
 #endif
-  if (oflag & O_CREAT) strcat (buf, "|O_CREAT");
+  if (oflags & O_CREAT) p = stpcpy (p, "|O_CREAT");
 #ifdef O_DSYNC
-  if (oflag & O_DSYNC) strcat (buf, "|O_DSYNC");
+  if (oflags & O_DSYNC) p = stpcpy (p, "|O_DSYNC");
 #endif
-  if (oflag & O_EXCL) strcat (buf, "|O_EXCL");
+  if (oflags & O_EXCL) p = stpcpy (p, "|O_EXCL");
 #ifdef O_LARGEFILE
-  if (oflag & O_LARGEFILE) strcat (buf, "|O_LARGEFILE");
+  if (oflags & O_LARGEFILE) p = stpcpy (p, "|O_LARGEFILE");
 #endif
 #ifdef O_NOCTTY
-  if (oflag & O_NOCTTY) strcat (buf, "|O_NOCTTY");
+  if (oflags & O_NOCTTY) p = stpcpy (p, "|O_NOCTTY");
 #endif
-  if (oflag & O_NONBLOCK) strcat (buf, "|O_NONBLOCK");
+  if (oflags & O_NONBLOCK) p = stpcpy (p, "|O_NONBLOCK");
 #ifdef O_RSYNC
-  if (oflag & O_RSYNC) strcat (buf, "|O_RSYNC");
+  if (oflags & O_RSYNC) p = stpcpy (p, "|O_RSYNC");
 #endif
 #ifdef O_SYNC
-  if (oflag & O_SYNC) strcat (buf, "|O_SYNC");
+  if (oflags & O_SYNC) p = stpcpy (p, "|O_SYNC");
 #endif
-  if (oflag & O_TRUNC) strcat (buf, "|O_TRUNC");
+  if (oflags & O_TRUNC) p = stpcpy (p, "|O_TRUNC");
+  return p;
 }
 
 /* Reset user and group IDs to be those of the real user.
-   Return NULL on success, a failing syscall name (setting errno) on error.  */
+   Return null on success, a failing syscall name (setting errno) on error.  */
 static char const *
 sys_reset_uid_gid (void)
 {
@@ -367,25 +371,25 @@ sys_reset_uid_gid (void)
 
   if (!pw)
     return "getpwuid";
-  if (initgroups (pw->pw_name, gid) != 0 && errno != EPERM)
+  if (initgroups (pw->pw_name, gid) < 0 && errno != EPERM)
     return "initgroups";
-  if (gid != getegid () && setgid (gid) != 0 && errno != EPERM)
+  if (gid != getegid () && setgid (gid) < 0 && errno != EPERM)
     return "setgid";
-  if (uid != geteuid () && setuid (uid) != 0 && errno != EPERM)
+  if (uid != geteuid () && setuid (uid) < 0 && errno != EPERM)
     return "setuid";
 #endif
 
-  return NULL;
+  return nullptr;
 }
 
 /* Open a file (a magnetic tape device?) on the system specified in
-   FILE_NAME, as the given user. FILE_NAME has the form `[USER@]HOST:FILE'.
-   OPEN_MODE is O_RDONLY, O_WRONLY, etc.  If successful, return the
+   FILE_NAME, as the given user. FILE_NAME has the form '[USER@]HOST:FILE'.
+   OFLAGS is O_RDONLY, O_WRONLY, etc.  If successful, return the
    remote pipe number plus BIAS.  REMOTE_SHELL may be overridden.  On
    error, return -1.  */
 int
-rmt_open__ (const char *file_name, int open_mode, int bias,
-            const char *remote_shell)
+rmt_open (char const *file_name, int oflags, int bias,
+	  char const *remote_shell, char const *rmt_command)
 {
   int remote_pipe_number;	/* pseudo, biased file descriptor */
   char *file_name_copy;		/* copy of file_name string */
@@ -398,8 +402,8 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
   for (remote_pipe_number = 0;
        remote_pipe_number < MAXUNIT;
        remote_pipe_number++)
-    if (READ_SIDE (remote_pipe_number) == -1
-	&& WRITE_SIDE (remote_pipe_number) == -1)
+    if (read_side (remote_pipe_number) < 0
+	&& write_side (remote_pipe_number) < 0)
       break;
 
   if (remote_pipe_number == MAXUNIT)
@@ -415,8 +419,8 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
 
     file_name_copy = xstrdup (file_name);
     remote_host = file_name_copy;
-    remote_user = 0;
-    remote_file = 0;
+    remote_user = nullptr;
+    remote_file = nullptr;
 
     for (cursor = file_name_copy; *cursor; cursor++)
       switch (*cursor)
@@ -453,25 +457,23 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
   assume (remote_file);
 
   /* FIXME: Should somewhat validate the decoding, here.  */
-  if (gethostbyname (remote_host) == NULL)
-    error (EXIT_ON_EXEC_ERROR, 0, _("Cannot connect to %s: resolve failed"),
-	   remote_host);
 
   if (remote_user && *remote_user == '\0')
-    remote_user = 0;
+    remote_user = nullptr;
 
 #if WITH_REXEC
 
   /* Execute the remote command using rexec.  */
 
-  READ_SIDE (remote_pipe_number) = _rmt_rexec (remote_host, remote_user);
-  if (READ_SIDE (remote_pipe_number) < 0)
+  int fd = _rmt_rexec (remote_host, remote_user, rmt_command);
+  if (fd < 0)
     {
       free (file_name_copy);
       return -1;
     }
 
-  WRITE_SIDE (remote_pipe_number) = READ_SIDE (remote_pipe_number);
+  from_remote[remote_pipe_number][PREAD] = fd;
+  to_remote[remote_pipe_number][PWRITE] = fd;
 
 #else /* not WITH_REXEC */
   {
@@ -492,7 +494,7 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
       }
     remote_shell_basename = last_component (remote_shell);
 
-    /* Set up the pipes for the `rsh' command, and fork.  */
+    /* Set up the pipes for the 'rsh' command, and fork.  */
 
     if (pipe (to_remote[remote_pipe_number]) < 0)
       {
@@ -511,7 +513,7 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
       }
 
     status = fork ();
-    if (status == -1)
+    if (status < 0)
       {
 	int e = errno;
 	close (from_remote[remote_pipe_number][PREAD]);
@@ -529,12 +531,12 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
 
 	if (dup2 (to_remote[remote_pipe_number][PREAD], STDIN_FILENO) < 0
 	    || (to_remote[remote_pipe_number][PREAD] != STDIN_FILENO
-		&& close (to_remote[remote_pipe_number][PREAD]) != 0)
+		&& close (to_remote[remote_pipe_number][PREAD]) < 0)
 	    || (to_remote[remote_pipe_number][PWRITE] != STDIN_FILENO
-		&& close (to_remote[remote_pipe_number][PWRITE]) != 0)
+		&& close (to_remote[remote_pipe_number][PWRITE]) < 0)
 	    || dup2 (from_remote[remote_pipe_number][PWRITE], STDOUT_FILENO) < 0
-	    || close (from_remote[remote_pipe_number][PREAD]) != 0
-	    || close (from_remote[remote_pipe_number][PWRITE]) != 0)
+	    || close (from_remote[remote_pipe_number][PREAD]) < 0
+	    || close (from_remote[remote_pipe_number][PWRITE]) < 0)
 	  error (EXIT_ON_EXEC_ERROR, errno,
 		 _("Cannot redirect files for remote shell"));
 
@@ -543,12 +545,14 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
 	  error (EXIT_ON_EXEC_ERROR, errno,
 		 _("Cannot reset uid and gid: %s"), reseterr);
 
+	char const *cmd = rmt_command ? rmt_command : DEFAULT_RMT_COMMAND;
+
 	if (remote_user)
 	  execl (remote_shell, remote_shell_basename, remote_host,
-		 "-l", remote_user, rmt_command, (char *) 0);
+		 "-l", remote_user, cmd, nullptr);
 	else
 	  execl (remote_shell, remote_shell_basename, remote_host,
-		 rmt_command, (char *) 0);
+		 cmd, nullptr);
 
 	/* Bad problems if we get here.  */
 
@@ -566,20 +570,30 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
   /* Attempt to open the tape device.  */
 
   {
-    size_t remote_file_len = strlen (remote_file);
-    char *command_buffer = xmalloc (remote_file_len + 1000);
-    sprintf (command_buffer, "O%s\n", remote_file);
-    encode_oflag (command_buffer + remote_file_len + 2, open_mode);
-    strcat (command_buffer, "\n");
-    if (do_command (remote_pipe_number, command_buffer) == -1
-	|| get_status (remote_pipe_number) == -1)
+    idx_t remote_file_len = strlen (remote_file);
+    char *command_buffer = ximalloc (remote_file_len + 1000);
+    char *p = command_buffer;
+    *p++ = 'O';
+    p = mempcpy (p, remote_file, remote_file_len);
+    *p++ = '\n';
+    p = encode_oflags (p, oflags);
+    if (!p)
       {
 	free (command_buffer);
+	free (file_name_copy);
+	_rmt_shutdown (remote_pipe_number, EINVAL);
+	return -1;
+      }
+    *p++ = '\n';
+    int done = do_command (remote_pipe_number,
+			   command_buffer, p - command_buffer);
+    free (command_buffer);
+    if (done < 0 || get_status (remote_pipe_number, INTMAX_MAX) < 0)
+      {
 	free (file_name_copy);
 	_rmt_shutdown (remote_pipe_number, errno);
 	return -1;
       }
-    free (command_buffer);
   }
 
   free (file_name_copy);
@@ -589,42 +603,46 @@ rmt_open__ (const char *file_name, int open_mode, int bias,
 /* Close remote tape connection HANDLE and shut down.  Return 0 if
    successful, -1 on error.  */
 int
-rmt_close__ (int handle)
+rmt_close (int handle)
 {
-  long int status;
+  int done = do_command (handle, "C\n", 2);
+  if (done < 0)
+    return done;
 
-  if (do_command (handle, "C\n") == -1)
-    return -1;
-
-  status = get_status (handle);
+  int status = get_status (handle, 0);
   _rmt_shutdown (handle, errno);
   return status;
 }
 
-/* Read up to LENGTH bytes into BUFFER from remote tape connection HANDLE.
-   Return the number of bytes read on success, SAFE_READ_ERROR on error.  */
-size_t
-rmt_read__ (int handle, char *buffer, size_t length)
+/* Read from remote tape connection HANDLE into BUFFER, of size LENGTH.
+   Return the number of bytes read on success, -1 (setting errno) on error.  */
+ptrdiff_t
+rmt_read (int handle, void *buffer, idx_t length)
 {
-  char command_buffer[sizeof "R\n" + INT_STRLEN_BOUND (size_t)];
-  size_t status;
-  size_t rlen;
-  size_t counter;
+  char command_buffer[sizeof "R\n" + INT_STRLEN_BOUND (idx_t)];
+  int done = do_command (handle, command_buffer,
+			 sprintf (command_buffer, "R%td\n", length));
+  if (done < 0)
+    return done;
 
-  sprintf (command_buffer, "R%zu\n", length);
-  if (do_command (handle, command_buffer) == -1
-      || (status = get_status (handle)) == SAFE_READ_ERROR
-      || status > length)
-    return SAFE_READ_ERROR;
-
-  for (counter = 0; counter < status; counter += rlen, buffer += rlen)
+  ptrdiff_t status = get_status (handle, length);
+  if (status < 0)
     {
-      rlen = safe_read (READ_SIDE (handle), buffer, status - counter);
-      if (rlen == SAFE_READ_ERROR || rlen == 0)
+      _rmt_shutdown (handle, EIO);
+      return -1;
+    }
+
+  char *buf = buffer;
+  for (idx_t counter = 0; counter < status; )
+    {
+      ptrdiff_t rlen = safe_read (read_side (handle),
+				  buf + counter, status - counter);
+      if (rlen <= 0)
 	{
 	  _rmt_shutdown (handle, EIO);
-	  return SAFE_READ_ERROR;
+	  return -1;
 	}
+      counter += rlen;
     }
 
   return status;
@@ -632,23 +650,21 @@ rmt_read__ (int handle, char *buffer, size_t length)
 
 /* Write LENGTH bytes from BUFFER to remote tape connection HANDLE.
    Return the number of bytes written.  */
-size_t
-rmt_write__ (int handle, char *buffer, size_t length)
+idx_t
+rmt_write (int handle, void const *buffer, idx_t length)
 {
-  char command_buffer[sizeof "W\n" + INT_STRLEN_BOUND (size_t)];
+  char command_buffer[sizeof "W\n" + INT_STRLEN_BOUND (idx_t)];
   void (*pipe_handler) (int);
-  size_t written;
-
-  sprintf (command_buffer, "W%zu\n", length);
-  if (do_command (handle, command_buffer) == -1)
+  int buflen = sprintf (command_buffer, "W%td\n", length);
+  if (do_command (handle, command_buffer, buflen) < 0)
     return 0;
 
   pipe_handler = signal (SIGPIPE, SIG_IGN);
-  written = full_write (WRITE_SIDE (handle), buffer, length);
+  idx_t written = full_write (write_side (handle), buffer, length);
   signal (SIGPIPE, pipe_handler);
   if (written == length)
     {
-      long int r = get_status (handle);
+      ptrdiff_t r = get_status (handle, length);
       if (r < 0)
 	return 0;
       if (r == length)
@@ -665,30 +681,31 @@ rmt_write__ (int handle, char *buffer, size_t length)
 /* Perform an imitation lseek operation on remote tape connection
    HANDLE.  Return the new file offset if successful, -1 if on error.  */
 off_t
-rmt_lseek__ (int handle, off_t offset, int whence)
+rmt_lseek (int handle, off_t offset, int whence)
 {
   char command_buffer[sizeof "L0\n\n" + INT_STRLEN_BOUND (offset)];
+  intmax_t off = offset;
 
   switch (whence)
     {
     case SEEK_SET: whence = 0; break;
     case SEEK_CUR: whence = 1; break;
     case SEEK_END: whence = 2; break;
-    default: abort ();
+    default: errno = EINVAL; return -1;
     }
 
-  sprintf (command_buffer, "L%d\n%jd\n", whence, (intmax_t) offset);
+  int done = do_command (handle, command_buffer,
+			 sprintf (command_buffer, "L%d\n%jd\n", whence, off));
+  if (done < 0)
+    return done;
 
-  if (do_command (handle, command_buffer) == -1)
-    return -1;
-
-  return get_status_off (handle);
+  return get_status (handle, TYPE_MAXIMUM (off_t));
 }
 
 /* Perform a raw tape operation on remote tape connection HANDLE.
    Return the results of the ioctl, or -1 on error.  */
 int
-rmt_ioctl__ (int handle, unsigned long int operation, void *argument)
+rmt_ioctl (int handle, unsigned long int operation, void *argument)
 {
   switch (operation)
     {
@@ -707,19 +724,20 @@ rmt_ioctl__ (int handle, unsigned long int operation, void *argument)
 	/* MTIOCTOP is the easy one.  Nothing is transferred in binary.  */
 
 	intmax_t count = mtop->mt_count;
-	sprintf (command_buffer, "I%d\n%jd\n", mtop->mt_op, count);
-	if (do_command (handle, command_buffer) == -1)
-	  return -1;
+	int done = do_command (handle, command_buffer,
+			       sprintf (command_buffer, "I%d\n%jd\n",
+					mtop->mt_op, count));
+	if (done < 0)
+	  return done;
 
-	return get_status (handle);
+	return get_status (handle, INT_MAX);
       }
 #endif /* MTIOCTOP */
 
 #ifdef MTIOCGET
     case MTIOCGET:
       {
-	ssize_t status;
-	size_t counter;
+	struct mtget *mtget = argument;
 
 	/* Grab the status and read it directly into the structure.  This
 	   assumes that the status buffer is not padded and that 2 shorts
@@ -727,24 +745,28 @@ rmt_ioctl__ (int handle, unsigned long int operation, void *argument)
 	   whole struct is contiguous.  NOTE - this is probably NOT a good
 	   assumption.  */
 
-	if (do_command (handle, "S") == -1
-	    || (status = get_status (handle), status == -1))
-	  return -1;
-
-	if (status > sizeof (struct mtop))
+	int done = do_command (handle, "S\n", 2);
+	if (done < 0)
+	  return done;
+	ptrdiff_t status = get_status (handle, sizeof *mtget);
+	if (status < 0)
+	  return status;
+	if (status != sizeof *mtget)
 	  {
-	    errno = EOVERFLOW;
+	    _rmt_shutdown (handle, EIO);
 	    return -1;
 	  }
 
-	for (char *p = argument; status > 0; status -= counter, p += counter)
+	for (char *p = argument; status > 0; )
 	  {
-	    counter = safe_read (READ_SIDE (handle), p, status);
-	    if (counter == SAFE_READ_ERROR || counter == 0)
+	    ptrdiff_t rlen = safe_read (read_side (handle), p, status);
+	    if (rlen <= 0)
 	      {
 		_rmt_shutdown (handle, EIO);
 		return -1;
 	      }
+	    status -= rlen;
+	    p += rlen;
 	  }
 
 	/* Check for byte position.  mt_type (or mt_model) is a small integer
@@ -752,12 +774,12 @@ rmt_ioctl__ (int handle, unsigned long int operation, void *argument)
 	   than 256, we will assume that the bytes are swapped and go through
 	   and reverse all the bytes.  */
 
-	struct mtget *mtget = argument;
 	if (mtget->MTIO_CHECK_FIELD < 256)
 	  return 0;
 
 	char *buf = argument;
-	for (counter = 0; counter < status; counter += 2)
+	static_assert (sizeof *mtget % 2 == 0);
+	for (idx_t counter = 0; counter < sizeof *mtget; counter += 2)
 	  {
 	    char copy = buf[counter];
 
