@@ -1,6 +1,6 @@
 /* Create a tar archive.
 
-   Copyright 1985-2023 Free Software Foundation, Inc.
+   Copyright 1985-2025 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -24,6 +24,7 @@
 #include <areadlink.h>
 #include <flexmember.h>
 #include <quotearg.h>
+#include <same-inode.h>
 
 #include "common.h"
 #include <hash.h>
@@ -34,8 +35,8 @@ enum { IMPOSTOR_ERRNO = ENOENT };
 
 struct link
   {
-    dev_t dev;
-    ino_t ino;
+    dev_t st_dev;
+    ino_t st_ino;
     nlink_t nlink;
     char name[FLEXIBLE_ARRAY_MEMBER];
   };
@@ -43,7 +44,7 @@ struct link
 struct exclusion_tag
 {
   const char *name;
-  size_t length;
+  idx_t length;
   enum exclusion_tag_type type;
   bool (*predicate) (int fd);
   struct exclusion_tag *next;
@@ -69,12 +70,11 @@ exclusion_tag_warning (const char *dirname, const char *tagname,
 		       const char *message)
 {
   if (verbose_option)
-    WARNOPT (WARN_CACHEDIR,
-	     (0, 0,
-	      _("%s: contains a cache directory tag %s; %s"),
-	      quotearg_colon (dirname),
-	      quotearg_n (1, tagname),
-	      message));
+    warnopt (WARN_CACHEDIR, 0,
+	     _("%s: contains a cache directory tag %s; %s"),
+	     quotearg_colon (dirname),
+	     quotearg_n (1, tagname),
+	     message);
 }
 
 enum exclusion_tag_type
@@ -103,50 +103,55 @@ check_exclusion_tags (struct tar_stat_info const *st, char const **tag_file_name
 
 /* Exclusion predicate to test if the named file (usually "CACHEDIR.TAG")
    contains a valid header, as described at:
-	http://www.brynosaurus.com/cachedir
+	https://bford.info/cachedir/
    Applications can write this file into directories they create
    for use as caches containing purely regenerable, non-precious data,
    allowing us to avoid archiving them if --exclude-caches is specified. */
 
-#define CACHEDIR_SIGNATURE "Signature: 8a477f597d28d172789f06886806bc55"
-#define CACHEDIR_SIGNATURE_SIZE (sizeof CACHEDIR_SIGNATURE - 1)
-
 bool
 cachedir_file_p (int fd)
 {
-  char tagbuf[CACHEDIR_SIGNATURE_SIZE];
-
-  return
-    (read (fd, tagbuf, CACHEDIR_SIGNATURE_SIZE) == CACHEDIR_SIGNATURE_SIZE
-     && memcmp (tagbuf, CACHEDIR_SIGNATURE, CACHEDIR_SIGNATURE_SIZE) == 0);
+  static char const sig[43] ATTRIBUTE_NONSTRING
+    = "Signature: 8a477f597d28d172789f06886806bc55";
+  char tagbuf[sizeof sig];
+  return (read (fd, tagbuf, sizeof sig) == sizeof sig
+	  && memcmp (tagbuf, sig, sizeof sig) == 0);
 }
 
 
 /* The maximum uintmax_t value that can be represented with DIGITS digits,
    assuming that each digit is BITS_PER_DIGIT wide.  */
-#define MAX_VAL_WITH_DIGITS(digits, bits_per_digit) \
-   ((digits) * (bits_per_digit) < sizeof (uintmax_t) * CHAR_BIT \
-    ? ((uintmax_t) 1 << ((digits) * (bits_per_digit))) - 1 \
-    : (uintmax_t) -1)
+static uintmax_t
+max_val_with_digits (int digits, int bits_per_digit)
+{
+  uintmax_t one = 1;
+  return (digits * bits_per_digit < UINTMAX_WIDTH
+	  ? (one << (digits * bits_per_digit)) - 1
+	  : UINTMAX_MAX);
+}
 
 /* The maximum uintmax_t value that can be represented with octal
-   digits and a trailing NUL in BUFFER.  */
-#define MAX_OCTAL_VAL(buffer) MAX_VAL_WITH_DIGITS (sizeof (buffer) - 1, LG_8)
+   digits and a trailing NUL in a buffer of size BUFSIZE.  */
+static uintmax_t
+max_octal_val (int bufsize)
+{
+  return max_val_with_digits (bufsize - 1, LG_8);
+}
 
 /* Convert VALUE to an octal representation suitable for tar headers.
    Output to buffer WHERE with size SIZE.
    The result is undefined if SIZE is 0 or if VALUE is too large to fit.  */
 
 static void
-to_octal (uintmax_t value, char *where, size_t size)
+to_octal (uintmax_t value, char *where, int size)
 {
   uintmax_t v = value;
-  size_t i = size;
+  int i = size;
 
   do
     {
-      where[--i] = '0' + (v & ((1 << LG_8) - 1));
-      v >>= LG_8;
+      where[--i] = '0' + v % 8;
+      v /= 8;
     }
   while (i);
 }
@@ -155,10 +160,9 @@ to_octal (uintmax_t value, char *where, size_t size)
    NUL unless SRC is LEN or more bytes long.  */
 
 static void
-tar_copy_str (char *dst, const char *src, size_t len)
+tar_copy_str (char *dst, const char *src, int len)
 {
-  size_t i;
-  for (i = 0; i < len; i++)
+  for (int i = 0; i < len; i++)
     if (! (dst[i] = src[i]))
       break;
 }
@@ -167,11 +171,11 @@ tar_copy_str (char *dst, const char *src, size_t len)
    is OLDGNU format */
 
 static void
-tar_name_copy_str (char *dst, const char *src, size_t len)
+tar_name_copy_str (char *dst, const char *src, int len)
 {
   tar_copy_str (dst, src, len);
   if (archive_format == OLDGNU_FORMAT)
-    dst[len-1] = 0;
+    dst[len - 1] = 0;
 }
 
 /* Convert NEGATIVE VALUE to a base-256 representation suitable for
@@ -181,17 +185,17 @@ tar_name_copy_str (char *dst, const char *src, size_t len)
    fit.  */
 
 static void
-to_base256 (bool negative, uintmax_t value, char *where, size_t size)
+to_base256 (bool negative, uintmax_t value, char *where, int size)
 {
   uintmax_t v = value;
-  uintmax_t propagated_sign_bits =
-    ((uintmax_t) - negative << (CHAR_BIT * sizeof v - LG_256));
-  size_t i = size;
+  uintmax_t sign_bits = - negative;
+  uintmax_t propagated_sign_bits = sign_bits << (UINTMAX_WIDTH - LG_256);
+  int i = size;
 
   do
     {
-      where[--i] = v & ((1 << LG_256) - 1);
-      v = propagated_sign_bits | (v >> LG_256);
+      where[--i] = v % 256;
+      v = propagated_sign_bits | (v / 256);
     }
   while (i);
 }
@@ -206,43 +210,24 @@ to_base256 (bool negative, uintmax_t value, char *where, size_t size)
 #define GNAME_TO_CHARS(name, buf) string_to_chars (name, buf, sizeof (buf))
 
 static bool
-to_chars (bool negative, uintmax_t value, size_t valsize,
+to_chars (bool negative, uintmax_t value, int valsize,
 	  uintmax_t (*substitute) (bool *),
-	  char *where, size_t size, const char *type);
+	  char *where, int size, const char *type);
 
 static bool
-to_chars_subst (bool negative, bool gnu_format, uintmax_t value, size_t valsize,
+to_chars_subst (bool negative, bool gnu_format, uintmax_t value, int valsize,
 		uintmax_t (*substitute) (bool *),
-		char *where, size_t size, const char *type)
+		char *where, int size, const char *type)
 {
   uintmax_t maxval = (gnu_format
-		      ? MAX_VAL_WITH_DIGITS (size - 1, LG_256)
-		      : MAX_VAL_WITH_DIGITS (size - 1, LG_8));
-  char valbuf[UINTMAX_STRSIZE_BOUND + 1];
-  char maxbuf[UINTMAX_STRSIZE_BOUND];
-  char minbuf[UINTMAX_STRSIZE_BOUND + 1];
-  char const *minval_string;
-  char const *maxval_string = STRINGIFY_BIGINT (maxval, maxbuf);
-  char const *value_string;
-
-  if (gnu_format)
-    {
-      uintmax_t m = maxval + 1 ? maxval + 1 : maxval / 2 + 1;
-      char *p = STRINGIFY_BIGINT (m, minbuf + 1);
-      *--p = '-';
-      minval_string = p;
-    }
-  else
-    minval_string = "0";
-
+		      ? max_val_with_digits (size - 1, LG_256)
+		      : max_val_with_digits (size - 1, LG_8));
+  intmax_t minval = (!gnu_format ? 0
+		     : ckd_sub (&minval, -1, maxval) ? INTMAX_MIN
+		     : minval);
+  char const *valuesign = &"-"[!negative];
   if (negative)
-    {
-      char *p = STRINGIFY_BIGINT (- value, valbuf + 1);
-      *--p = '-';
-      value_string = p;
-    }
-  else
-    value_string = STRINGIFY_BIGINT (value, valbuf);
+    value = -value;
 
   if (substitute)
     {
@@ -256,18 +241,15 @@ to_chars_subst (bool negative, bool gnu_format, uintmax_t value, size_t valsize,
 
 	 Apart from this they are completely identical. */
       uintmax_t s = (negsub &= archive_format == GNU_FORMAT) ? - sub : sub;
-      char subbuf[UINTMAX_STRSIZE_BOUND + 1];
-      char *sub_string = STRINGIFY_BIGINT (s, subbuf + 1);
-      if (negsub)
-	*--sub_string = '-';
-      WARN ((0, 0, _("value %s out of %s range %s..%s; substituting %s"),
-	     value_string, type, minval_string, maxval_string,
-	     sub_string));
+      char const *ssign = &"-"[!negsub];
+      paxwarn (0, _("value %s%ju out of %s range %jd..%ju;"
+		    " substituting %s%ju"),
+	       valuesign, value, type, minval, maxval, ssign, s);
       return to_chars (negsub, s, valsize, 0, where, size, type);
     }
   else
-    ERROR ((0, 0, _("value %s out of %s range %s..%s"),
-	    value_string, type, minval_string, maxval_string));
+    paxerror (0, _("value %s%ju out of %s range %jd..%ju"),
+	      valuesign, value, type, minval, maxval);
   return false;
 }
 
@@ -286,15 +268,15 @@ to_chars_subst (bool negative, bool gnu_format, uintmax_t value, size_t valsize,
    substitute value is negative.  */
 
 static bool
-to_chars (bool negative, uintmax_t value, size_t valsize,
+to_chars (bool negative, uintmax_t value, int valsize,
 	  uintmax_t (*substitute) (bool *),
-	  char *where, size_t size, const char *type)
+	  char *where, int size, const char *type)
 {
   bool gnu_format = (archive_format == GNU_FORMAT
 		     || archive_format == OLDGNU_FORMAT);
 
   /* Generate the POSIX octal representation if the number fits.  */
-  if (! negative && value <= MAX_VAL_WITH_DIGITS (size - 1, LG_8))
+  if (! negative && value <= max_val_with_digits (size - 1, LG_8))
     {
       where[size - 1] = '\0';
       to_octal (value, where, size - 1);
@@ -307,7 +289,7 @@ to_chars (bool negative, uintmax_t value, size_t valsize,
 
       /* Generate the base-256 representation if the number fits.  */
       if (((negative ? -1 - value : value)
-	   <= MAX_VAL_WITH_DIGITS (size - 1, LG_256)))
+	   <= max_val_with_digits (size - 1, LG_256)))
 	{
 	  where[0] = (char) (negative ? -1 : 1 << (LG_256 - 1));
 	  to_base256 (negative, value, where + 1, size - 1);
@@ -326,10 +308,10 @@ to_chars (bool negative, uintmax_t value, size_t valsize,
 	  if (! warned_once)
 	    {
 	      warned_once = true;
-	      WARN ((0, 0, _("Generating negative octal headers")));
+	      paxwarn (0, _("Generating negative octal headers"));
 	    }
 	  where[size - 1] = '\0';
-	  to_octal (value & MAX_VAL_WITH_DIGITS (valsize * CHAR_BIT, 1),
+	  to_octal (value & max_val_with_digits (valsize * CHAR_BIT, 1),
 		    where, size - 1);
 	  return true;
 	}
@@ -359,25 +341,25 @@ gid_substitute (bool *negative)
 }
 
 static bool
-gid_to_chars (gid_t v, char *p, size_t s)
+gid_to_chars (gid_t v, char *p, int s)
 {
-  return to_chars (v < 0, (uintmax_t) v, sizeof v, gid_substitute, p, s, "gid_t");
+  return to_chars (v < 0, v, sizeof v, gid_substitute, p, s, "gid_t");
 }
 
 static bool
-major_to_chars (major_t v, char *p, size_t s)
+major_to_chars (major_t v, char *p, int s)
 {
-  return to_chars (v < 0, (uintmax_t) v, sizeof v, 0, p, s, "major_t");
+  return to_chars (v < 0, v, sizeof v, 0, p, s, "major_t");
 }
 
 static bool
-minor_to_chars (minor_t v, char *p, size_t s)
+minor_to_chars (minor_t v, char *p, int s)
 {
-  return to_chars (v < 0, (uintmax_t) v, sizeof v, 0, p, s, "minor_t");
+  return to_chars (v < 0, v, sizeof v, 0, p, s, "minor_t");
 }
 
 static bool
-mode_to_chars (mode_t v, char *p, size_t s)
+mode_to_chars (mode_t v, char *p, int s)
 {
   /* In the common case where the internal and external mode bits are the same,
      and we are not using POSIX or GNU format,
@@ -417,15 +399,15 @@ mode_to_chars (mode_t v, char *p, size_t s)
 }
 
 bool
-off_to_chars (off_t v, char *p, size_t s)
+off_to_chars (off_t v, char *p, int s)
 {
-  return to_chars (v < 0, (uintmax_t) v, sizeof v, 0, p, s, "off_t");
+  return to_chars (v < 0, v, sizeof v, 0, p, s, "off_t");
 }
 
 bool
-time_to_chars (time_t v, char *p, size_t s)
+time_to_chars (time_t v, char *p, int s)
 {
-  return to_chars (v < 0, (uintmax_t) v, sizeof v, 0, p, s, "time_t");
+  return to_chars (v < 0, v, sizeof v, 0, p, s, "time_t");
 }
 
 static uintmax_t
@@ -445,19 +427,13 @@ uid_substitute (bool *negative)
 }
 
 static bool
-uid_to_chars (uid_t v, char *p, size_t s)
+uid_to_chars (uid_t v, char *p, int s)
 {
-  return to_chars (v < 0, (uintmax_t) v, sizeof v, uid_substitute, p, s, "uid_t");
-}
-
-static bool
-uintmax_to_chars (uintmax_t v, char *p, size_t s)
-{
-  return to_chars (false, v, sizeof v, 0, p, s, "uintmax_t");
+  return to_chars (v < 0, v, sizeof v, uid_substitute, p, s, "uid_t");
 }
 
 static void
-string_to_chars (char const *str, char *p, size_t s)
+string_to_chars (char const *str, char *p, int s)
 {
   tar_copy_str (p, str, s);
   p[s - 1] = '\0';
@@ -498,13 +474,13 @@ write_eot (void)
   memset (pointer->buffer, 0, BLOCKSIZE);
   set_next_block_after (pointer);
   pointer = find_next_block ();
-  memset (pointer->buffer, 0, available_space_after (pointer));
+  memset (charptr (pointer), 0, available_space_after (pointer));
   set_next_block_after (pointer);
 }
 
 /* Write a "private" header */
 union block *
-start_private_header (const char *name, size_t size, time_t t)
+start_private_header (const char *name, idx_t size, time_t t)
 {
   union block *header = find_next_block ();
 
@@ -513,7 +489,8 @@ start_private_header (const char *name, size_t size, time_t t)
   tar_name_copy_str (header->header.name, name, NAME_FIELD_SIZE);
   OFF_TO_CHARS (size, header->header.size);
 
-  TIME_TO_CHARS (t < 0 ? 0 : min (t, MAX_OCTAL_VAL (header->header.mtime)),
+  TIME_TO_CHARS ((t < 0 ? 0
+		  : min (t, max_octal_val (sizeof header->header.mtime))),
 		 header->header.mtime);
   MODE_TO_CHARS (S_IFREG|S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, header->header.mode);
   UID_TO_CHARS (0, header->header.uid);
@@ -539,11 +516,8 @@ write_short_name (struct tar_stat_info *st)
 static void
 write_gnu_long_link (struct tar_stat_info *st, const char *p, char type)
 {
-  size_t size = strlen (p) + 1;
-  size_t bufsize;
-  union block *header;
-
-  header = start_private_header ("././@LongLink", size, 0);
+  idx_t size = strlen (p) + 1;
+  union block *header = start_private_header ("././@LongLink", size, 0);
   if (! numeric_owner_option)
     {
       static char *uname, *gname;
@@ -563,26 +537,26 @@ write_gnu_long_link (struct tar_stat_info *st, const char *p, char type)
 
   header = find_next_block ();
 
-  bufsize = available_space_after (header);
+  idx_t bufsize = available_space_after (header);
 
   while (bufsize < size)
     {
-      memcpy (header->buffer, p, bufsize);
+      memcpy (charptr (header), p, bufsize);
       p += bufsize;
       size -= bufsize;
-      set_next_block_after (header + (bufsize - 1) / BLOCKSIZE);
+      set_next_block_after (charptr (header) + bufsize - 1);
       header = find_next_block ();
       bufsize = available_space_after (header);
     }
-  memcpy (header->buffer, p, size);
-  memset (header->buffer + size, 0, bufsize - size);
-  set_next_block_after (header + (size - 1) / BLOCKSIZE);
+  memcpy (charptr (header), p, size);
+  memset (charptr (header) + size, 0, bufsize - size);
+  set_next_block_after (charptr (header) + size - 1);
 }
 
-static size_t
-split_long_name (const char *name, size_t length)
+static int
+split_long_name (const char *name, idx_t length)
 {
-  size_t i;
+  int i;
 
   if (length > PREFIX_FIELD_SIZE + 1)
     length = PREFIX_FIELD_SIZE + 1;
@@ -597,28 +571,26 @@ split_long_name (const char *name, size_t length)
 static union block *
 write_ustar_long_name (const char *name)
 {
-  size_t length = strlen (name);
-  size_t i, nlen;
-  union block *header;
+  idx_t length = strlen (name);
 
   if (length > PREFIX_FIELD_SIZE + NAME_FIELD_SIZE + 1)
     {
-      ERROR ((0, 0, _("%s: file name is too long (max %d); not dumped"),
-	      quotearg_colon (name),
-	      PREFIX_FIELD_SIZE + NAME_FIELD_SIZE + 1));
+      paxerror (0, _("%s: file name is too long (max %d); not dumped"),
+		quotearg_colon (name),
+		PREFIX_FIELD_SIZE + NAME_FIELD_SIZE + 1);
       return NULL;
     }
 
-  i = split_long_name (name, length);
-  if (i == 0 || (nlen = length - i - 1) > NAME_FIELD_SIZE || nlen == 0)
+  int i = split_long_name (name, length);
+  idx_t nlen = length - i - 1;
+  if (i == 0 || ! (0 < nlen && nlen <= NAME_FIELD_SIZE))
     {
-      ERROR ((0, 0,
-	      _("%s: file name is too long (cannot be split); not dumped"),
-	      quotearg_colon (name)));
+      paxerror (0, _("%s: file name is too long (cannot be split); not dumped"),
+		quotearg_colon (name));
       return NULL;
     }
 
-  header = find_next_block ();
+  union block *header = find_next_block ();
   memset (header->buffer, 0, sizeof (header->buffer));
   memcpy (header->header.prefix, name, i);
   memcpy (header->header.name, name + i + 1, length - i - 1);
@@ -639,9 +611,8 @@ write_long_link (struct tar_stat_info *st)
     case V7_FORMAT:			/* old V7 tar format */
     case USTAR_FORMAT:
     case STAR_FORMAT:
-      ERROR ((0, 0,
-	      _("%s: link name is too long; not dumped"),
-	      quotearg_colon (st->link_name)));
+      paxerror (0, _("%s: link name is too long; not dumped"),
+		quotearg_colon (st->link_name));
       break;
 
     case OLDGNU_FORMAT:
@@ -666,9 +637,8 @@ write_long_name (struct tar_stat_info *st)
     case V7_FORMAT:
       if (strlen (st->file_name) > NAME_FIELD_SIZE-1)
 	{
-	  ERROR ((0, 0, _("%s: file name is too long (max %d); not dumped"),
-		  quotearg_colon (st->file_name),
-		  NAME_FIELD_SIZE - 1));
+	  paxerror (0, _("%s: file name is too long (max %d); not dumped"),
+		    quotearg_colon (st->file_name), NAME_FIELD_SIZE - 1);
 	  return NULL;
 	}
       break;
@@ -793,7 +763,7 @@ start_header (struct tar_stat_info *st)
   {
     uid_t uid = st->stat.st_uid;
     if (archive_format == POSIX_FORMAT
-	&& MAX_OCTAL_VAL (header->header.uid) < uid)
+	&& max_octal_val (sizeof header->header.uid) < uid)
       {
 	xheader_store ("uid", st, NULL);
 	uid = 0;
@@ -805,7 +775,7 @@ start_header (struct tar_stat_info *st)
   {
     gid_t gid = st->stat.st_gid;
     if (archive_format == POSIX_FORMAT
-	&& MAX_OCTAL_VAL (header->header.gid) < gid)
+	&& max_octal_val (sizeof header->header.gid) < gid)
       {
 	xheader_store ("gid", st, NULL);
 	gid = 0;
@@ -817,7 +787,7 @@ start_header (struct tar_stat_info *st)
   {
     off_t size = st->stat.st_size;
     if (archive_format == POSIX_FORMAT
-	&& MAX_OCTAL_VAL (header->header.size) < size)
+	&& max_octal_val (sizeof header->header.size) < size)
       {
 	xheader_store ("size", st, NULL);
 	size = 0;
@@ -843,14 +813,23 @@ start_header (struct tar_stat_info *st)
 	mtime = timespec_cmp (st->mtime, mtime_option) > 0
 	           ? mtime_option : st->mtime;
 	break;
+
+      case COMMAND_MTIME:
+	if (!sys_exec_setmtime_script (set_mtime_command,
+				       chdir_fd,
+				       st->orig_file_name,
+				       set_mtime_format,
+				       &mtime))
+	  mtime = st->mtime;
+	break;
       }
 
     if (archive_format == POSIX_FORMAT)
       {
-	if (MAX_OCTAL_VAL (header->header.mtime) < mtime.tv_sec
+	if (max_octal_val (sizeof header->header.mtime) < mtime.tv_sec
 	    || mtime.tv_nsec != 0)
 	  xheader_store ("mtime", st, &mtime);
-	if (MAX_OCTAL_VAL (header->header.mtime) < mtime.tv_sec)
+	if (max_octal_val (sizeof header->header.mtime) < mtime.tv_sec)
 	  mtime.tv_sec = 0;
       }
     if (!TIME_TO_CHARS (mtime.tv_sec, header->header.mtime))
@@ -865,7 +844,7 @@ start_header (struct tar_stat_info *st)
       minor_t devminor = minor (st->stat.st_rdev);
 
       if (archive_format == POSIX_FORMAT
-	  && MAX_OCTAL_VAL (header->header.devmajor) < devmajor)
+	  && max_octal_val (sizeof header->header.devmajor) < devmajor)
 	{
 	  xheader_store ("devmajor", st, NULL);
 	  devmajor = 0;
@@ -874,7 +853,7 @@ start_header (struct tar_stat_info *st)
 	return NULL;
 
       if (archive_format == POSIX_FORMAT
-	  && MAX_OCTAL_VAL (header->header.devminor) < devminor)
+	  && max_octal_val (sizeof header->header.devminor) < devminor)
 	{
 	  xheader_store ("devminor", st, NULL);
 	  devminor = 0;
@@ -959,12 +938,9 @@ start_header (struct tar_stat_info *st)
         }
       if ((selinux_context_option > 0) && st->cntx_name)
         xheader_store ("RHT.security.selinux", st, NULL);
-      if (xattrs_option > 0)
-        {
-          size_t i;
-	  for (i = 0; i < st->xattr_map.xm_size; i++)
-	    xheader_store (st->xattr_map.xm_map[i].xkey, st, &i);
-        }
+      if (xattrs_option)
+	for (idx_t i = 0; i < st->xattr_map.xm_size; i++)
+	  xheader_store (st->xattr_map.xm_map[i].xkey, st, &i);
     }
 
   return header;
@@ -973,15 +949,12 @@ start_header (struct tar_stat_info *st)
 void
 simple_finish_header (union block *header)
 {
-  size_t i;
-  int sum;
-  char *p;
+  /* Fill checksum field with spaces while the checksum is computed.  */
+  memset (header->header.chksum, ' ', sizeof header->header.chksum);
 
-  memcpy (header->header.chksum, CHKBLANKS, sizeof header->header.chksum);
-
-  sum = 0;
-  p = header->buffer;
-  for (i = sizeof *header; i-- != 0; )
+  int sum = 0;
+  char *p = header->buffer;
+  for (int i = sizeof *header; i-- != 0; )
     /* We can't use unsigned char here because of old compilers, e.g. V7.  */
     sum += 0xFF & *p++;
 
@@ -993,9 +966,10 @@ simple_finish_header (union block *header)
 
      This is a fast way to do:
 
-     sprintf(header->header.chksum, "%6o", sum);  */
+     sprintf (header->header.chksum, "%6o", sum);  */
 
-  uintmax_to_chars ((uintmax_t) sum, header->header.chksum, 7);
+  header->header.chksum[6] = '\0';
+  to_octal (sum, header->header.chksum, 6);
 
   set_next_block_after (header);
 }
@@ -1060,43 +1034,38 @@ dump_regular_file (int fd, struct tar_stat_info *st)
   mv_begin_write (st->file_name, st->stat.st_size, st->stat.st_size);
   while (size_left > 0)
     {
-      size_t bufsize, count;
-
       blk = find_next_block ();
 
-      bufsize = available_space_after (blk);
+      idx_t bufsize = available_space_after (blk);
 
       if (size_left < bufsize)
 	{
 	  /* Last read -- zero out area beyond.  */
 	  bufsize = size_left;
-	  count = bufsize % BLOCKSIZE;
-	  if (count)
-	    memset (blk->buffer + size_left, 0, BLOCKSIZE - count);
+	  idx_t beyond = bufsize & (BLOCKSIZE - 1);
+	  if (beyond)
+	    memset (blk->buffer + size_left, 0, BLOCKSIZE - beyond);
 	}
 
-      count = (fd <= 0) ? bufsize : blocking_read (fd, blk->buffer, bufsize);
-      if (count == SAFE_READ_ERROR)
-	{
-	  read_diag_details (st->orig_file_name,
-	                     st->stat.st_size - size_left, bufsize);
-	  pad_archive (size_left);
-	  return dump_status_short;
-	}
+      idx_t count = (fd <= 0 ? bufsize
+		     : blocking_read (fd, charptr (blk), bufsize));
       size_left -= count;
-      set_next_block_after (blk + (bufsize - 1) / BLOCKSIZE);
+      set_next_block_after (charptr (blk) + bufsize - 1);
 
       if (count != bufsize)
 	{
-	  char buf[UINTMAX_STRSIZE_BOUND];
-	  memset (blk->buffer + count, 0, bufsize - count);
-	  WARNOPT (WARN_FILE_SHRANK,
-		   (0, 0,
-		    ngettext ("%s: File shrank by %s byte; padding with zeros",
-			      "%s: File shrank by %s bytes; padding with zeros",
-			      size_left),
-		    quotearg_colon (st->orig_file_name),
-		    STRINGIFY_BIGINT (size_left, buf)));
+	  if (errno)
+	    read_diag_details (st->orig_file_name,
+			       st->stat.st_size - size_left, bufsize);
+	  memset (charptr (blk) + count, 0, bufsize - count);
+	  warnopt (WARN_FILE_SHRANK, 0,
+		   ngettext (("%s: File shrank by %jd byte;"
+			      " padding with zeros"),
+			     ("%s: File shrank by %jd bytes;"
+			      " padding with zeros"),
+			     size_left),
+		   quotearg_colon (st->orig_file_name),
+		   intmax (size_left));
 	  if (! ignore_failed_read_option)
 	    set_exit_status (TAREXIT_DIFFERS);
 	  pad_archive (size_left - (bufsize - count));
@@ -1145,36 +1114,31 @@ dump_dir0 (struct tar_stat_info *st, char const *directory)
 	}
       else
 	{
-	  off_t size_left;
-	  off_t totsize;
-	  size_t bufsize;
-	  ssize_t count;
-	  const char *buffer, *p_buffer;
-
 	  block_ordinal = current_block_ordinal ();
-	  buffer = safe_directory_contents (gnu_list_name->directory);
-	  totsize = dumpdir_size (buffer);
+	  char const *buffer
+	    = safe_directory_contents (gnu_list_name->directory);
+	  off_t totsize = dumpdir_size (buffer);
 	  OFF_TO_CHARS (totsize, blk->header.size);
 	  finish_header (st, blk, block_ordinal);
-	  p_buffer = buffer;
-	  size_left = totsize;
+	  char const *p_buffer = buffer;
+	  off_t size_left = totsize;
 
 	  mv_begin_write (st->file_name, totsize, totsize);
 	  while (size_left > 0)
 	    {
 	      blk = find_next_block ();
-	      bufsize = available_space_after (blk);
+	      idx_t bufsize = available_space_after (blk);
 	      if (size_left < bufsize)
 		{
 		  bufsize = size_left;
-		  count = bufsize % BLOCKSIZE;
+		  idx_t count = bufsize & (BLOCKSIZE - 1);
 		  if (count)
 		    memset (blk->buffer + size_left, 0, BLOCKSIZE - count);
 		}
-	      memcpy (blk->buffer, p_buffer, bufsize);
+	      memcpy (charptr (blk), p_buffer, bufsize);
 	      size_left -= bufsize;
 	      p_buffer += bufsize;
-	      set_next_block_after (blk + (bufsize - 1) / BLOCKSIZE);
+	      set_next_block_after (charptr (blk) + bufsize - 1);
 	    }
 	}
       return;
@@ -1188,15 +1152,14 @@ dump_dir0 (struct tar_stat_info *st, char const *directory)
       && st->parent->stat.st_dev != st->stat.st_dev)
     {
       if (verbose_option)
-	WARNOPT (WARN_XDEV,
-		 (0, 0,
-		  _("%s: file is on a different filesystem; not dumped"),
-		  quotearg_colon (st->orig_file_name)));
+	warnopt (WARN_XDEV, 0,
+		 _("%s: file is on a different filesystem; not dumped"),
+		 quotearg_colon (st->orig_file_name));
     }
   else
     {
       char *name_buf;
-      size_t name_size;
+      idx_t name_size;
 
       switch (check_exclusion_tags (st, &tag_file_name))
 	{
@@ -1206,15 +1169,13 @@ dump_dir0 (struct tar_stat_info *st, char const *directory)
 
 	case exclusion_tag_none:
 	  {
-	    char const *entry;
-	    size_t entry_len;
-	    size_t name_len;
-
 	    name_buf = xstrdup (st->orig_file_name);
-	    name_size = name_len = strlen (name_buf);
+	    idx_t name_len = name_size = strlen (name_buf);
 
 	    /* Now output all the files in the directory.  */
-	    for (entry = directory; (entry_len = strlen (entry)) != 0;
+	    idx_t entry_len;
+	    for (char const *entry = directory;
+		 (entry_len = strlen (entry)) != 0;
 		 entry += entry_len + 1)
 	      {
 		if (name_size < name_len + entry_len)
@@ -1254,7 +1215,7 @@ dump_dir0 (struct tar_stat_info *st, char const *directory)
 static void
 ensure_slash (char **pstr)
 {
-  size_t len = strlen (*pstr);
+  idx_t len = strlen (*pstr);
   while (len >= 1 && ISSLASH ((*pstr)[len - 1]))
     len--;
   if (!ISSLASH ((*pstr)[len]))
@@ -1345,7 +1306,7 @@ create_archive (void)
 
   if (incremental_option)
     {
-      size_t buffer_size = 0;
+      idx_t buffer_size = 0;
       char *buffer = NULL;
       const char *q;
 
@@ -1360,9 +1321,10 @@ create_archive (void)
 	if (!excluded_name (p->name, NULL))
 	  {
 	    struct tar_stat_info st;
-	    size_t plen = strlen (p->name);
-	    while (buffer_size <= plen)
-	      buffer = x2realloc (buffer, &buffer_size);
+	    idx_t plen = strlen (p->name);
+	    if (buffer_size <= plen)
+	      buffer = xpalloc (buffer, &buffer_size,
+				plen - buffer_size + 1, -1, 1);
 	    memcpy (buffer, p->name, plen);
 	    if (! ISSLASH (buffer[plen - 1]))
 	      buffer[plen++] = DIRECTORY_SEPARATOR;
@@ -1371,7 +1333,7 @@ create_archive (void)
 	    if (q)
 	      while (*q)
 		{
-		  size_t qlen = strlen (q);
+		  idx_t qlen = strlen (q);
 		  if (*q == 'Y')
 		    {
 		      if (! st.orig_file_name)
@@ -1385,7 +1347,7 @@ create_archive (void)
 			      break;
 			    }
 			  st.fd = fd;
-			  if (fstat (fd, &st.stat) != 0)
+			  if (fstat (fd, &st.stat) < 0)
 			    {
 			      file_removed_diag (p->name, !p->parent,
 						 stat_diag);
@@ -1393,8 +1355,9 @@ create_archive (void)
 			    }
 			  st.orig_file_name = xstrdup (p->name);
 			}
-		      while (buffer_size < plen + qlen)
-			buffer = x2realloc (buffer, &buffer_size);
+		      if (buffer_size < plen + qlen)
+			buffer = xpalloc (buffer, &buffer_size,
+					  plen + qlen - buffer_size, -1, 1);
 		      strcpy (buffer + plen, q + 1);
 		      dump_file (&st, q + 1, buffer);
 		    }
@@ -1407,7 +1370,7 @@ create_archive (void)
   else
     {
       const char *name;
-      while ((name = name_next (1)) != NULL)
+      while ((name = name_next (true)))
 	if (!excluded_name (name, NULL))
 	  dump_file (0, name, name);
     }
@@ -1425,7 +1388,7 @@ static size_t
 hash_link (void const *entry, size_t n_buckets)
 {
   struct link const *l = entry;
-  uintmax_t num = l->dev ^ l->ino;
+  uintmax_t num = l->st_dev ^ l->st_ino;
   return num % n_buckets;
 }
 
@@ -1435,15 +1398,14 @@ compare_links (void const *entry1, void const *entry2)
 {
   struct link const *link1 = entry1;
   struct link const *link2 = entry2;
-  return ((link1->dev ^ link2->dev) | (link1->ino ^ link2->ino)) == 0;
+  return PSAME_INODE (link1, link2);
 }
 
 static void
 unknown_file_error (char const *p)
 {
-  WARNOPT (WARN_FILE_IGNORED,
-	   (0, 0, _("%s: Unknown file type; file ignored"),
-	    quotearg_colon (p)));
+  warnopt (WARN_FILE_IGNORED, 0,
+	   _("%s: Unknown file type; file ignored"), quotearg_colon (p));
   if (!ignore_failed_read_option)
     set_exit_status (TAREXIT_FAILURE);
 }
@@ -1469,8 +1431,8 @@ dump_hard_link (struct tar_stat_info *st)
       off_t block_ordinal;
       union block *blk;
 
-      lp.ino = st->stat.st_ino;
-      lp.dev = st->stat.st_dev;
+      lp.st_dev = st->stat.st_dev;
+      lp.st_ino = st->stat.st_ino;
 
       if ((duplicate = hash_lookup (link_table, &lp)))
 	{
@@ -1517,11 +1479,15 @@ file_count_links (struct tar_stat_info *st)
 
       assign_string (&linkname, safer_name_suffix (st->orig_file_name, true,
 						   absolute_names_option));
-      transform_name (&linkname, XFORM_LINK);
+      if (!transform_name (&linkname, XFORM_LINK))
+	{
+	  free (linkname);
+	  return;
+	}
 
       lp = xmalloc (FLEXNSIZEOF (struct link, name, strlen (linkname) + 1));
-      lp->ino = st->stat.st_ino;
-      lp->dev = st->stat.st_dev;
+      lp->st_dev = st->stat.st_dev;
+      lp->st_ino = st->stat.st_ino;
       lp->nlink = st->stat.st_nlink;
       strcpy (lp->name, linkname);
       free (linkname);
@@ -1543,19 +1509,13 @@ file_count_links (struct tar_stat_info *st)
 void
 check_links (void)
 {
-  struct link *lp;
-
   if (!link_table)
     return;
 
-  for (lp = hash_get_first (link_table); lp;
+  for (struct link *lp = hash_get_first (link_table); lp;
        lp = hash_get_next (link_table, lp))
-    {
-      if (lp->nlink)
-	{
-	  WARN ((0, 0, _("Missing links to %s."), quote (lp->name)));
-	}
-    }
+    if (lp->nlink)
+      paxwarn (0, _("Missing links to %s."), quote (lp->name));
 }
 
 /* Assuming DIR is the working directory, open FILE, using FLAGS to
@@ -1600,9 +1560,8 @@ restore_parent_fd (struct tar_stat_info const *st)
 
       if (parentfd < 0)
 	parentfd = - errno;
-      else if (! (fstat (parentfd, &parentstat) == 0
-		  && parent->stat.st_ino == parentstat.st_ino
-		  && parent->stat.st_dev == parentstat.st_dev))
+      else if (fstat (parentfd, &parentstat) < 0
+	       || !psame_inode (&parent->stat, &parentstat))
 	{
 	  close (parentfd);
 	  parentfd = IMPOSTOR_ERRNO;
@@ -1614,12 +1573,11 @@ restore_parent_fd (struct tar_stat_info const *st)
 			       open_searchdir_flags);
 	  if (0 <= origfd)
 	    {
-	      if (fstat (parentfd, &parentstat) == 0
-		  && parent->stat.st_ino == parentstat.st_ino
-		  && parent->stat.st_dev == parentstat.st_dev)
-		parentfd = origfd;
-	      else
+	      if (fstat (parentfd, &parentstat) < 0
+		  || !psame_inode (&parent->stat, &parentstat))
 		close (origfd);
+	      else
+		parentfd = origfd;
 	    }
 	}
 
@@ -1629,12 +1587,15 @@ restore_parent_fd (struct tar_stat_info const *st)
 
 /* Dump a single file, recursing on directories.  ST is the file's
    status info, NAME its name relative to the parent directory, and P
-   its full name (which may be relative to the working directory).  */
+   its full name (which may be relative to the working directory).
+
+   Return the address of dynamically allocated storage that the caller
+   should free, or the null pointer if there is no such storage.  */
 
 /* FIXME: One should make sure that for *every* path leading to setting
    exit_status to failure, a clear diagnostic has been issued.  */
 
-static void
+static void *
 dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
 {
   union block *header;
@@ -1648,20 +1609,21 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
   void (*diag) (char const *) = 0;
 
   if (interactive_option && !confirm ("add", p))
-    return;
+    return NULL;
 
   assign_string (&st->orig_file_name, p);
   assign_string (&st->file_name,
                  safer_name_suffix (p, false, absolute_names_option));
 
-  transform_name (&st->file_name, XFORM_REGFILE);
+  if (!transform_name (&st->file_name, XFORM_REGFILE))
+    return NULL;
 
   if (parentfd < 0 && ! top_level)
     {
       errno = - parentfd;
       diag = open_diag;
     }
-  else if (fstatat (parentfd, name, &st->stat, fstatat_flags) != 0)
+  else if (fstatat (parentfd, name, &st->stat, fstatat_flags) < 0)
     diag = stat_diag;
   else if (file_dumpable_p (&st->stat))
     {
@@ -1671,14 +1633,14 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
       else
 	{
 	  st->fd = fd;
-	  if (fstat (fd, &st->stat) != 0)
+	  if (fstat (fd, &st->stat) < 0)
 	    diag = stat_diag;
 	}
     }
   if (diag)
     {
       file_removed_diag (p, top_level, diag);
-      return;
+      return NULL;
     }
 
   struct stat st1 = st->stat;
@@ -1687,16 +1649,13 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
   st->mtime = get_stat_mtime (&st->stat);
   st->ctime = get_stat_ctime (&st->stat);
 
+  void *allocated = NULL;
 #ifdef S_ISHIDDEN
   if (S_ISHIDDEN (st->stat.st_mode))
     {
-      char *new = (char *) alloca (strlen (p) + 2);
-      if (new)
-	{
-	  strcpy (new, p);
-	  strcat (new, "@");
-	  p = new;
-	}
+      allocated = xmalloc (strlen (p) + 2);
+      strcpy (stpcpy (allocated, p), "@");
+      p = allocated;
     }
 #endif
 
@@ -1708,29 +1667,28 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
 
   if (! (incremental_option && ! top_level)
       && !S_ISDIR (st->stat.st_mode)
-      && OLDER_TAR_STAT_TIME (*st, m)
-      && (!after_date_option || OLDER_TAR_STAT_TIME (*st, c)))
+      && timespec_cmp (st->mtime, newer_mtime_option) < 0
+      && (!after_date_option
+	  || timespec_cmp (st->ctime, newer_mtime_option) < 0))
     {
       if (!incremental_option && verbose_option)
-	WARNOPT (WARN_FILE_UNCHANGED,
-		 (0, 0, _("%s: file is unchanged; not dumped"),
-		  quotearg_colon (p)));
-      return;
+	warnopt (WARN_FILE_UNCHANGED, 0, _("%s: file is unchanged; not dumped"),
+		 quotearg_colon (p));
+      return allocated;
     }
 
   /* See if we are trying to dump the archive.  */
   if (sys_file_is_archive (st))
     {
-      WARNOPT (WARN_IGNORE_ARCHIVE,
-	       (0, 0, _("%s: archive cannot contain itself; not dumped"),
-		quotearg_colon (p)));
-      return;
+      warnopt (WARN_IGNORE_ARCHIVE, 0,
+	       _("%s: archive cannot contain itself; not dumped"), quotearg_colon (p));
+      return allocated;
     }
 
   is_dir = S_ISDIR (st->stat.st_mode) != 0;
 
   if (!is_dir && dump_hard_link (st))
-    return;
+    return allocated;
 
   if (is_dir || S_ISREG (st->stat.st_mode) || S_ISCTG (st->stat.st_mode))
     {
@@ -1751,7 +1709,7 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
 	    {
 	      exclusion_tag_warning (st->orig_file_name, tag_file_name,
 				     _("directory not dumped"));
-	      return;
+	      return allocated;
 	    }
 
 	  ok = dump_dir (st);
@@ -1843,15 +1801,15 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
 
 	  if (!ok)
 	    {
-	      WARNOPT (WARN_FILE_CHANGED,
-		       (0, 0, _("%s: file changed as we read it"),
-			quotearg_colon (p)));
+	      warnopt (WARN_FILE_CHANGED, 0,
+		       _("%s: file changed as we read it"),
+		       quotearg_colon (p));
 	      if (! ignore_failed_read_option)
 		set_exit_status (TAREXIT_DIFFERS);
 	    }
 	  else if (atime_preserve_option == replace_atime_preserve
 		   && timespec_cmp (st->atime, get_stat_atime (&st2)) != 0
-		   && set_file_atime (fd, parentfd, name, st->atime) != 0)
+		   && set_file_atime (fd, parentfd, name, st->atime) < 0)
 	    utime_error (p);
 	}
 
@@ -1859,9 +1817,8 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
       if (ok && remove_files_option)
 	queue_deferred_unlink (p, is_dir);
 
-      return;
+      return allocated;
     }
-#ifdef HAVE_READLINK
   else if (S_ISLNK (st->stat.st_mode))
     {
       st->link_name = areadlinkat_with_size (parentfd, name, st->stat.st_size);
@@ -1870,9 +1827,10 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
 	  if (errno == ENOMEM)
 	    xalloc_die ();
 	  file_removed_diag (p, top_level, readlink_diag);
-	  return;
+	  return allocated;
 	}
-      transform_name (&st->link_name, XFORM_SYMLINK);
+      if (!transform_name (&st->link_name, XFORM_SYMLINK))
+	return allocated;
       if (NAME_FIELD_SIZE - (archive_format == OLDGNU_FORMAT)
 	  < strlen (st->link_name))
 	write_long_link (st);
@@ -1884,7 +1842,7 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
       st->stat.st_size = 0;	/* force 0 size on symlink */
       header = start_header (st);
       if (!header)
-	return;
+	return allocated;
       tar_copy_str (header->header.linkname, st->link_name, NAME_FIELD_SIZE);
       header->header.typeflag = SYMTYPE;
       finish_header (st, header, block_ordinal);
@@ -1894,9 +1852,8 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
 	queue_deferred_unlink (p, false);
 
       file_count_links (st);
-      return;
+      return allocated;
     }
-#endif
   else if (S_ISCHR (st->stat.st_mode))
     {
       type = CHRTYPE;
@@ -1920,37 +1877,36 @@ dump_file0 (struct tar_stat_info *st, char const *name, char const *p)
     }
   else if (S_ISSOCK (st->stat.st_mode))
     {
-      WARNOPT (WARN_FILE_IGNORED,
-	       (0, 0, _("%s: socket ignored"), quotearg_colon (p)));
-      return;
+      warnopt (WARN_FILE_IGNORED, 0, _("%s: socket ignored"), quotearg_colon (p));
+      return allocated;
     }
   else if (S_ISDOOR (st->stat.st_mode))
     {
-      WARNOPT (WARN_FILE_IGNORED,
-	       (0, 0, _("%s: door ignored"), quotearg_colon (p)));
-      return;
+      warnopt (WARN_FILE_IGNORED, 0, _("%s: door ignored"), quotearg_colon (p));
+      return allocated;
     }
   else
     {
       unknown_file_error (p);
-      return;
+      return allocated;
     }
 
   if (archive_format == V7_FORMAT)
     {
       unknown_file_error (p);
-      return;
+      return allocated;
     }
 
   block_ordinal = current_block_ordinal ();
   st->stat.st_size = 0;	/* force 0 size */
   header = start_header (st);
   if (!header)
-    return;
+    return allocated;
   header->header.typeflag = type;
   finish_header (st, header, block_ordinal);
   if (remove_files_option)
     queue_deferred_unlink (p, false);
+  return allocated;
 }
 
 /* Dump a file, recursively.  PARENT describes the file's parent
@@ -1965,7 +1921,7 @@ dump_file (struct tar_stat_info *parent, char const *name,
   struct tar_stat_info st;
   tar_stat_init (&st);
   st.parent = parent;
-  dump_file0 (&st, name, fullname);
+  free (dump_file0 (&st, name, fullname));
   if (parent && listed_incremental_option)
     update_parent_directory (parent);
   tar_stat_destroy (&st);
