@@ -1,6 +1,6 @@
 /* Owner/group mapping for tar
 
-   Copyright 2015-2023 Free Software Foundation, Inc.
+   Copyright 2015-2025 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -45,28 +45,26 @@ map_compare (void const *entry1, void const *entry2)
   return map1->orig_id == map2->orig_id;
 }
 
-static int
+static bool
 parse_id (uintmax_t *retval,
 	  char const *arg, char const *what, uintmax_t maxval,
-	  char const *file, unsigned line)
+	  char const *file, intmax_t line)
 {
-  uintmax_t v;
   char *p;
-  
-  errno = 0;
-  v = strtoumax (arg, &p, 10);
-  if (*p || errno)
+  bool overflow;
+  *retval = stoint (arg, &p, &overflow, 0, maxval);
+
+  if ((p == arg) | *p)
     {
-      error (0, 0, _("%s:%u: invalid %s: %s"),  file, line, what, arg);
-      return -1;
+      error (0, 0, _("%s:%jd: invalid %s: %s"),  file, line, what, arg);
+      return false;
     }
-  if (v > maxval)
+  if (overflow)
     {
-      error (0, 0, _("%s:%u: %s out of range: %s"), file, line, what, arg);
-      return -1;
+      error (0, 0, _("%s:%jd: %s out of range: %s"), file, line, what, arg);
+      return false;
     }
-  *retval = v;
-  return 0;
+  return true;
 }
 
 static void
@@ -80,9 +78,9 @@ map_read (Hash_table **ptab, char const *file,
   ssize_t n;
   struct wordsplit ws;
   int wsopt;
-  unsigned line;
-  int err = 0;
-  
+  intmax_t line;
+  bool err = false;
+
   fp = fopen (file, "r");
   if (!fp)
     open_fatal (file);
@@ -97,26 +95,26 @@ map_read (Hash_table **ptab, char const *file,
       uintmax_t orig_id, new_id;
       char *name = NULL;
       char *colon;
-      
+
       ++line;
-      if (wordsplit (buf, &ws, wsopt))
-	FATAL_ERROR ((0, 0, _("%s:%u: cannot split line: %s"),
-		      file, line, wordsplit_strerror (&ws)));
+      if (wordsplit (buf, &ws, wsopt) != WRDSE_OK)
+	paxfatal (0, _("%s:%jd: cannot split line: %s"),
+		  file, line, wordsplit_strerror (&ws));
       wsopt |= WRDSF_REUSE;
       if (ws.ws_wordc == 0)
 	continue;
       if (ws.ws_wordc != 2)
 	{
-	  error (0, 0, _("%s:%u: malformed line"), file, line);
-	  err = 1;
+	  error (0, 0, _("%s:%jd: malformed line"), file, line);
+	  err = true;
 	  continue;
 	}
 
       if (ws.ws_wordv[0][0] == '+')
 	{
-	  if (parse_id (&orig_id, ws.ws_wordv[0]+1, what, maxval, file, line)) 
+	  if (!parse_id (&orig_id, ws.ws_wordv[0]+1, what, maxval, file, line))
 	    {
-	      err = 1;
+	      err = true;
 	      continue;
 	    }
 	}
@@ -125,9 +123,9 @@ map_read (Hash_table **ptab, char const *file,
 	  orig_id = name_to_id (ws.ws_wordv[0]);
 	  if (orig_id == UINTMAX_MAX)
 	    {
-	      error (0, 0, _("%s:%u: can't obtain %s of %s"),
+	      error (0, 0, _("%s:%jd: can't obtain %s of %s"),
 		     file, line, what, ws.ws_wordv[0]);
-	      err = 1;
+	      err = true;
 	      continue;
 	    }
 	}
@@ -138,17 +136,17 @@ map_read (Hash_table **ptab, char const *file,
 	  if (colon > ws.ws_wordv[1])
 	    name = ws.ws_wordv[1];
 	  *colon++ = 0;
-	  if (parse_id (&new_id, colon, what, maxval, file, line)) 
+	  if (!parse_id (&new_id, colon, what, maxval, file, line))
 	    {
-	      err = 1;
+	      err = true;
 	      continue;
 	    }
 	}
       else if (ws.ws_wordv[1][0] == '+')
 	{
-	  if (parse_id (&new_id, ws.ws_wordv[1], what, maxval, file, line)) 
+	  if (!parse_id (&new_id, ws.ws_wordv[1], what, maxval, file, line))
 	    {
-	      err = 1;
+	      err = true;
 	      continue;
 	    }
 	}
@@ -158,9 +156,9 @@ map_read (Hash_table **ptab, char const *file,
 	  new_id = name_to_id (ws.ws_wordv[1]);
 	  if (new_id == UINTMAX_MAX)
 	    {
-	      error (0, 0, _("%s:%u: can't obtain %s of %s"),
+	      error (0, 0, _("%s:%jd: can't obtain %s of %s"),
 		     file, line, what, ws.ws_wordv[1]);
-	      err = 1;
+	      err = true;
 	      continue;
 	    }
 	}
@@ -169,7 +167,7 @@ map_read (Hash_table **ptab, char const *file,
       ent->orig_id = orig_id;
       ent->new_id = new_id;
       ent->new_name = name ? xstrdup (name) : NULL;
-      
+
       if (!((*ptab
 	     || (*ptab = hash_initialize (0, 0, map_hash, map_compare, 0)))
 	    && hash_insert (*ptab, ent)))
@@ -179,7 +177,7 @@ map_read (Hash_table **ptab, char const *file,
     wordsplit_free (&ws);
   fclose (fp);
   if (err)
-    FATAL_ERROR ((0, 0, _("errors reading map file")));
+    paxfatal (0, _("errors reading map file"));
 }
 
 /* UID translation */
@@ -199,37 +197,28 @@ owner_map_read (char const *file)
   map_read (&owner_map, file, name_to_uid, "UID", TYPE_MAXIMUM (uid_t));
 }
 
-int
+void
 owner_map_translate (uid_t uid, uid_t *new_uid, char const **new_name)
 {
-  int rc = 1;
-  
   if (owner_map)
     {
       struct mapentry ent, *res;
-  
+
       ent.orig_id = uid;
       res = hash_lookup (owner_map, &ent);
       if (res)
 	{
 	  *new_uid = res->new_id;
 	  *new_name = res->new_name;
-	  return 0;
+	  return;
 	}
     }
 
-  if (owner_option != (uid_t) -1)
-    {
-      *new_uid = owner_option;
-      rc = 0;
-    }
+  uid_t minus_1 = -1;
+  if (owner_option != minus_1)
+    *new_uid = owner_option;
   if (owner_name_option)
-    {
-      *new_name = owner_name_option;
-      rc = 0;
-    }
-
-  return rc;
+    *new_name = owner_name_option;
 }
 
 /* GID translation */
@@ -249,35 +238,26 @@ group_map_read (char const *file)
   map_read (&group_map, file, name_to_gid, "GID", TYPE_MAXIMUM (gid_t));
 }
 
-int
+void
 group_map_translate (gid_t gid, gid_t *new_gid, char const **new_name)
 {
-  int rc = 1;
-  
   if (group_map)
     {
       struct mapentry ent, *res;
-  
+
       ent.orig_id = gid;
       res = hash_lookup (group_map, &ent);
       if (res)
 	{
 	  *new_gid = res->new_id;
 	  *new_name = res->new_name;
-	  return 0;
+	  return;
 	}
     }
 
-  if (group_option != (uid_t) -1)
-    {
-      *new_gid = group_option;
-      rc = 0;
-    }
+  gid_t minus_1 = -1;
+  if (group_option != minus_1)
+    *new_gid = group_option;
   if (group_name_option)
-    {
-      *new_name = group_name_option;
-      rc = 0;
-    }
-  
-  return rc;
+    *new_name = group_name_option;
 }
