@@ -1,6 +1,6 @@
 /* Functions for dealing with sparse files
 
-   Copyright 2003-2023 Free Software Foundation, Inc.
+   Copyright 2003-2026 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -16,6 +16,7 @@
    with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #include <system.h>
+#include <c-ctype.h>
 #include <inttostr.h>
 #include <quotearg.h>
 #include "common.h"
@@ -40,8 +41,8 @@ struct tar_sparse_optab
   bool (*decode_header) (struct tar_sparse_file *);
   bool (*scan_block) (struct tar_sparse_file *, enum sparse_scan_state,
 		      void *);
-  bool (*dump_region) (struct tar_sparse_file *, size_t);
-  bool (*extract_region) (struct tar_sparse_file *, size_t);
+  bool (*dump_region) (struct tar_sparse_file *, idx_t);
+  bool (*extract_region) (struct tar_sparse_file *, idx_t);
 };
 
 struct tar_sparse_file
@@ -73,12 +74,8 @@ dump_zeros (struct tar_sparse_file *file, off_t offset)
 
   while (file->offset < offset)
     {
-      size_t size = (BLOCKSIZE < offset - file->offset
-		     ? BLOCKSIZE
-		     : offset - file->offset);
-      ssize_t wrbytes;
-
-      wrbytes = write (file->fd, zero_buf, size);
+      idx_t size = min (BLOCKSIZE, offset - file->offset);
+      ssize_t wrbytes = write (file->fd, zero_buf, size);
       if (wrbytes <= 0)
 	{
 	  if (wrbytes == 0)
@@ -131,7 +128,7 @@ tar_sparse_scan (struct tar_sparse_file *file, enum sparse_scan_state state,
 }
 
 static bool
-tar_sparse_dump_region (struct tar_sparse_file *file, size_t i)
+tar_sparse_dump_region (struct tar_sparse_file *file, idx_t i)
 {
   if (file->optab->dump_region)
     return file->optab->dump_region (file, i);
@@ -139,7 +136,7 @@ tar_sparse_dump_region (struct tar_sparse_file *file, size_t i)
 }
 
 static bool
-tar_sparse_extract_region (struct tar_sparse_file *file, size_t i)
+tar_sparse_extract_region (struct tar_sparse_file *file, idx_t i)
 {
   if (file->optab->extract_region)
     return file->optab->extract_region (file, i);
@@ -188,9 +185,9 @@ lseek_or_error (struct tar_sparse_file *file, off_t offset)
    it's made *entirely* of zeros, returning a 0 the instant it finds
    something that is a nonzero, i.e., useful data.  */
 static bool
-zero_block_p (char const *buffer, size_t size)
+zero_block_p (char const *buffer, idx_t size)
 {
-  while (size--)
+  for (; size; size--)
     if (*buffer++)
       return false;
   return true;
@@ -200,10 +197,10 @@ static void
 sparse_add_map (struct tar_stat_info *st, struct sp_array const *sp)
 {
   struct sp_array *sparse_map = st->sparse_map;
-  size_t avail = st->sparse_map_avail;
+  idx_t avail = st->sparse_map_avail;
   if (avail == st->sparse_map_size)
     st->sparse_map = sparse_map =
-      x2nrealloc (sparse_map, &st->sparse_map_size, sizeof *sparse_map);
+      xpalloc (sparse_map, &st->sparse_map_size, 1, -1, sizeof *sparse_map);
   sparse_map[avail] = *sp;
   st->sparse_map_avail = avail + 1;
 }
@@ -215,7 +212,6 @@ sparse_scan_file_raw (struct tar_sparse_file *file)
   struct tar_stat_info *st = file->stat_info;
   int fd = file->fd;
   char buffer[BLOCKSIZE];
-  size_t count = 0;
   off_t offset = 0;
   struct sp_array sp = {0, 0};
 
@@ -224,9 +220,17 @@ sparse_scan_file_raw (struct tar_sparse_file *file)
   if (!tar_sparse_scan (file, scan_begin, NULL))
     return false;
 
-  while ((count = blocking_read (fd, buffer, sizeof buffer)) != 0
-         && count != SAFE_READ_ERROR)
+  while (true)
     {
+      idx_t count = blocking_read (fd, buffer, sizeof buffer);
+      if (count < sizeof buffer)
+	{
+	  if (errno)
+	    read_diag_details (st->orig_file_name, offset, sizeof buffer);
+	  if (count == 0)
+	    break;
+	}
+
       /* Analyze the block.  */
       if (zero_block_p (buffer, count))
         {
@@ -249,6 +253,8 @@ sparse_scan_file_raw (struct tar_sparse_file *file)
         }
 
       offset += count;
+      if (count < sizeof buffer)
+	break;
     }
 
   /* save one more sparse segment of length 0 to indicate that
@@ -257,7 +263,6 @@ sparse_scan_file_raw (struct tar_sparse_file *file)
     sp.offset = offset;
 
   sparse_add_map (st, &sp);
-  st->archive_file_size += count;
   return tar_sparse_scan (file, scan_end, NULL);
 }
 
@@ -300,7 +305,7 @@ sparse_scan_file_seek (struct tar_sparse_file *file)
       /* locate first chunk of data */
       data_offset = lseek (fd, offset, SEEK_DATA);
 
-      if (data_offset == (off_t)-1)
+      if (data_offset < 0)
         /* ENXIO == EOF; error otherwise */
         {
           if (errno == ENXIO)
@@ -354,8 +359,7 @@ sparse_scan_file (struct tar_sparse_file *file)
         return true;
 #else
       if (hole_detection == HOLE_DETECTION_SEEK)
-	WARN((0, 0,
-	      _("\"seek\" hole detection is not supported, using \"raw\".")));
+	paxwarn (0, _("\"seek\" hole detection is not supported, using \"raw\"."));
       /* fall back to "raw" for this and all other files */
       hole_detection = HOLE_DETECTION_RAW;
 #endif
@@ -401,9 +405,8 @@ sparse_select_optab (struct tar_sparse_file *file)
 }
 
 static bool
-sparse_dump_region (struct tar_sparse_file *file, size_t i)
+sparse_dump_region (struct tar_sparse_file *file, idx_t i)
 {
-  union block *blk;
   off_t bytes_left = file->stat_info->sparse_map[i].numbytes;
 
   if (!lseek_or_error (file, file->stat_info->sparse_map[i].offset))
@@ -411,68 +414,50 @@ sparse_dump_region (struct tar_sparse_file *file, size_t i)
 
   while (bytes_left > 0)
     {
-      size_t bufsize = (bytes_left > BLOCKSIZE) ? BLOCKSIZE : bytes_left;
-      size_t bytes_read;
-
-      blk = find_next_block ();
-      bytes_read = full_read (file->fd, blk->buffer, bufsize);
-      if (bytes_read == SAFE_READ_ERROR)
-	{
-          read_diag_details (file->stat_info->orig_file_name,
-	                     (file->stat_info->sparse_map[i].offset
-			      + file->stat_info->sparse_map[i].numbytes
-			      - bytes_left),
-			     bufsize);
-	  return false;
-	}
-      else if (bytes_read == 0)
-	{
-	  if (errno != 0)
-	    {
-	      read_diag_details (file->stat_info->orig_file_name,
-				 (file->stat_info->sparse_map[i].offset
-				  + file->stat_info->sparse_map[i].numbytes
-				  - bytes_left),
-				 bufsize);
-	      return false;
-	    }
-	  else
-	    {
-	      char buf[UINTMAX_STRSIZE_BOUND];
-	      struct stat st;
-	      size_t n;
-	      if (fstat (file->fd, &st) == 0)
-		n = file->stat_info->stat.st_size - st.st_size;
-	      else
-		n = file->stat_info->stat.st_size
-		  - (file->stat_info->sparse_map[i].offset
-		     + file->stat_info->sparse_map[i].numbytes
-		     - bytes_left);
-
-	      WARNOPT (WARN_FILE_SHRANK,
-		       (0, 0,
-			ngettext ("%s: File shrank by %s byte; padding with zeros",
-				  "%s: File shrank by %s bytes; padding with zeros",
-				  n),
-			quotearg_colon (file->stat_info->orig_file_name),
-			STRINGIFY_BIGINT (n, buf)));
-	      if (! ignore_failed_read_option)
-		set_exit_status (TAREXIT_DIFFERS);
-	      return false;
-	    }
-	}
-
-      memset (blk->buffer + bytes_read, 0, BLOCKSIZE - bytes_read);
+      union block *blk = find_next_block ();
+      idx_t avail = available_space_after (blk);
+      idx_t bufsize = min (bytes_left, avail);
+      idx_t bytes_read = full_read (file->fd, charptr (blk), bufsize);
+      if (bytes_read < BLOCKSIZE)
+	memset (blk->buffer + bytes_read, 0, BLOCKSIZE - bytes_read);
       bytes_left -= bytes_read;
       file->dumped_size += bytes_read;
-      set_next_block_after (blk);
+
+      if (bytes_read < bufsize)
+	{
+	  off_t current_offset = (file->stat_info->sparse_map[i].offset
+				  + file->stat_info->sparse_map[i].numbytes
+				  - bytes_left);
+	  if (errno != 0)
+	    read_diag_details (file->stat_info->orig_file_name,
+			       current_offset, bufsize - bytes_read);
+	  else
+	    {
+	      off_t cursize = current_offset;
+	      struct stat st;
+	      if (fstat (file->fd, &st) == 0 && st.st_size < cursize)
+		cursize = st.st_size;
+	      intmax_t n = file->stat_info->stat.st_size - cursize;
+	      warnopt (WARN_FILE_SHRANK, 0,
+		       ngettext ("%s: File shrank by %jd byte; padding with zeros",
+				 "%s: File shrank by %jd bytes; padding with zeros",
+				 n),
+		       quotearg_colon (file->stat_info->orig_file_name),
+		       n);
+	    }
+	  if (! ignore_failed_read_option)
+	    set_exit_status (TAREXIT_DIFFERS);
+	  return false;
+	}
+
+      set_next_block_after (charptr (blk) + bufsize - 1);
     }
 
   return true;
 }
 
 static bool
-sparse_extract_region (struct tar_sparse_file *file, size_t i)
+sparse_extract_region (struct tar_sparse_file *file, idx_t i)
 {
   off_t write_size;
 
@@ -489,17 +474,17 @@ sparse_extract_region (struct tar_sparse_file *file, size_t i)
     }
   else while (write_size > 0)
     {
-      size_t count;
-      size_t wrbytes = (write_size > BLOCKSIZE) ? BLOCKSIZE : write_size;
       union block *blk = find_next_block ();
       if (!blk)
 	{
-	  ERROR ((0, 0, _("Unexpected EOF in archive")));
+	  paxerror (0, _("Unexpected EOF in archive"));
 	  return false;
 	}
-      set_next_block_after (blk);
-      file->dumped_size += BLOCKSIZE;
-      count = blocking_write (file->fd, blk->buffer, wrbytes);
+      idx_t avail = available_space_after (blk);
+      idx_t wrbytes = min (write_size, avail);
+      set_next_block_after (charptr (blk) + wrbytes - 1);
+      file->dumped_size += avail;
+      idx_t count = blocking_write (file->fd, charptr (blk), wrbytes);
       write_size -= count;
       mv_size_left (file->stat_info->archive_file_size - file->dumped_size);
       file->offset += count;
@@ -536,12 +521,10 @@ sparse_dump_file (int fd, struct tar_stat_info *st)
 
       if (fd >= 0)
 	{
-	  size_t i;
-
 	  mv_begin_write (file.stat_info->file_name,
 		          file.stat_info->stat.st_size,
 		          file.stat_info->archive_file_size - file.dumped_size);
-	  for (i = 0; rc && i < file.stat_info->sparse_map_avail; i++)
+	  for (idx_t i = 0; rc && i < file.stat_info->sparse_map_avail; i++)
 	    rc = tar_sparse_dump_region (&file, i);
 	}
     }
@@ -577,7 +560,6 @@ sparse_extract_file (int fd, struct tar_stat_info *st, off_t *size)
 {
   bool rc = true;
   struct tar_sparse_file file;
-  size_t i;
 
   if (!tar_sparse_init (&file))
     {
@@ -591,7 +573,7 @@ sparse_extract_file (int fd, struct tar_stat_info *st, off_t *size)
   file.offset = 0;
 
   rc = tar_sparse_decode_header (&file);
-  for (i = 0; rc && i < file.stat_info->sparse_map_avail; i++)
+  for (idx_t i = 0; rc && i < file.stat_info->sparse_map_avail; i++)
     rc = tar_sparse_extract_region (&file, i);
   *size = file.stat_info->archive_file_size - file.dumped_size;
   return (tar_sparse_done (&file) && rc) ? dump_status_ok : dump_status_short;
@@ -623,24 +605,14 @@ check_sparse_region (struct tar_sparse_file *file, off_t beg, off_t end)
 
   while (beg < end)
     {
-      size_t bytes_read;
-      size_t rdsize = BLOCKSIZE < end - beg ? BLOCKSIZE : end - beg;
+      idx_t rdsize = min (end - beg, BLOCKSIZE);
       char diff_buffer[BLOCKSIZE];
 
-      bytes_read = full_read (file->fd, diff_buffer, rdsize);
-      if (bytes_read == SAFE_READ_ERROR)
+      idx_t bytes_read = full_read (file->fd, diff_buffer, rdsize);
+      if (bytes_read < rdsize)
 	{
-          read_diag_details (file->stat_info->orig_file_name,
-	                     beg,
-			     rdsize);
-	  return false;
-	}
-      else if (bytes_read == 0)
-	{
-	  if (errno != 0)
-	    read_diag_details (file->stat_info->orig_file_name,
-			       beg,
-			       rdsize);
+	  if (errno)
+	    read_diag_details (file->stat_info->orig_file_name, beg, rdsize);
 	  else
 	    report_difference (file->stat_info, _("Size differs"));
 	  return false;
@@ -662,7 +634,7 @@ check_sparse_region (struct tar_sparse_file *file, off_t beg, off_t end)
 }
 
 static bool
-check_data_region (struct tar_sparse_file *file, size_t i)
+check_data_region (struct tar_sparse_file *file, idx_t i)
 {
   off_t size_left;
 
@@ -673,47 +645,38 @@ check_data_region (struct tar_sparse_file *file, size_t i)
 
   while (size_left > 0)
     {
-      size_t bytes_read;
-      size_t rdsize = (size_left > BLOCKSIZE) ? BLOCKSIZE : size_left;
       char diff_buffer[BLOCKSIZE];
 
       union block *blk = find_next_block ();
       if (!blk)
 	{
-	  ERROR ((0, 0, _("Unexpected EOF in archive")));
+	  paxerror (0, _("Unexpected EOF in archive"));
 	  return false;
 	}
       set_next_block_after (blk);
       file->dumped_size += BLOCKSIZE;
-      bytes_read = full_read (file->fd, diff_buffer, rdsize);
-      if (bytes_read == SAFE_READ_ERROR)
+      idx_t rdsize = min (size_left, BLOCKSIZE);
+      idx_t bytes_read = full_read (file->fd, diff_buffer, rdsize);
+      size_left -= bytes_read;
+      mv_size_left (file->stat_info->archive_file_size - file->dumped_size);
+      if (!memeq (blk->buffer, diff_buffer, bytes_read))
 	{
-          read_diag_details (file->stat_info->orig_file_name,
-			     (file->stat_info->sparse_map[i].offset
-			      + file->stat_info->sparse_map[i].numbytes
-			      - size_left),
-			     rdsize);
+	  report_difference (file->stat_info, _("Contents differ"));
 	  return false;
 	}
-      else if (bytes_read == 0)
+
+      if (bytes_read < rdsize)
 	{
 	  if (errno != 0)
 	    read_diag_details (file->stat_info->orig_file_name,
 			       (file->stat_info->sparse_map[i].offset
 				+ file->stat_info->sparse_map[i].numbytes
 				- size_left),
-			       rdsize);
+			       rdsize - bytes_read);
 	  else
 	    report_difference (&current_stat_info, _("Size differs"));
 	  return false;
 	}
-      size_left -= bytes_read;
-      mv_size_left (file->stat_info->archive_file_size - file->dumped_size);
-      if (memcmp (blk->buffer, diff_buffer, bytes_read))
-	{
-	  report_difference (file->stat_info, _("Contents differ"));
-	  return false;
-	}
     }
   return true;
 }
@@ -723,7 +686,6 @@ sparse_diff_file (int fd, struct tar_stat_info *st)
 {
   bool rc = true;
   struct tar_sparse_file file;
-  size_t i;
   off_t offset = 0;
 
   if (!tar_sparse_init (&file))
@@ -735,7 +697,7 @@ sparse_diff_file (int fd, struct tar_stat_info *st)
 
   rc = tar_sparse_decode_header (&file);
   mv_begin_read (st);
-  for (i = 0; rc && i < file.stat_info->sparse_map_avail; i++)
+  for (idx_t i = 0; rc && i < file.stat_info->sparse_map_avail; i++)
     {
       rc = check_sparse_region (&file,
 				offset, file.stat_info->sparse_map[i].offset)
@@ -794,9 +756,10 @@ oldgnu_add_sparse (struct tar_sparse_file *file, struct sparse *s)
     return add_finish;
   sp.offset = OFF_FROM_HEADER (s->offset);
   sp.numbytes = OFF_FROM_HEADER (s->numbytes);
+  off_t size;
   if (sp.offset < 0 || sp.numbytes < 0
-      || INT_ADD_OVERFLOW (sp.offset, sp.numbytes)
-      || file->stat_info->stat.st_size < sp.offset + sp.numbytes
+      || ckd_add (&size, sp.offset, sp.numbytes)
+      || file->stat_info->stat.st_size < size
       || file->stat_info->archive_file_size < 0)
     return add_fail;
 
@@ -819,45 +782,43 @@ oldgnu_fixup_header (struct tar_sparse_file *file)
 static bool
 oldgnu_get_sparse_info (struct tar_sparse_file *file)
 {
-  size_t i;
   union block *h = current_header;
-  int ext_p;
   enum oldgnu_add_status rc;
 
   file->stat_info->sparse_map_avail = 0;
-  for (i = 0; i < SPARSES_IN_OLDGNU_HEADER; i++)
+  for (idx_t i = 0; i < SPARSES_IN_OLDGNU_HEADER; i++)
     {
       rc = oldgnu_add_sparse (file, &h->oldgnu_header.sp[i]);
       if (rc != add_ok)
 	break;
     }
 
-  for (ext_p = h->oldgnu_header.isextended;
+  for (char ext_p = h->oldgnu_header.isextended;
        rc == add_ok && ext_p; ext_p = h->sparse_header.isextended)
     {
       h = find_next_block ();
       if (!h)
 	{
-	  ERROR ((0, 0, _("Unexpected EOF in archive")));
+	  paxerror (0, _("Unexpected EOF in archive"));
 	  return false;
 	}
       set_next_block_after (h);
-      for (i = 0; i < SPARSES_IN_SPARSE_HEADER && rc == add_ok; i++)
+      for (idx_t i = 0; i < SPARSES_IN_SPARSE_HEADER && rc == add_ok; i++)
 	rc = oldgnu_add_sparse (file, &h->sparse_header.sp[i]);
     }
 
   if (rc == add_fail)
     {
-      ERROR ((0, 0, _("%s: invalid sparse archive member"),
-	      file->stat_info->orig_file_name));
+      paxerror (0, _("%s: invalid sparse archive member"),
+		file->stat_info->orig_file_name);
       return false;
     }
   return true;
 }
 
 static void
-oldgnu_store_sparse_info (struct tar_sparse_file *file, size_t *pindex,
-			  struct sparse *sp, size_t sparse_size)
+oldgnu_store_sparse_info (struct tar_sparse_file *file, idx_t *pindex,
+			  struct sparse *sp, idx_t sparse_size)
 {
   for (; *pindex < file->stat_info->sparse_map_avail
 	 && sparse_size > 0; sparse_size--, sp++, ++*pindex)
@@ -874,7 +835,6 @@ oldgnu_dump_header (struct tar_sparse_file *file)
 {
   off_t block_ordinal = current_block_ordinal ();
   union block *blk;
-  size_t i;
 
   blk = start_header (file->stat_info);
   blk->header.typeflag = GNUTYPE_SPARSE;
@@ -886,7 +846,7 @@ oldgnu_dump_header (struct tar_sparse_file *file)
   /* Store the effective (shrunken) file size */
   OFF_TO_CHARS (file->stat_info->archive_file_size, blk->header.size);
 
-  i = 0;
+  idx_t i = 0;
   oldgnu_store_sparse_info (file, &i,
 			    blk->oldgnu_header.sp,
 			    SPARSES_IN_OLDGNU_HEADER);
@@ -943,9 +903,8 @@ star_fixup_header (struct tar_sparse_file *file)
 static bool
 star_get_sparse_info (struct tar_sparse_file *file)
 {
-  size_t i;
   union block *h = current_header;
-  int ext_p;
+  char ext_p;
   enum oldgnu_add_status rc = add_ok;
 
   file->stat_info->sparse_map_avail = 0;
@@ -954,7 +913,7 @@ star_get_sparse_info (struct tar_sparse_file *file)
       && h->star_in_header.sp[0].offset[10] != '\0')
     {
       /* Old star format */
-      for (i = 0; i < SPARSES_IN_STAR_HEADER; i++)
+      for (idx_t i = 0; i < SPARSES_IN_STAR_HEADER; i++)
 	{
 	  rc = oldgnu_add_sparse (file, &h->star_in_header.sp[i]);
 	  if (rc != add_ok)
@@ -970,19 +929,19 @@ star_get_sparse_info (struct tar_sparse_file *file)
       h = find_next_block ();
       if (!h)
 	{
-	  ERROR ((0, 0, _("Unexpected EOF in archive")));
+	  paxerror (0, _("Unexpected EOF in archive"));
 	  return false;
 	}
       set_next_block_after (h);
-      for (i = 0; i < SPARSES_IN_STAR_EXT_HEADER && rc == add_ok; i++)
+      for (idx_t i = 0; i < SPARSES_IN_STAR_EXT_HEADER && rc == add_ok; i++)
 	rc = oldgnu_add_sparse (file, &h->star_ext_header.sp[i]);
       file->dumped_size += BLOCKSIZE;
     }
 
   if (rc == add_fail)
     {
-      ERROR ((0, 0, _("%s: invalid sparse archive member"),
-	      file->stat_info->orig_file_name));
+      paxerror (0, _("%s: invalid sparse archive member"),
+		file->stat_info->orig_file_name);
       return false;
     }
   return true;
@@ -1102,7 +1061,6 @@ pax_dump_header_0 (struct tar_sparse_file *file)
 {
   off_t block_ordinal = current_block_ordinal ();
   union block *blk;
-  size_t i;
   char nbuf[UINTMAX_STRSIZE_BOUND];
   struct sp_array *map = file->stat_info->sparse_map;
   char *save_file_name = NULL;
@@ -1114,7 +1072,7 @@ pax_dump_header_0 (struct tar_sparse_file *file)
   if (xheader_keyword_deleted_p ("GNU.sparse.map")
       || tar_sparse_minor == 0)
     {
-      for (i = 0; i < file->stat_info->sparse_map_avail; i++)
+      for (idx_t i = 0; i < file->stat_info->sparse_map_avail; i++)
 	{
 	  xheader_store ("GNU.sparse.offset", file->stat_info, &i);
 	  xheader_store ("GNU.sparse.numbytes", file->stat_info, &i);
@@ -1128,7 +1086,7 @@ pax_dump_header_0 (struct tar_sparse_file *file)
 					       "%d/GNUSparseFile.%p/%f", 0);
 
       xheader_string_begin (&file->stat_info->xhdr);
-      for (i = 0; i < file->stat_info->sparse_map_avail; i++)
+      for (idx_t i = 0; i < file->stat_info->sparse_map_avail; i++)
 	{
 	  if (i)
 	    xheader_string_add (&file->stat_info->xhdr, ",");
@@ -1156,48 +1114,65 @@ pax_dump_header_0 (struct tar_sparse_file *file)
   return true;
 }
 
+/* An output block BLOCK, and a pointer PTR into it.  */
+struct block_ptr
+{
+  union block *block;
+  char *ptr;
+};
+
+/* Append to BP the contents of the string SRC, followed by a newline.
+   If the string doesn't fit, put any overflow into the succeeding blocks.
+   Return the updated BP.  */
+static struct block_ptr
+dump_str_nl (struct block_ptr bp, char const *str)
+{
+  char *endp = bp.block->buffer + BLOCKSIZE;
+  char c;
+  do
+    {
+      c = *str++;
+      if (bp.ptr == endp)
+	{
+	  set_next_block_after (bp.block);
+	  bp.block = find_next_block ();
+	  bp.ptr = bp.block->buffer;
+	  endp = bp.block->buffer + BLOCKSIZE;
+	}
+      *bp.ptr++ = c ? c : '\n';
+    }
+  while (c);
+
+  return bp;
+}
+
+/* Return the floor of the log base 10 of N.  If N is 0, return 0.  */
+static int
+floorlog10 (uintmax_t n)
+{
+  for (int f = 0; ; f++)
+    if ((n /= 10) == 0)
+      return f;
+}
+
 static bool
 pax_dump_header_1 (struct tar_sparse_file *file)
 {
   off_t block_ordinal = current_block_ordinal ();
-  union block *blk;
-  char *p, *q;
-  size_t i;
   char nbuf[UINTMAX_STRSIZE_BOUND];
-  off_t size = 0;
   struct sp_array *map = file->stat_info->sparse_map;
   char *save_file_name = file->stat_info->file_name;
 
-#define COPY_STRING(b,dst,src) do                \
- {                                               \
-   char *endp = b->buffer + BLOCKSIZE;           \
-   char const *srcp = src;                       \
-   while (*srcp)                                 \
-     {                                           \
-       if (dst == endp)                          \
-	 {                                       \
-	   set_next_block_after (b);             \
-	   b = find_next_block ();               \
-           dst = b->buffer;                      \
-	   endp = b->buffer + BLOCKSIZE;         \
-	 }                                       \
-       *dst++ = *srcp++;                         \
-     }                                           \
-   } while (0)
-
   /* Compute stored file size */
-  p = umaxtostr (file->stat_info->sparse_map_avail, nbuf);
-  size += strlen (p) + 1;
-  for (i = 0; i < file->stat_info->sparse_map_avail; i++)
+  off_t size = floorlog10 (file->stat_info->sparse_map_avail) + 2;
+  for (idx_t i = 0; i < file->stat_info->sparse_map_avail; i++)
     {
-      p = umaxtostr (map[i].offset, nbuf);
-      size += strlen (p) + 1;
-      p = umaxtostr (map[i].numbytes, nbuf);
-      size += strlen (p) + 1;
+      size += floorlog10 (map[i].offset) + 2;
+      size += floorlog10 (map[i].numbytes) + 2;
     }
-  size = (size + BLOCKSIZE - 1) / BLOCKSIZE;
-  file->stat_info->archive_file_size += size * BLOCKSIZE;
-  file->dumped_size += size * BLOCKSIZE;
+  size = (size + BLOCKSIZE - 1) & ~(BLOCKSIZE - 1);
+  file->stat_info->archive_file_size += size;
+  file->dumped_size += size;
 
   /* Store sparse file identification */
   xheader_store ("GNU.sparse.major", file->stat_info, NULL);
@@ -1211,27 +1186,22 @@ pax_dump_header_1 (struct tar_sparse_file *file)
   if (strlen (file->stat_info->file_name) > NAME_FIELD_SIZE)
     file->stat_info->file_name[NAME_FIELD_SIZE] = 0;
 
-  blk = pax_start_header (file->stat_info);
-  finish_header (file->stat_info, blk, block_ordinal);
+  struct block_ptr bp;
+  bp.block = pax_start_header (file->stat_info);
+  finish_header (file->stat_info, bp.block, block_ordinal);
   free (file->stat_info->file_name);
   file->stat_info->file_name = save_file_name;
 
-  blk = find_next_block ();
-  q = blk->buffer;
-  p = umaxtostr (file->stat_info->sparse_map_avail, nbuf);
-  COPY_STRING (blk, q, p);
-  COPY_STRING (blk, q, "\n");
-  for (i = 0; i < file->stat_info->sparse_map_avail; i++)
+  bp.block = find_next_block ();
+  bp.ptr = bp.block->buffer;
+  bp = dump_str_nl (bp, umaxtostr (file->stat_info->sparse_map_avail, nbuf));
+  for (idx_t i = 0; i < file->stat_info->sparse_map_avail; i++)
     {
-      p = umaxtostr (map[i].offset, nbuf);
-      COPY_STRING (blk, q, p);
-      COPY_STRING (blk, q, "\n");
-      p = umaxtostr (map[i].numbytes, nbuf);
-      COPY_STRING (blk, q, p);
-      COPY_STRING (blk, q, "\n");
+      bp = dump_str_nl (bp, umaxtostr (map[i].offset, nbuf));
+      bp = dump_str_nl (bp, umaxtostr (map[i].numbytes, nbuf));
     }
-  memset (q, 0, BLOCKSIZE - (q - blk->buffer));
-  set_next_block_after (blk);
+  memset (bp.ptr, 0, BLOCKSIZE - (bp.ptr - bp.block->buffer));
+  set_next_block_after (bp.block);
   return true;
 }
 
@@ -1245,23 +1215,52 @@ pax_dump_header (struct tar_sparse_file *file)
            pax_dump_header_0 (file) : pax_dump_header_1 (file);
 }
 
-static bool
-decode_num (uintmax_t *num, char const *arg, uintmax_t maxval)
+/* A success flag OK, a computed integer N, and block + ptr BP.  */
+struct ok_n_block_ptr
 {
-  uintmax_t u;
-  char *arg_lim;
+  bool ok;
+  uintmax_t n;
+  struct block_ptr bp;
+};
 
-  if (!ISDIGIT (*arg))
-    return false;
+static struct ok_n_block_ptr
+decode_num (struct block_ptr bp, uintmax_t nmax, struct tar_sparse_file *file)
+{
+  char *endp = bp.block->buffer + BLOCKSIZE;
+  uintmax_t n = 0;
+  bool digit_seen = false, nondigit_seen = false, overflow = false;
+  while (true)
+    {
+      if (bp.ptr == endp)
+	{
+	  set_next_block_after (bp.block);
+	  bp.block = find_next_block ();
+	  if (!bp.block)
+	    paxfatal (0, _("Unexpected EOF in archive"));
+	  bp.ptr = bp.block->buffer;
+	  endp = bp.block->buffer + BLOCKSIZE;
+	}
+      char c = *bp.ptr++;
+      if (c == '\n')
+	break;
+      if (c_isdigit (c))
+	{
+	  digit_seen = true;
+	  overflow |= ckd_mul (&n, n, 10);
+	  overflow |= ckd_add (&n, n, c - '0');
+	}
+      else
+	nondigit_seen = true;
+    }
 
-  errno = 0;
-  u = strtoumax (arg, &arg_lim, 10);
-
-  if (! (u <= maxval && errno != ERANGE) || *arg_lim)
-    return false;
-
-  *num = u;
-  return true;
+  overflow |= nmax < n;
+  char const *msgid
+    = ((!digit_seen | nondigit_seen) ? N_("%s: malformed sparse archive member")
+       : overflow ? N_("%s: numeric overflow in sparse archive member")
+       : NULL);
+  if (msgid)
+    paxerror (0, gettext (msgid), file->stat_info->orig_file_name);
+  return (struct ok_n_block_ptr) { .ok = !msgid, .n = n, .bp = bp };
 }
 
 static bool
@@ -1269,82 +1268,38 @@ pax_decode_header (struct tar_sparse_file *file)
 {
   if (file->stat_info->sparse_major > 0)
     {
-      uintmax_t u;
-      char nbuf[UINTMAX_STRSIZE_BOUND];
-      union block *blk;
-      char *p;
-      size_t i;
-      off_t start;
-
-#define COPY_BUF(b,buf,src) do                                     \
- {                                                                 \
-   char *endp = b->buffer + BLOCKSIZE;                             \
-   char *dst = buf;                                                \
-   do                                                              \
-     {                                                             \
-       if (dst == buf + UINTMAX_STRSIZE_BOUND -1)                  \
-         {                                                         \
-           ERROR ((0, 0, _("%s: numeric overflow in sparse archive member"), \
-	          file->stat_info->orig_file_name));               \
-           return false;                                           \
-         }                                                         \
-       if (src == endp)                                            \
-	 {                                                         \
-	   set_next_block_after (b);                               \
-           b = find_next_block ();                                 \
-           if (!b)                                                 \
-             FATAL_ERROR ((0, 0, _("Unexpected EOF in archive"))); \
-           src = b->buffer;                                        \
-	   endp = b->buffer + BLOCKSIZE;                           \
-	 }                                                         \
-       *dst = *src++;                                              \
-     }                                                             \
-   while (*dst++ != '\n');                                         \
-   dst[-1] = 0;                                                    \
- } while (0)
-
-      start = current_block_ordinal ();
+      off_t start = current_block_ordinal ();
       set_next_block_after (current_header);
-      blk = find_next_block ();
-      if (!blk)
-        FATAL_ERROR ((0, 0, _("Unexpected EOF in archive")));
-      p = blk->buffer;
-      COPY_BUF (blk,nbuf,p);
-      if (!decode_num (&u, nbuf, TYPE_MAXIMUM (size_t)))
-	{
-	  ERROR ((0, 0, _("%s: malformed sparse archive member"),
-		  file->stat_info->orig_file_name));
-	  return false;
-	}
-      file->stat_info->sparse_map_size = u;
-      file->stat_info->sparse_map = xcalloc (file->stat_info->sparse_map_size,
-					     sizeof (*file->stat_info->sparse_map));
+      struct block_ptr bp;
+      bp.block = find_next_block ();
+      if (!bp.block)
+	paxfatal (0, _("Unexpected EOF in archive"));
+      bp.ptr = bp.block->buffer;
+      struct ok_n_block_ptr onbp = decode_num (bp, SIZE_MAX, file);
+      if (!onbp.ok)
+	return false;
+      bp = onbp.bp;
+      file->stat_info->sparse_map_size = onbp.n;
+      file->stat_info->sparse_map
+	= xicalloc (file->stat_info->sparse_map_size,
+		    sizeof *file->stat_info->sparse_map);
       file->stat_info->sparse_map_avail = 0;
-      for (i = 0; i < file->stat_info->sparse_map_size; i++)
+      for (idx_t i = 0; i < file->stat_info->sparse_map_size; i++)
 	{
 	  struct sp_array sp;
-
-	  COPY_BUF (blk,nbuf,p);
-	  if (!decode_num (&u, nbuf, TYPE_MAXIMUM (off_t)))
-	    {
-	      ERROR ((0, 0, _("%s: malformed sparse archive member"),
-		      file->stat_info->orig_file_name));
-	      return false;
-	    }
-	  sp.offset = u;
-	  COPY_BUF (blk,nbuf,p);
-	  if (!decode_num (&u, nbuf, TYPE_MAXIMUM (off_t))
-	      || INT_ADD_OVERFLOW (sp.offset, u)
-	      || file->stat_info->stat.st_size < sp.offset + u)
-	    {
-	      ERROR ((0, 0, _("%s: malformed sparse archive member"),
-		      file->stat_info->orig_file_name));
-	      return false;
-	    }
-	  sp.numbytes = u;
+	  onbp = decode_num (bp, file->stat_info->stat.st_size, file);
+	  if (!onbp.ok)
+	    return false;
+	  sp.offset = onbp.n;
+	  off_t numbytes_max = file->stat_info->stat.st_size - sp.offset;
+	  onbp = decode_num (onbp.bp, numbytes_max, file);
+	  if (!onbp.ok)
+	    return false;
+	  sp.numbytes = onbp.n;
+	  bp = onbp.bp;
 	  sparse_add_map (file->stat_info, &sp);
 	}
-      set_next_block_after (blk);
+      set_next_block_after (bp.block);
 
       file->dumped_size += BLOCKSIZE * (current_block_ordinal () - start);
     }
