@@ -1,6 +1,6 @@
 /* System-dependent calls for tar.
 
-   Copyright 2003-2021 Free Software Foundation, Inc.
+   Copyright 2003-2026 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -17,11 +17,21 @@
 
 #include <system.h>
 
+#if HAVE_SYS_MTIO_H
+# include <sys/ioctl.h>
+# include <sys/mtio.h>
+#endif
+
 #include "common.h"
 #include <priv-set.h>
 #include <rmt.h>
+#include <same-inode.h>
 #include <signal.h>
 #include <wordsplit.h>
+#include <poll.h>
+#include <parse-datetime.h>
+
+bool dev_null_output;
 
 static _Noreturn void
 xexec (const char *cmd)
@@ -37,7 +47,36 @@ xexec (const char *cmd)
   exec_fatal (cmd);
 }
 
-#if MSDOS
+/* True if the archive is seekable via ioctl and MTIOCTOP,
+   or if it is not known whether it is seekable.
+   False if it is known to be not seekable.  */
+static bool mtioseekable_archive;
+
+bool
+mtioseek (bool count_files, off_t count)
+{
+  if (mtioseekable_archive)
+    {
+#ifdef MTIOCTOP
+      struct mtop operation;
+      operation.mt_op = (count_files
+			 ? (count < 0 ? MTBSF : MTFSF)
+			 : (count < 0 ? MTBSR : MTFSR));
+      if (! (count < 0
+	     ? ckd_sub (&operation.mt_count, 0, count)
+	     : ckd_add (&operation.mt_count, count, 0))
+	  && (0 <= rmtioctl (archive, MTIOCTOP, &operation)
+	      || (errno == EIO
+		  && 0 <= rmtioctl (archive, MTIOCTOP, &operation))))
+	return true;
+#endif
+
+      mtioseekable_archive = false;
+    }
+  return false;
+}
+
+#if !HAVE_WAITPID /* MingW, MSVC 14.  */
 
 bool
 sys_get_archive_stat (void)
@@ -51,17 +90,12 @@ sys_file_is_archive (struct tar_stat_info *p)
   return false;
 }
 
-void
-sys_save_archive_dev_ino (void)
-{
-}
-
 void
 sys_detect_dev_null_output (void)
 {
   static char const dev_null[] = "nul";
 
-  dev_null_output = (strcmp (archive_name_array[0], dev_null) == 0
+  dev_null_output = (streq (archive_name_array[0], dev_null)
 		     || (! _isrmt (archive)));
 }
 
@@ -92,82 +126,86 @@ sys_compare_gid (struct stat *a, struct stat *b)
   return true;
 }
 
-void
-sys_compare_links (struct stat *link_data, struct stat *stat_data)
-{
-  return true;
-}
-
 int
 sys_truncate (int fd)
 {
   return write (fd, "", 0);
 }
 
-size_t
+idx_t
 sys_write_archive_buffer (void)
 {
-  return full_write (archive, record_start->buffer, record_size);
+  return full_write (archive, charptr (record_start), record_size);
 }
 
 /* Set ARCHIVE for writing, then compressing an archive.  */
 void
 sys_child_open_for_compress (void)
 {
-  FATAL_ERROR ((0, 0, _("Cannot use compressed or remote archives")));
+  paxfatal (0, _("Cannot use compressed or remote archives"));
 }
 
 /* Set ARCHIVE for uncompressing, then reading an archive.  */
 void
 sys_child_open_for_uncompress (void)
 {
-  FATAL_ERROR ((0, 0, _("Cannot use compressed or remote archives")));
+  paxfatal (0, _("Cannot use compressed or remote archives"));
 }
 
+bool
+sys_exec_setmtime_script (const char *script_name,
+			  int dirfd,
+			  const char *file_name,
+			  const char *fmt,
+			  struct timespec *ts)
+{
+  paxfatal (0, _("--set-mtime-command not implemented on this platform"));
+}
 #else
 
-extern union block *record_start; /* FIXME */
-
-static struct stat archive_stat; /* stat block for archive file */
-
 bool
 sys_get_archive_stat (void)
 {
-  return fstat (archive, &archive_stat) == 0;
+  bool remote = _isrmt (archive);
+  mtioseekable_archive = true;
+  if (!remote && 0 <= archive && fstat (archive, &archive_stat) == 0)
+    {
+      if (!S_ISCHR (archive_stat.st_mode))
+	mtioseekable_archive = false;
+      return true;
+    }
+  else
+    {
+      /* FIXME: This memset should not be needed.  It is present only
+	 because other parts of tar may incorrectly access
+	 archive_stat even if it's not the archive status.  */
+      memset (&archive_stat, 0, sizeof archive_stat);
+
+      return remote;
+    }
 }
 
 bool
 sys_file_is_archive (struct tar_stat_info *p)
 {
-  return (ar_dev && p->stat.st_dev == ar_dev && p->stat.st_ino == ar_ino);
+  return (!dev_null_output && !_isrmt (archive)
+	  && psame_inode (&p->stat, &archive_stat));
 }
 
-/* Save archive file inode and device numbers */
-void
-sys_save_archive_dev_ino (void)
-{
-  if (!_isrmt (archive) && S_ISREG (archive_stat.st_mode))
-    {
-      ar_dev = archive_stat.st_dev;
-      ar_ino = archive_stat.st_ino;
-    }
-  else
-    ar_dev = 0;
-}
+static char const dev_null[] = "/dev/null";
 
 /* Detect if outputting to "/dev/null".  */
 void
 sys_detect_dev_null_output (void)
 {
-  static char const dev_null[] = "/dev/null";
-  struct stat dev_null_stat;
+  static struct stat dev_null_stat;
 
-  dev_null_output = (strcmp (archive_name_array[0], dev_null) == 0
+  dev_null_output = (streq (archive_name_array[0], dev_null)
 		     || (! _isrmt (archive)
 			 && S_ISCHR (archive_stat.st_mode)
-			 && stat (dev_null, &dev_null_stat) == 0
-			 && archive_stat.st_dev == dev_null_stat.st_dev
-			 && archive_stat.st_ino == dev_null_stat.st_ino));
+			 && (dev_null_stat.st_ino != 0
+			     || stat (dev_null, &dev_null_stat) == 0)
+			 && psame_inode (&archive_stat, &dev_null_stat)));
 }
 
 void
@@ -177,7 +215,7 @@ sys_wait_for_child (pid_t child_pid, bool eof)
     {
       int wait_status;
 
-      while (waitpid (child_pid, &wait_status, 0) == -1)
+      while (waitpid (child_pid, &wait_status, 0) < 0)
 	if (errno != EINTR)
 	  {
 	    waitpid_error (use_compress_program_option);
@@ -188,11 +226,11 @@ sys_wait_for_child (pid_t child_pid, bool eof)
 	{
 	  int sig = WTERMSIG (wait_status);
 	  if (!(!eof && sig == SIGPIPE))
-	    FATAL_ERROR ((0, 0, _("Child died with signal %d"), sig));
+	    paxfatal (0, _("Child died with signal %d"), sig);
 	}
       else if (WEXITSTATUS (wait_status) != 0)
-	FATAL_ERROR ((0, 0, _("Child returned status %d"),
-		      WEXITSTATUS (wait_status)));
+	paxfatal (0, _("Child returned status %d"),
+		  WEXITSTATUS (wait_status));
     }
 }
 
@@ -213,7 +251,7 @@ sys_spawn_shell (void)
   else
     {
       int wait_status;
-      while (waitpid (child, &wait_status, 0) == -1)
+      while (waitpid (child, &wait_status, 0) < 0)
 	if (errno != EINTR)
 	  {
 	    waitpid_error (shell);
@@ -234,41 +272,47 @@ sys_compare_gid (struct stat *a, struct stat *b)
   return a->st_gid == b->st_gid;
 }
 
-bool
-sys_compare_links (struct stat *link_data, struct stat *stat_data)
-{
-  return stat_data->st_dev == link_data->st_dev
-         && stat_data->st_ino == link_data->st_ino;
-}
-
 int
 sys_truncate (int fd)
 {
-  off_t pos = lseek (fd, (off_t) 0, SEEK_CUR);
+  off_t pos = lseek (fd, 0, SEEK_CUR);
   return pos < 0 ? -1 : ftruncate (fd, pos);
 }
 
-/* Return nonzero if NAME is the name of a regular file, or if the file
+/* Return true if NAME is the name of a regular file, or if the file
    does not exist (so it would be created as a regular file).  */
-static int
+static bool
 is_regular_file (const char *name)
 {
   struct stat stbuf;
 
   if (stat (name, &stbuf) == 0)
-    return S_ISREG (stbuf.st_mode);
+    return !!S_ISREG (stbuf.st_mode);
   else
     return errno == ENOENT;
 }
 
-size_t
+idx_t
 sys_write_archive_buffer (void)
 {
-  return rmtwrite (archive, record_start->buffer, record_size);
+  return rmtwrite (archive, charptr (record_start), record_size);
 }
 
-#define	PREAD 0			/* read file descriptor from pipe() */
-#define	PWRITE 1		/* write file descriptor from pipe() */
+/* Read and write file descriptors from a pipe(pipefd) call.  */
+enum { PREAD, PWRITE };
+
+/* Work around GCC bug 109839.  */
+#if 13 <= __GNUC__
+# pragma GCC diagnostic ignored "-Wanalyzer-fd-leak"
+#endif
+
+/* Close file having descriptor FD, and abort if close unsuccessful.  */
+static void
+xclose (int fd)
+{
+  if (close (fd) < 0)
+    close_error (_("(pipe)"));
+}
 
 /* Duplicate file descriptor FROM into becoming INTO.
    INTO is closed first and has to be the next available slot.  */
@@ -277,37 +321,20 @@ xdup2 (int from, int into)
 {
   if (from != into)
     {
-      int status = close (into);
-
-      if (status != 0 && errno != EBADF)
-	{
-	  int e = errno;
-	  FATAL_ERROR ((0, e, _("Cannot close")));
-	}
-      status = dup (from);
-      if (status != into)
-	{
-	  if (status < 0)
-	    {
-	      int e = errno;
-	      FATAL_ERROR ((0, e, _("Cannot dup")));
-	    }
-	  abort ();
-	}
+      if (dup2 (from, into) < 0)
+	paxfatal (errno, _("Cannot dup2"));
       xclose (from);
     }
 }
 
-static void wait_for_grandchild (pid_t pid) __attribute__ ((__noreturn__));
-
 /* Propagate any failure of the grandchild back to the parent.  */
-static void
+static _Noreturn void
 wait_for_grandchild (pid_t pid)
 {
   int wait_status;
   int exit_code = 0;
 
-  while (waitpid (pid, &wait_status, 0) == -1)
+  while (waitpid (pid, &wait_status, 0) < 0)
     if (errno != EINTR)
       {
 	waitpid_error (use_compress_program_option);
@@ -364,7 +391,7 @@ sys_child_open_for_compress (void)
 
       /* We don't need a grandchild tar.  Open the archive and launch the
 	 compressor.  */
-      if (strcmp (archive_name_array[0], "-"))
+      if (!streq (archive_name_array[0], "-"))
 	{
 	  archive = creat (archive_name_array[0], MODE_RW);
 	  if (archive < 0)
@@ -406,7 +433,7 @@ sys_child_open_for_compress (void)
   xdup2 (child_pipe[PREAD], STDIN_FILENO);
   xclose (child_pipe[PWRITE]);
 
-  if (strcmp (archive_name_array[0], "-") == 0)
+  if (streq (archive_name_array[0], "-"))
     archive = STDOUT_FILENO;
   else
     {
@@ -419,20 +446,20 @@ sys_child_open_for_compress (void)
 
   while (1)
     {
-      size_t status = 0;
+      ptrdiff_t status = 0;
       char *cursor;
-      size_t length;
+      idx_t length;
 
       /* Assemble a record.  */
 
-      for (length = 0, cursor = record_start->buffer;
+      for (length = 0, cursor = charptr (record_start);
 	   length < record_size;
 	   length += status, cursor += status)
 	{
-	  size_t size = record_size - length;
+	  idx_t size = record_size - length;
 
 	  status = safe_read (STDIN_FILENO, cursor, size);
-	  if (status == SAFE_READ_ERROR)
+	  if (status < 0)
 	    read_fatal (use_compress_program_option);
 	  if (status == 0)
 	    break;
@@ -448,7 +475,7 @@ sys_child_open_for_compress (void)
 
 	  if (length > 0)
 	    {
-	      memset (record_start->buffer + length, 0, record_size - length);
+	      memset (charptr (record_start) + length, 0, record_size - length);
 	      status = sys_write_archive_buffer ();
 	      if (status != record_size)
 		archive_write_error (status);
@@ -481,24 +508,22 @@ run_decompress_program (void)
     {
       if (prog)
 	{
-	  WARNOPT (WARN_DECOMPRESS_PROGRAM,
-		   (0, errno, _("cannot run %s"), prog));
-	  WARNOPT (WARN_DECOMPRESS_PROGRAM,
-		   (0, 0, _("trying %s"), p));
+	  warnopt (WARN_DECOMPRESS_PROGRAM, errno, _("cannot run %s"), prog);
+	  warnopt (WARN_DECOMPRESS_PROGRAM, 0, _("trying %s"), p);
 	}
-      if (wordsplit (p, &ws, wsflags))
-	FATAL_ERROR ((0, 0, _("cannot split string '%s': %s"),
-		      p, wordsplit_strerror (&ws)));
+      if (wordsplit (p, &ws, wsflags) != WRDSE_OK)
+	paxfatal (0, _("cannot split string '%s': %s"),
+		  p, wordsplit_strerror (&ws));
       wsflags |= WRDSF_REUSE;
-      memmove(ws.ws_wordv, ws.ws_wordv + ws.ws_offs,
-	      sizeof(ws.ws_wordv[0])*ws.ws_wordc);
+      memmove (ws.ws_wordv, ws.ws_wordv + ws.ws_offs,
+	       ws.ws_wordc * sizeof *ws.ws_wordv);
       ws.ws_wordv[ws.ws_wordc] = (char *) "-d";
       prog = p;
       execvp (ws.ws_wordv[0], ws.ws_wordv);
       ws.ws_wordv[ws.ws_wordc] = NULL;
     }
   if (!prog)
-    FATAL_ERROR ((0, 0, _("unable to run decompression program")));
+    paxfatal (0, _("unable to run decompression program"));
   exec_fatal (prog);
 }
 
@@ -536,11 +561,11 @@ sys_child_open_for_uncompress (void)
      b) the file is to be accessed by rmt: compressor doesn't know how;
      c) the file is not a plain file.  */
 
-  if (strcmp (archive_name_array[0], "-") != 0
+  if (!streq (archive_name_array[0], "-")
       && !_remdev (archive_name_array[0])
       && is_regular_file (archive_name_array[0]))
     {
-      /* We don't need a grandchild tar.  Open the archive and lauch the
+      /* We don't need a grandchild tar.  Open the archive and launch the
 	 uncompressor.  */
 
       archive = open (archive_name_array[0], O_RDONLY | O_BINARY, MODE_RW);
@@ -576,7 +601,7 @@ sys_child_open_for_uncompress (void)
   xdup2 (child_pipe[PWRITE], STDOUT_FILENO);
   xclose (child_pipe[PREAD]);
 
-  if (strcmp (archive_name_array[0], "-") == 0)
+  if (streq (archive_name_array[0], "-"))
     archive = STDIN_FILENO;
   else
     archive = rmtopen (archive_name_array[0], O_RDONLY | O_BINARY,
@@ -586,34 +611,26 @@ sys_child_open_for_uncompress (void)
 
   /* Let's read the archive and pipe it into stdout.  */
 
-  while (1)
+  while (true)
     {
-      char *cursor;
-      size_t maximum;
-      size_t count;
-      size_t status;
-
       clear_read_error_count ();
 
-    error_loop:
-      status = rmtread (archive, record_start->buffer, record_size);
-      if (status == SAFE_READ_ERROR)
-	{
-	  archive_read_error ();
-	  goto error_loop;
-	}
-      if (status == 0)
+      ptrdiff_t n;
+      while ((n = rmtread (archive, charptr (record_start), record_size)) < 0)
+	archive_read_error ();
+      if (n == 0)
 	break;
-      cursor = record_start->buffer;
-      maximum = status;
-      while (maximum)
+
+      char *cursor = charptr (record_start);
+      do
 	{
-	  count = maximum < BLOCKSIZE ? maximum : BLOCKSIZE;
+	  idx_t count = min (n, BLOCKSIZE);
 	  if (full_write (STDOUT_FILENO, cursor, count) != count)
 	    write_error (use_compress_program_option);
 	  cursor += count;
-	  maximum -= count;
+	  n -= count;
 	}
+      while (n);
     }
 
   xclose (STDOUT_FILENO);
@@ -626,11 +643,8 @@ sys_child_open_for_uncompress (void)
 static void
 dec_to_env (char const *envar, uintmax_t num)
 {
-  char buf[UINTMAX_STRSIZE_BOUND];
-  char *numstr;
-
-  numstr = STRINGIFY_BIGINT (num, buf);
-  if (setenv (envar, numstr, 1) != 0)
+  char numstr[UINTMAX_STRSIZE_BOUND];
+  if (setenv (envar, umaxtostr (num, numstr), 1) < 0)
     xalloc_die ();
 }
 
@@ -638,17 +652,19 @@ static void
 time_to_env (char const *envar, struct timespec t)
 {
   char buf[TIMESPEC_STRSIZE_BOUND];
-  if (setenv (envar, code_timespec (t, buf), 1) != 0)
+  if (setenv (envar, code_timespec (t, buf), 1) < 0)
     xalloc_die ();
 }
 
 static void
-oct_to_env (char const *envar, unsigned long num)
+oct_to_env (char const *envar, mode_t m)
 {
-  char buf[1+1+(sizeof(unsigned long)*CHAR_BIT+2)/3];
-
-  snprintf (buf, sizeof buf, "0%lo", num);
-  if (setenv (envar, buf, 1) != 0)
+  char buf[sizeof "0" + (UINTMAX_WIDTH + 2) / 3];
+  uintmax_t um = m;
+  if (EXPR_SIGNED (m) && sizeof m < sizeof um)
+    um &= ~ (UINTMAX_MAX << TYPE_WIDTH (m));
+  sprintf (buf, "%#"PRIoMAX, um);
+  if (setenv (envar, buf, 1) < 0)
     xalloc_die ();
 }
 
@@ -657,7 +673,7 @@ str_to_env (char const *envar, char const *str)
 {
   if (str)
     {
-      if (setenv (envar, str, 1) != 0)
+      if (setenv (envar, str, 1) < 0)
 	xalloc_die ();
     }
   else
@@ -670,7 +686,7 @@ chr_to_env (char const *envar, char c)
   char buf[2];
   buf[0] = c;
   buf[1] = 0;
-  if (setenv (envar, buf, 1) != 0)
+  if (setenv (envar, buf, 1) < 0)
     xalloc_die ();
 }
 
@@ -725,7 +741,7 @@ static pid_t global_pid;
 static void (*pipe_handler) (int sig);
 
 int
-sys_exec_command (char *file_name, int typechar, struct tar_stat_info *st)
+sys_exec_command (char *file_name, char typechar, struct tar_stat_info *st)
 {
   int p[2];
 
@@ -758,7 +774,7 @@ sys_wait_command (void)
     return;
 
   signal (SIGPIPE, pipe_handler);
-  while (waitpid (global_pid, &status, 0) == -1)
+  while (waitpid (global_pid, &status, 0) < 0)
     if (errno != EINTR)
       {
         global_pid = -1;
@@ -769,53 +785,65 @@ sys_wait_command (void)
   if (WIFEXITED (status))
     {
       if (!ignore_command_error_option && WEXITSTATUS (status))
-	ERROR ((0, 0, _("%lu: Child returned status %d"),
-		(unsigned long) global_pid, WEXITSTATUS (status)));
+	paxerror (0, _("%jd: Child returned status %d"),
+		  intmax (global_pid), WEXITSTATUS (status));
     }
   else if (WIFSIGNALED (status))
     {
-      WARN ((0, 0, _("%lu: Child terminated on signal %d"),
-	     (unsigned long) global_pid, WTERMSIG (status)));
+      paxwarn (0, _("%jd: Child terminated on signal %d"),
+	       intmax (global_pid), WTERMSIG (status));
     }
   else
-    ERROR ((0, 0, _("%lu: Child terminated on unknown reason"),
-	    (unsigned long) global_pid));
+    paxerror (0, _("%jd: Child terminated on unknown reason"),
+	      intmax (global_pid));
 
   global_pid = -1;
 }
 
 int
-sys_exec_info_script (const char **archive_name, int volume_number)
+sys_exec_info_script (const char **archive_name, intmax_t volume_number)
 {
-  pid_t pid;
-  char uintbuf[UINTMAX_STRSIZE_BOUND];
   int p[2];
   static void (*saved_handler) (int sig);
 
   xpipe (p);
   saved_handler = signal (SIGPIPE, SIG_IGN);
 
-  pid = xfork ();
+  pid_t pid = xfork ();
 
   if (pid != 0)
     {
       /* Master */
 
-      int rc;
       int status;
       char *buf = NULL;
       size_t size = 0;
-      FILE *fp;
 
       xclose (p[PWRITE]);
-      fp = fdopen (p[PREAD], "r");
-      rc = getline (&buf, &size, fp);
-      fclose (fp);
+      FILE *fp = fdopen (p[PREAD], "r");
+      if (!fp)
+	{
+	  signal (SIGPIPE, saved_handler);
+	  call_arg_error ("fdopen", info_script_option);
+	  return -1;
+	}
+      ssize_t rc = getline (&buf, &size, fp);
+      if (rc < 0)
+	{
+	  signal (SIGPIPE, saved_handler);
+	  read_error (info_script_option);
+	  return -1;
+	}
+      *archive_name = buf;
+      buf[rc - 1] = '\0';
+      if (fclose (fp) < 0)
+	{
+	  signal (SIGPIPE, saved_handler);
+	  close_error (info_script_option);
+	  return -1;
+	}
 
-      if (rc > 0 && buf[rc-1] == '\n')
-	buf[--rc] = 0;
-
-      while (waitpid (pid, &status, 0) == -1)
+      while (waitpid (pid, &status, 0) < 0)
 	if (errno != EINTR)
 	  {
 	    signal (SIGPIPE, saved_handler);
@@ -824,31 +852,19 @@ sys_exec_info_script (const char **archive_name, int volume_number)
 	  }
 
       signal (SIGPIPE, saved_handler);
-
-      if (WIFEXITED (status))
-	{
-	  if (WEXITSTATUS (status) == 0 && rc > 0)
-	    *archive_name = buf;
-	  else
-	    free (buf);
-	  return WEXITSTATUS (status);
-	}
-
-      free (buf);
-      return -1;
+      return WIFEXITED (status) ? WEXITSTATUS (status) : -1;
     }
 
   /* Child */
-  setenv ("TAR_VERSION", PACKAGE_VERSION, 1);
-  setenv ("TAR_ARCHIVE", *archive_name, 1);
-  setenv ("TAR_VOLUME", STRINGIFY_BIGINT (volume_number, uintbuf), 1);
-  setenv ("TAR_BLOCKING_FACTOR",
-	  STRINGIFY_BIGINT (blocking_factor, uintbuf), 1);
+  str_to_env ("TAR_VERSION", PACKAGE_VERSION);
+  str_to_env ("TAR_ARCHIVE", *archive_name);
+  dec_to_env ("TAR_VOLUME", volume_number);
+  dec_to_env ("TAR_BLOCKING_FACTOR", blocking_factor);
   setenv ("TAR_SUBCOMMAND", subcommand_string (subcommand_option), 1);
   setenv ("TAR_FORMAT",
 	  archive_format_string (current_format == DEFAULT_FORMAT ?
 				 archive_format : current_format), 1);
-  setenv ("TAR_FD", STRINGIFY_BIGINT (p[PWRITE], uintbuf), 1);
+  dec_to_env ("TAR_FD", p[PWRITE]);
 
   xclose (p[PREAD]);
 
@@ -859,12 +875,9 @@ sys_exec_info_script (const char **archive_name, int volume_number)
 void
 sys_exec_checkpoint_script (const char *script_name,
 			    const char *archive_name,
-			    int checkpoint_number)
+			    intmax_t checkpoint_number)
 {
-  pid_t pid;
-  char uintbuf[UINTMAX_STRSIZE_BOUND];
-
-  pid = xfork ();
+  pid_t pid = xfork ();
 
   if (pid != 0)
     {
@@ -872,7 +885,7 @@ sys_exec_checkpoint_script (const char *script_name,
 
       int status;
 
-      while (waitpid (pid, &status, 0) == -1)
+      while (waitpid (pid, &status, 0) < 0)
 	if (errno != EINTR)
 	  {
 	    waitpid_error (script_name);
@@ -883,17 +896,175 @@ sys_exec_checkpoint_script (const char *script_name,
     }
 
   /* Child */
-  setenv ("TAR_VERSION", PACKAGE_VERSION, 1);
-  setenv ("TAR_ARCHIVE", archive_name, 1);
-  setenv ("TAR_CHECKPOINT", STRINGIFY_BIGINT (checkpoint_number, uintbuf), 1);
-  setenv ("TAR_BLOCKING_FACTOR",
-	  STRINGIFY_BIGINT (blocking_factor, uintbuf), 1);
-  setenv ("TAR_SUBCOMMAND", subcommand_string (subcommand_option), 1);
-  setenv ("TAR_FORMAT",
-	  archive_format_string (current_format == DEFAULT_FORMAT ?
-				 archive_format : current_format), 1);
+  str_to_env ("TAR_VERSION", PACKAGE_VERSION);
+  str_to_env ("TAR_ARCHIVE", archive_name);
+  dec_to_env ("TAR_CHECKPOINT", checkpoint_number);
+  dec_to_env ("TAR_BLOCKING_FACTOR", blocking_factor);
+  str_to_env ("TAR_SUBCOMMAND", subcommand_string (subcommand_option));
+  str_to_env ("TAR_FORMAT",
+	      archive_format_string (current_format == DEFAULT_FORMAT
+				     ? archive_format : current_format));
   priv_set_restore_linkdir ();
   xexec (script_name);
 }
 
+bool
+sys_exec_setmtime_script (const char *script_name,
+			  int dirfd,
+			  const char *file_name,
+			  const char *fmt,
+			  struct timespec *ts)
+{
+  pid_t pid;
+  int p[2];
+  bool stop = false;
+  struct pollfd pfd;
+
+  char *buffer = NULL;
+  idx_t buflen = 0;
+  idx_t bufsize = 0;
+  char *cp;
+  bool rc = true;
+
+  if (pipe (p) < 0)
+    paxfatal (errno, _("pipe failed"));
+
+  if ((pid = xfork ()) == 0)
+    {
+      char *command = xmalloc (strlen (script_name) + strlen (file_name) + 2);
+
+      strcpy (command, script_name);
+      strcat (command, " ");
+      strcat (command, file_name);
+
+      if (dirfd != AT_FDCWD && fchdir (dirfd) < 0)
+	paxfatal (errno, _("chdir failed"));
+
+      close (p[0]);
+      if (dup2 (p[1], STDOUT_FILENO) < 0)
+	paxfatal (errno, _("dup2 failed"));
+      if (p[1] != STDOUT_FILENO)
+	close (p[1]);
+
+      close (STDIN_FILENO);
+      if (open (dev_null, O_RDONLY) != STDIN_FILENO)
+	open_error (dev_null);
+
+      priv_set_restore_linkdir ();
+      /* FIXME: This mishandles shell metacharacters in the file name.
+	 Come to think of it, isn't every use of xexec suspect?  */
+      xexec (command);
+    }
+  close (p[1]);
+
+  pfd.fd = p[0];
+  pfd.events = POLLIN;
+
+  while (1)
+    {
+      int n = poll (&pfd, 1, -1);
+      if (n < 0)
+	{
+	  if (errno != EINTR)
+	    {
+	      paxerror (errno, _("poll failed"));
+	      stop = true;
+	      break;
+	    }
+	}
+      if (n == 0)
+	break;
+      if (pfd.revents & POLLIN)
+	{
+	  if (buflen == bufsize)
+	    buffer = xpalloc (buffer, &bufsize, 1, -1, 1);
+	  ssize_t nread = read (pfd.fd, buffer + buflen, bufsize - buflen);
+	  if (nread < 0)
+	    {
+	      paxerror (errno, _("error reading output of %s"), script_name);
+	      stop = true;
+	      break;
+	    }
+	  if (nread == 0)
+	    break;
+	  buflen += n;
+	}
+      else if (pfd.revents & POLLHUP)
+	break;
+    }
+  close (pfd.fd);
+
+  if (stop)
+    kill (SIGKILL, pid);
+
+  sys_wait_for_child (pid, false);
+
+  if (stop)
+    {
+      free (buffer);
+      return false;
+    }
+
+  if (buflen == 0)
+    {
+      paxerror (0, _("empty output from \"%s %s\""), script_name, file_name);
+      return false;
+    }
+
+  cp = memchr (buffer, '\n', buflen);
+  if (cp)
+    *cp = 0;
+  else
+    {
+      if (buflen == bufsize)
+	buffer = xirealloc (buffer, ++bufsize);
+      buffer[buflen] = 0;
+    }
+
+  if (fmt)
+    {
+      struct tm tm;
+      time_t t;
+      cp = strptime (buffer, fmt, &tm);
+      if (cp == NULL)
+	{
+	  paxerror (0, _("output from \"%s %s\" does not satisfy format string:"
+			 " %s"),
+		    script_name, file_name, buffer);
+	  rc = false;
+	}
+      else if (*cp != 0)
+	{
+	  paxwarn (0, _("unconsumed output from \"%s %s\": %s"),
+		   script_name, file_name, cp);
+	  rc = false;
+	}
+      else
+	{
+	  tm.tm_wday = -1;
+	  t = mktime (&tm);
+	  if (tm.tm_wday < 0)
+	    {
+	      paxerror (errno, _("mktime failed"));
+	      rc = false;
+	    }
+	  else
+	    {
+	      ts->tv_sec = t;
+	      ts->tv_nsec = 0;
+	    }
+	}
+    }
+  else if (! parse_datetime (ts, buffer, NULL))
+    {
+      paxerror (0, _("unparsable output from \"%s %s\": %s"),
+		script_name, file_name, buffer);
+      rc = false;
+    }
+
+  free (buffer);
+
+  return rc;
+}
+
 #endif /* not MSDOS */
