@@ -1,5 +1,5 @@
 /* This file is part of GNU tar.
-   Copyright 2006-2023 Free Software Foundation, Inc.
+   Copyright 2006-2026 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -16,6 +16,8 @@
 
 #include <system.h>
 #include <regex.h>
+#include <mcel.h>
+#include <quotearg.h>
 #include "common.h"
 
 enum transform_type
@@ -49,9 +51,9 @@ struct replace_segm
     struct
     {
       char *ptr;
-      size_t size;
+      idx_t size;
     } literal;                /* type == segm_literal */
-    size_t ref;               /* type == segm_backref */
+    idx_t ref;		      /* type == segm_backref */
     enum case_ctl_type ctl;   /* type == segm_case_ctl */
   } v;
 };
@@ -61,11 +63,11 @@ struct transform
   struct transform *next;
   enum transform_type transform_type;
   int flags;
-  unsigned match_number;
+  idx_t match_number;
   regex_t regex;
   /* Compiled replacement expression */
   struct replace_segm *repl_head, *repl_tail;
-  size_t segm_count; /* Number of elements in the above list */
+  idx_t segm_count; /* Number of elements in the above list */
 };
 
 
@@ -102,7 +104,7 @@ add_segment (struct transform *tf)
 static void
 add_literal_segment (struct transform *tf, const char *str, const char *end)
 {
-  size_t len = end - str;
+  idx_t len = end - str;
   if (len)
     {
       struct replace_segm *segm = add_segment (tf);
@@ -115,7 +117,7 @@ add_literal_segment (struct transform *tf, const char *str, const char *end)
 }
 
 static void
-add_char_segment (struct transform *tf, int chr)
+add_char_segment (struct transform *tf, char chr)
 {
   struct replace_segm *segm = add_segment (tf);
   segm->type = segm_literal;
@@ -126,15 +128,15 @@ add_char_segment (struct transform *tf, int chr)
 }
 
 static void
-add_backref_segment (struct transform *tf, size_t ref)
+add_backref_segment (struct transform *tf, idx_t ref)
 {
   struct replace_segm *segm = add_segment (tf);
   segm->type = segm_backref;
   segm->v.ref = ref;
 }
 
-static int
-parse_xform_flags (int *pflags, int c)
+static bool
+parse_xform_flags (int *pflags, char c)
 {
   switch (c)
     {
@@ -163,9 +165,9 @@ parse_xform_flags (int *pflags, int c)
       break;
 
     default:
-      return 1;
+      return false;
     }
-  return 0;
+  return true;
 }
 
 static void
@@ -179,8 +181,7 @@ add_case_ctl_segment (struct transform *tf, enum case_ctl_type ctl)
 static const char *
 parse_transform_expr (const char *expr)
 {
-  int delim;
-  int i, j, rc;
+  idx_t i, j;
   char *str, *beg, *cur;
   const char *p;
   int cflags = 0;
@@ -198,18 +199,17 @@ parse_transform_expr (const char *expr)
 		  expr++;
 		  break;
 		}
-	      if (parse_xform_flags (&transform_flags, *expr))
-		USAGE_ERROR ((0, 0, _("Unknown transform flag: %c"),
-			      *expr));
+	      if (!parse_xform_flags (&transform_flags, *expr))
+		paxusage (_("Unknown transform flag: %c"), *expr);
 	    }
 	  return expr;
 	}
-      USAGE_ERROR ((0, 0, _("Invalid transform expression")));
+      paxusage (_("Invalid transform expression"));
     }
 
-  delim = expr[1];
+  char delim = expr[1];
   if (!delim)
-    USAGE_ERROR ((0, 0, _("Invalid transform expression")));
+    paxusage (_("Invalid transform expression"));
 
   /* Scan regular expression */
   for (i = 2; expr[i] && expr[i] != delim; i++)
@@ -217,7 +217,7 @@ parse_transform_expr (const char *expr)
       i++;
 
   if (expr[i] != delim)
-    USAGE_ERROR ((0, 0, _("Invalid transform expression")));
+    paxusage (_("Invalid transform expression"));
 
   /* Scan replacement expression */
   for (j = i + 1; expr[j] && expr[j] != delim; j++)
@@ -225,7 +225,7 @@ parse_transform_expr (const char *expr)
       j++;
 
   if (expr[j] != delim)
-    USAGE_ERROR ((0, 0, _("Invalid transform expression")));
+    paxusage (_("Invalid transform expression"));
 
   /* Check flags */
   tf->transform_type = transform_first;
@@ -247,14 +247,16 @@ parse_transform_expr (const char *expr)
 
       case '0': case '1': case '2': case '3': case '4':
       case '5': case '6': case '7': case '8': case '9':
-	tf->match_number = strtoul (p, (char**) &p, 0);
-	p--;
+	{
+	  char *endp;
+	  tf->match_number = stoint (p, &endp, NULL, 0, IDX_MAX);
+	  p = endp - 1;
+	}
 	break;
 
       default:
-	if (parse_xform_flags (&tf->flags, *p))
-	  USAGE_ERROR ((0, 0, _("Unknown flag in transform expression: %c"),
-			*p));
+	if (!parse_xform_flags (&tf->flags, *p))
+	  paxusage (_("Unknown flag in transform expression: %c"), *p);
       }
 
   if (*p == ';')
@@ -265,13 +267,13 @@ parse_transform_expr (const char *expr)
   memcpy (str, expr + 2, i - 2);
   str[i - 2] = 0;
 
-  rc = regcomp (&tf->regex, str, cflags);
+  int rc = regcomp (&tf->regex, str, cflags);
 
   if (rc)
     {
       char errbuf[512];
       regerror (rc, &tf->regex, errbuf, sizeof (errbuf));
-      USAGE_ERROR ((0, 0, _("Invalid transform expression: %s"), errbuf));
+      paxusage (_("Invalid transform expression: %s"), errbuf);
     }
 
   if (str[0] == '^' || (i > 2 && str[i - 3] == '$'))
@@ -289,17 +291,18 @@ parse_transform_expr (const char *expr)
     {
       if (*cur == '\\')
 	{
-	  size_t n;
-
 	  add_literal_segment (tf, beg, cur);
 	  switch (*++cur)
 	    {
 	    case '0': case '1': case '2': case '3': case '4':
 	    case '5': case '6': case '7': case '8': case '9':
-	      n = strtoul (cur, &cur, 10);
-	      if (n > tf->regex.re_nsub)
-		USAGE_ERROR ((0, 0, _("Invalid transform replacement: back reference out of range")));
-	      add_backref_segment (tf, n);
+	      {
+		idx_t n = stoint (cur, &cur, NULL, 0, IDX_MAX);
+		if (tf->regex.re_nsub < n)
+		  paxusage (_("Invalid transform replacement:"
+			      " back reference out of range"));
+		add_backref_segment (tf, n);
+	      }
 	      break;
 
 	    case '\\':
@@ -416,74 +419,56 @@ set_transform_expr (const char *expr)
     expr = parse_transform_expr (expr);
 }
 
-/* Run case conversion specified by CASE_CTL on array PTR of SIZE
-   characters. Returns pointer to statically allocated storage. */
-static char *
-run_case_conv (enum case_ctl_type case_ctl, char *ptr, size_t size)
-{
-  static char *case_ctl_buffer;
-  static size_t case_ctl_bufsize;
-  char *p;
-
-  if (case_ctl_bufsize < size)
-    {
-      case_ctl_bufsize = size;
-      case_ctl_buffer = xrealloc (case_ctl_buffer, case_ctl_bufsize);
-    }
-  memcpy (case_ctl_buffer, ptr, size);
-  switch (case_ctl)
-    {
-    case ctl_upcase_next:
-      case_ctl_buffer[0] = toupper ((unsigned char) case_ctl_buffer[0]);
-      break;
-
-    case ctl_locase_next:
-      case_ctl_buffer[0] = tolower ((unsigned char) case_ctl_buffer[0]);
-      break;
-
-    case ctl_upcase:
-      for (p = case_ctl_buffer; p < case_ctl_buffer + size; p++)
-	*p = toupper ((unsigned char) *p);
-      break;
-
-    case ctl_locase:
-      for (p = case_ctl_buffer; p < case_ctl_buffer + size; p++)
-	*p = tolower ((unsigned char) *p);
-      break;
-
-    case ctl_stop:
-      break;
-    }
-  return case_ctl_buffer;
-}
-
 
 static struct obstack stk;
 static bool stk_init;
 
+/* Run case conversion specified by CASE_CTL on array PTR of SIZE
+   characters.  Append the result to STK.  */
+static void
+run_case_conv (enum case_ctl_type case_ctl, char *ptr, idx_t size)
+{
+  char const *p = ptr, *plim = ptr + size;
+  mbstate_t mbs; mbszero (&mbs);
+  while (p < plim)
+    {
+      mcel_t g = mcel_scan (p, plim);
+      char32_t ch;
+      switch (case_ctl)
+	{
+	case ctl_upcase: case ctl_upcase_next: ch = c32toupper (g.ch); break;
+	case ctl_locase: case ctl_locase_next: ch = c32tolower (g.ch); break;
+	default: ch = g.ch; break;
+	}
+      if (ch == g.ch)
+	obstack_grow (&stk, p, g.len);
+      else
+	{
+	  obstack_make_room (&stk, MB_LEN_MAX);
+	  mbstate_t ombs; mbszero (&ombs);
+	  idx_t outbytes = c32rtomb (obstack_next_free (&stk), ch, &ombs);
+	  obstack_blank_fast (&stk, outbytes);
+	}
+      p += g.len;
+      if (case_ctl != ctl_upcase && case_ctl != ctl_locase)
+	break;
+    }
+
+  obstack_grow (&stk, p, plim - p);
+}
+
 static void
 _single_transform_name_to_obstack (struct transform *tf, char *input)
 {
-  regmatch_t *rmp;
   int rc;
-  size_t nmatches = 0;
+  idx_t nmatches = 0;
   enum case_ctl_type case_ctl = ctl_stop,  /* Current case conversion op */
                      save_ctl = ctl_stop;  /* Saved case_ctl for \u and \l */
-
-  /* Reset case conversion after a single-char operation */
-#define CASE_CTL_RESET()  if (case_ctl == ctl_upcase_next     \
-			      || case_ctl == ctl_locase_next) \
-                            {                                 \
-                              case_ctl = save_ctl;            \
-                              save_ctl = ctl_stop;            \
-			    }
-
-  rmp = xmalloc ((tf->regex.re_nsub + 1) * sizeof (*rmp));
+  regmatch_t *rmp = xinmalloc (tf->regex.re_nsub + 1, sizeof *rmp);
 
   while (*input)
     {
-      size_t disp;
-      char *ptr;
+      idx_t disp;
 
       rc = regexec (&tf->regex, input, tf->regex.re_nsub + 1, rmp, 0);
 
@@ -509,32 +494,28 @@ _single_transform_name_to_obstack (struct transform *tf, char *input)
 	      switch (segm->type)
 		{
 		case segm_literal:    /* Literal segment */
-		  if (case_ctl == ctl_stop)
-		    ptr = segm->v.literal.ptr;
-		  else
+		  run_case_conv (case_ctl,
+				 segm->v.literal.ptr,
+				 segm->v.literal.size);
+		case_ctl_reset:
+		  /* Reset case conversion after a single-char operation.  */
+		  if (case_ctl == ctl_upcase_next
+		      || case_ctl == ctl_locase_next)
 		    {
-		      ptr = run_case_conv (case_ctl,
-					   segm->v.literal.ptr,
-					   segm->v.literal.size);
-		      CASE_CTL_RESET();
+		      case_ctl = save_ctl;
+		      save_ctl = ctl_stop;
 		    }
-		  obstack_grow (&stk, ptr, segm->v.literal.size);
 		  break;
 
 		case segm_backref:    /* Back-reference segment */
-		  if (rmp[segm->v.ref].rm_so != -1
-		      && rmp[segm->v.ref].rm_eo != -1)
+		  if (0 <= rmp[segm->v.ref].rm_so
+		      && 0 <= rmp[segm->v.ref].rm_eo)
 		    {
-		      size_t size = rmp[segm->v.ref].rm_eo
-			              - rmp[segm->v.ref].rm_so;
-		      ptr = input + rmp[segm->v.ref].rm_so;
-		      if (case_ctl != ctl_stop)
-			{
-			  ptr = run_case_conv (case_ctl, ptr, size);
-			  CASE_CTL_RESET();
-			}
-
-		      obstack_grow (&stk, ptr, size);
+		      idx_t size = (rmp[segm->v.ref].rm_eo
+				    - rmp[segm->v.ref].rm_so);
+		      run_case_conv (case_ctl,
+				     input + rmp[segm->v.ref].rm_so, size);
+		      goto case_ctl_reset;
 		    }
 		  break;
 
@@ -581,11 +562,11 @@ _single_transform_name_to_obstack (struct transform *tf, char *input)
   free (rmp);
 }
 
-static bool
+static void
 _transform_name_to_obstack (int flags, char *input, char **output)
 {
   struct transform *tf;
-  bool alloced = false;
+  bool ok = false;
 
   if (!stk_init)
     {
@@ -599,38 +580,53 @@ _transform_name_to_obstack (int flags, char *input, char **output)
 	{
 	  _single_transform_name_to_obstack (tf, input);
 	  input = obstack_finish (&stk);
-	  alloced = true;
+	  ok = true;
 	}
     }
+  if (!ok)
+    {
+      obstack_grow0 (&stk, input, strlen (input));
+      input = obstack_finish (&stk);
+    }
   *output = input;
-  return alloced;
 }
 
+/* Transform name *PINPUT of a file or archive member of type TYPE
+   (a single XFORM_* bit).  If FUN is not NULL, call this function
+   to further transform the result.  Arguments to FUN are the transformed
+   name and type, it's return value is the new transformed name.
+
+   If transformation results in a non-empty string, store the result in
+   *PINPUT and return true.  Otherwise, if it results in an empty string,
+   issue a warning, return false and don't modify PINPUT.
+ */
 bool
-transform_name_fp (char **pinput, int flags,
-		   char *(*fun)(char *, void *), void *dat)
+transform_name_fp (char **pinput, int type,
+		   char const *(*fun) (char const *, int))
 {
-    char *str;
-    bool ret = _transform_name_to_obstack (flags, *pinput, &str);
-    if (ret)
-      {
-	assign_string (pinput, fun ? fun (str, dat) : str);
-	obstack_free (&stk, str);
-      }
-    else if (fun)
-      {
-	*pinput = NULL;
-	assign_string (pinput, fun (str, dat));
-	free (str);
-	ret = true;
-      }
-    return ret;
+  char *str;
+  char const *result;
+
+  _transform_name_to_obstack (type, *pinput, &str);
+  result = (str[0] != 0 && fun) ? fun (str, type) : str;
+
+  if (result[0] == 0)
+    {
+      warnopt (WARN_EMPTY_TRANSFORM, 0,
+	       _("%s: transforms to empty name"), quotearg_colon (*pinput));
+      obstack_free (&stk, str);
+      return false;
+    }
+
+  assign_string (pinput, result);
+  obstack_free (&stk, str);
+  return true;
 }
 
 bool
 transform_name (char **pinput, int type)
 {
-  return transform_name_fp (pinput, type, NULL, NULL);
+  return transform_name_fp (pinput, type, NULL);
 }
 
 bool
