@@ -1,6 +1,6 @@
 /* Unlink files.
 
-   Copyright 2009-2023 Free Software Foundation, Inc.
+   Copyright 2009-2026 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -24,7 +24,7 @@
 struct deferred_unlink
   {
     struct deferred_unlink *next;   /* Next unlink in the queue */
-    int dir_idx;                    /* Directory index in wd */
+    idx_t dir_idx;		    /* Directory index in wd */
     char *file_name;                /* Name of the file to unlink, relative
 				       to dir_idx */
     bool is_dir;                    /* True if file_name is a directory */
@@ -32,35 +32,25 @@ struct deferred_unlink
 				       entry got added to the queue */
   };
 
-#define IS_CWD(p) \
-  ((p)->is_dir \
-   && ((p)->file_name[0] == 0 || strcmp ((p)->file_name, ".") == 0))
+static bool
+is_cwd (struct deferred_unlink const *p)
+{
+  return p->is_dir && !p->file_name[p->file_name[0] == '.'];
+}
 
 /* The unlink queue */
 static struct deferred_unlink *dunlink_head, *dunlink_tail;
 
-/* Number of entries in the queue */
-static size_t dunlink_count;
-
 /* List of entries available for allocation */
 static struct deferred_unlink *dunlink_avail;
 
-/* Delay (number of records written) between adding entry to the
-   list and its actual removal. */
-static size_t deferred_unlink_delay = 0;
-
 static struct deferred_unlink *
 dunlink_alloc (void)
 {
-  struct deferred_unlink *p;
-  if (dunlink_avail)
-    {
-      p = dunlink_avail;
-      dunlink_avail = p->next;
-      p->next  = NULL;
-    }
-  else
-    p = xmalloc (sizeof (*p));
+  struct deferred_unlink *p = dunlink_avail;
+  if (!p)
+    return xmalloc (sizeof *p);
+  dunlink_avail = p->next;
   return p;
 }
 
@@ -72,14 +62,13 @@ dunlink_insert (struct deferred_unlink *anchor, struct deferred_unlink *p)
       p->next = anchor->next;
       anchor->next = p;
     }
-  else 
+  else
     {
       p->next = dunlink_head;
       dunlink_head = p;
     }
   if (!p->next)
     dunlink_tail = p;
-  dunlink_count++;
 }
 
 static void
@@ -94,21 +83,21 @@ static void
 flush_deferred_unlinks (bool force)
 {
   struct deferred_unlink *p, *prev = NULL;
-  int saved_chdir = chdir_current;
-  
+  idx_t saved_chdir = chdir_current;
+
   for (p = dunlink_head; p; )
     {
       struct deferred_unlink *next = p->next;
 
       if (force
-	  || records_written > p->records_written + deferred_unlink_delay)
+	  || p->records_written < records_written)
 	{
 	  chdir_do (p->dir_idx);
 	  if (p->is_dir)
 	    {
 	      const char *fname;
 
-	      if (p->dir_idx && IS_CWD (p))
+	      if (p->dir_idx && is_cwd (p))
 		{
 		  prev = p;
 		  p = next;
@@ -117,7 +106,10 @@ flush_deferred_unlinks (bool force)
 	      else
 		fname = p->file_name;
 
-	      if (unlinkat (chdir_fd, fname, AT_REMOVEDIR) != 0)
+	      struct fdbase f = fdbase (fname);
+	      if (f.fd != BADFD && unlinkat (f.fd, f.base, AT_REMOVEDIR) == 0)
+		fdbase_clear ();
+	      else
 		{
 		  switch (errno)
 		    {
@@ -143,11 +135,13 @@ flush_deferred_unlinks (bool force)
 	    }
 	  else
 	    {
-	      if (unlinkat (chdir_fd, p->file_name, 0) != 0 && errno != ENOENT)
+	      struct fdbase f = fdbase (p->file_name);
+	      if (f.fd != BADFD && unlinkat (f.fd, f.base, 0) == 0)
+		fdbase_clear ();
+	      else if (errno != ENOENT)
 		unlink_error (p->file_name);
 	    }
 	  dunlink_reclaim (p);
-	  dunlink_count--;
 	  p = next;
 	  if (prev)
 	    prev->next = p;
@@ -170,7 +164,7 @@ flush_deferred_unlinks (bool force)
 	  const char *fname;
 
 	  chdir_do (p->dir_idx);
-	  if (p->dir_idx && IS_CWD (p))
+	  if (p->dir_idx && is_cwd (p))
 	    {
 	      fname = tar_dirname ();
 	      chdir_do (p->dir_idx - 1);
@@ -178,18 +172,18 @@ flush_deferred_unlinks (bool force)
 	  else
 	    fname = p->file_name;
 
-	  if (unlinkat (chdir_fd, fname, AT_REMOVEDIR) != 0)
-	    {
-	      if (errno != ENOENT)
-		rmdir_error (fname);
-	    }
+	  struct fdbase f = fdbase (fname);
+	  if (f.fd != BADFD && unlinkat (f.fd, f.base, AT_REMOVEDIR) == 0)
+	    fdbase_clear ();
+	  else if (errno != ENOENT)
+	    rmdir_error (fname);
+
 	  dunlink_reclaim (p);
-	  dunlink_count--;
 	  p = next;
 	}
       dunlink_head = dunlink_tail = NULL;
-    }	  
-	    
+    }
+
   chdir_do (saved_chdir);
 }
 
@@ -197,7 +191,7 @@ void
 finish_deferred_unlinks (void)
 {
   flush_deferred_unlinks (true);
-  
+
   while (dunlink_avail)
     {
       struct deferred_unlink *next = dunlink_avail->next;
@@ -212,7 +206,7 @@ queue_deferred_unlink (const char *name, bool is_dir)
   struct deferred_unlink *p;
 
   if (dunlink_head
-      && records_written > dunlink_head->records_written + deferred_unlink_delay)
+      && records_written > dunlink_head->records_written)
     flush_deferred_unlinks (false);
 
   p = dunlink_alloc ();
@@ -223,11 +217,11 @@ queue_deferred_unlink (const char *name, bool is_dir)
   p->is_dir = is_dir;
   p->records_written = records_written;
 
-  if (IS_CWD (p))
+  if (is_cwd (p))
     {
       struct deferred_unlink *q, *prev;
       for (q = dunlink_head, prev = NULL; q; prev = q, q = q->next)
-	if (IS_CWD (q) && q->dir_idx < p->dir_idx)
+	if (is_cwd (q) && q->dir_idx < p->dir_idx)
 	  break;
       if (q)
 	dunlink_insert (prev, p);
