@@ -1,6 +1,6 @@
 /* Update a tar archive.
 
-   Copyright 1988-2023 Free Software Foundation, Inc.
+   Copyright 1988-2025 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -26,10 +26,6 @@
 #include <quotearg.h>
 #include "common.h"
 
-/* FIXME: This module should not directly handle the following variable,
-   instead, this should be done in buffer.c only.  */
-extern union block *current_block;
-
 /* We've hit the end of the old stuff, and its time to start writing new
    stuff to the tape.  This involves seeking back one record and
    re-writing the current record (which has been changed).
@@ -60,23 +56,19 @@ append_file (char *file_name)
   while (true)
     {
       union block *start = find_next_block ();
-      size_t status = full_read (handle, start->buffer,
-				 available_space_after (start));
-      if (status == 0)
-	{
-	  if (errno == 0)
-	    break;
-	  read_fatal (file_name);
-	}
-      if (status == SAFE_READ_ERROR)
+      idx_t bufsize = available_space_after (start);
+      idx_t status = full_read (handle, charptr (start), bufsize);
+      if (status < bufsize && errno)
 	read_fatal (file_name);
-      if (status % BLOCKSIZE)
-	memset (start->buffer + status - status % BLOCKSIZE, 0,
-		BLOCKSIZE - status % BLOCKSIZE);
-      set_next_block_after (start + (status - 1) / BLOCKSIZE);
+      if (status == 0)
+	break;
+      idx_t rem = status % BLOCKSIZE;
+      if (rem)
+	memset (charptr (start) + (status - rem), 0, BLOCKSIZE - rem);
+      set_next_block_after (charptr (start) + status - 1);
     }
 
-  if (close (handle) != 0)
+  if (close (handle) < 0)
     close_error (file_name);
 }
 
@@ -130,7 +122,7 @@ update_archive (void)
 	    struct name *name;
 
 	    decode_header (current_header, &current_stat_info,
-			   &current_format, 0);
+			   &current_format, false);
 	    transform_stat_info (current_header->header.typeflag,
 				 &current_stat_info);
 	    archive_format = current_format;
@@ -145,7 +137,9 @@ update_archive (void)
 		  {
 		    if (S_ISDIR (s.st_mode))
 		      {
-			char *p, *dirp = tar_savedir (current_stat_info.file_name, 1);
+			char *p;
+			char *dirp = tar_savedir (current_stat_info.file_name,
+						  true);
 			if (dirp)
 			  {
 			    namebuf_t nbuf = namebuf_create (current_stat_info.file_name);
@@ -190,11 +184,11 @@ update_archive (void)
 	  switch (previous_status)
 	    {
 	    case HEADER_STILL_UNREAD:
-	      WARN ((0, 0, _("This does not look like a tar archive")));
+	      paxwarn (0, _("This does not look like a tar archive"));
 	      FALLTHROUGH;
 	    case HEADER_SUCCESS:
 	    case HEADER_ZERO_BLOCK:
-	      ERROR ((0, 0, _("Skipping to next header")));
+	      paxerror (0, _("Skipping to next header"));
 	      FALLTHROUGH;
 	    case HEADER_FAILURE:
 	      break;
@@ -212,7 +206,7 @@ update_archive (void)
 
   reset_eof ();
   time_to_start_writing = true;
-  output_start = current_block->buffer;
+  output_start = charptr (current_block);
 
   {
     struct name const *p;
