@@ -1,6 +1,6 @@
 /* Charset handling for GNU tar.
 
-   Copyright 2004-2023 Free Software Foundation, Inc.
+   Copyright 2004-2026 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -35,7 +35,8 @@
 # define iconv_open(tocode, fromcode) ((iconv_t) -1)
 
 # undef iconv
-# define iconv(cd, inbuf, inbytesleft, outbuf, outbytesleft) (errno = ENOSYS, (size_t) -1)
+# define iconv(cd, inbuf, inbytesleft, outbuf, outbytesleft) \
+    (errno = ENOSYS, SIZE_MAX)
 
 # undef iconv_close
 # define iconv_close(cd) 0
@@ -53,14 +54,14 @@ static iconv_t conv_desc[2] = { (iconv_t) -1, (iconv_t) -1 };
 static iconv_t
 utf8_init (bool to_utf)
 {
-  if (conv_desc[(int) to_utf] == (iconv_t) -1)
+  if (conv_desc[to_utf] == (iconv_t) -1)
     {
       if (to_utf)
-	conv_desc[(int) to_utf] = iconv_open ("UTF-8", locale_charset ());
+	conv_desc[to_utf] = iconv_open ("UTF-8", locale_charset ());
       else
-	conv_desc[(int) to_utf] = iconv_open (locale_charset (), "UTF-8");
+	conv_desc[to_utf] = iconv_open (locale_charset (), "UTF-8");
     }
-  return conv_desc[(int) to_utf];
+  return conv_desc[to_utf];
 }
 
 bool
@@ -72,16 +73,19 @@ utf8_convert (bool to_utf, char const *input, char **output)
   size_t outlen;
   iconv_t cd = utf8_init (to_utf);
 
-  if (cd == 0)
+  if (!cd)
     {
       *output = xstrdup (input);
       return true;
     }
-  else if (cd == (iconv_t)-1)
+  else if (cd == (iconv_t) -1)
     return false;
 
   inlen = strlen (input) + 1;
-  outlen = inlen * MB_LEN_MAX + 1;
+  bool overflow = ckd_mul (&outlen, inlen, MB_LEN_MAX);
+  overflow |= ckd_add (&outlen, outlen, 1);
+  if (overflow)
+    xalloc_die ();
   ob = ret = xmalloc (outlen);
   ib = (char ICONV_CONST *) input;
   /* According to POSIX, "if iconv() encounters a character in the input
@@ -90,7 +94,7 @@ utf8_convert (bool to_utf, char const *input, char **output)
      implementation-defined conversion on this character." It will "update
      the variables pointed to by the arguments to reflect the extent of the
      conversion and return the number of non-identical conversions performed".
-     On error, it returns -1. 
+     On error, it returns SIZE_MAX.
      In other words, non-zero return always indicates failure, either because
      the input was not fully converted, or because it was converted in a
      non-reversible way.
