@@ -1,6 +1,6 @@
 /* Support for extended attributes.
 
-   Copyright (C) 2006-2021 Free Software Foundation, Inc.
+   Copyright (C) 2006-2026 Free Software Foundation, Inc.
 
    This file is part of GNU tar.
 
@@ -30,11 +30,82 @@
 #include "xattr-at.h"
 #include "selinux-at.h"
 
+static char const XATTRS_PREFIX[] = "SCHILY.xattr.";
+enum { XATTRS_PREFIX_LEN = sizeof XATTRS_PREFIX - 1 };
+
+void
+xheader_xattr_init (struct tar_stat_info *st)
+{
+  xattr_map_init (&st->xattr_map);
+
+  st->acls_a_ptr = NULL;
+  st->acls_a_len = 0;
+  st->acls_d_ptr = NULL;
+  st->acls_d_len = 0;
+  st->cntx_name = NULL;
+}
+
+void
+xattr_map_init (struct xattr_map *map)
+{
+  memset (map, 0, sizeof *map);
+}
+
+void
+xattr_map_free (struct xattr_map *xattr_map)
+{
+  for (idx_t i = 0; i < xattr_map->xm_size; i++)
+    {
+      free (xattr_map->xm_map[i].xkey);
+      free (xattr_map->xm_map[i].xval_ptr);
+    }
+  free (xattr_map->xm_map);
+}
+
+void
+xattr_map_add (struct xattr_map *map,
+	       const char *key, const char *val, idx_t len)
+{
+  if (map->xm_size == map->xm_max)
+    map->xm_map = xpalloc (map->xm_map, &map->xm_max, 1, -1,
+			   sizeof *map->xm_map);
+  struct xattr_array *p = &map->xm_map[map->xm_size];
+  p->xkey = xstrdup (key);
+  p->xval_ptr = ximemdup (val, len + 1);
+  p->xval_len = len;
+  map->xm_size++;
+}
+
+MAYBE_UNUSED static void
+xheader_xattr_add (struct tar_stat_info *st,
+		   const char *key, const char *val, idx_t len)
+{
+  idx_t klen = strlen (key);
+  char *xkey = xmalloc (XATTRS_PREFIX_LEN + klen + 1);
+  char *tmp = xkey;
+
+  tmp = stpcpy (tmp, XATTRS_PREFIX);
+  stpcpy (tmp, key);
+
+  xattr_map_add (&st->xattr_map, xkey, val, len);
+
+  free (xkey);
+}
+
+void
+xattr_map_copy (struct xattr_map *dst, const struct xattr_map *src)
+{
+  for (idx_t i = 0; i < src->xm_size; i++)
+    xattr_map_add (dst, src->xm_map[i].xkey,
+		   src->xm_map[i].xval_ptr,
+		   src->xm_map[i].xval_len);
+}
+
 struct xattrs_mask_map
 {
   const char **masks;
-  size_t size;
-  size_t used;
+  idx_t size;
+  idx_t used;
 };
 
 /* list of fnmatch patterns */
@@ -53,6 +124,10 @@ static struct
 #ifdef HAVE_POSIX_ACLS
 # include "acl.h"
 # include <sys/acl.h>
+# ifdef HAVE_ACL_LIBACL_H
+#  /* needed for numeric-owner support */
+#  include <acl/libacl.h>
+# endif
 #endif
 
 #ifdef HAVE_POSIX_ACLS
@@ -179,7 +254,7 @@ fixup_extra_acl_fields (char *ptr)
   while (*src)
     {
       const char *old = src;
-      size_t len = 0;
+      idx_t len = 0;
 
       src = skip_to_ext_fields (src);
       len = src - old;
@@ -203,14 +278,13 @@ fixup_extra_acl_fields (char *ptr)
    attribute.  Called only when acls_option > 0. */
 static void
 xattrs__acls_set (struct tar_stat_info const *st,
-                  char const *file_name, int type,
-                  char *ptr, size_t len, bool def)
+		  char const *file_name, acl_type_t type,
+		  char *ptr, bool def)
 {
   acl_t acl;
 
   if (ptr)
     {
-      /* assert (strlen (ptr) == len); */
       ptr = fixup_extra_acl_fields (ptr);
       acl = acl_from_text (ptr);
     }
@@ -219,12 +293,12 @@ xattrs__acls_set (struct tar_stat_info const *st,
       /* No "default" IEEE 1003.1e ACL set for directory.  At this moment,
          FILE_NAME may already have inherited default acls from parent
          directory;  clean them up. */
-      if (acl_delete_def_file_at (chdir_fd, file_name))
-        WARNOPT (WARN_XATTR_WRITE,
-                (0, errno,
+      struct fdbase f1 = fdbase (file_name);
+      if (f1.fd == BADFD || acl_delete_def_file_at (f1.fd, f1.base) < 0)
+	warnopt (WARN_XATTR_WRITE, errno,
                  _("acl_delete_def_file_at: Cannot drop default POSIX ACLs "
                    "for file '%s'"),
-                 file_name));
+		 quote (file_name));
       return;
     }
   else
@@ -236,12 +310,12 @@ xattrs__acls_set (struct tar_stat_info const *st,
       return;
     }
 
-  if (acl_set_file_at (chdir_fd, file_name, type, acl) == -1)
+  struct fdbase f = fdbase (file_name);
+  if (f.fd == BADFD || acl_set_file_at (f.fd, f.base, type, acl) < 0)
     /* warn even if filesystem does not support acls */
-    WARNOPT (WARN_XATTR_WRITE,
-	     (0, errno,
-	      _ ("acl_set_file_at: Cannot set POSIX ACLs for file '%s'"),
-	      file_name));
+    warnopt (WARN_XATTR_WRITE, errno,
+	     _ ("acl_set_file_at: Cannot set POSIX ACLs for file '%s'"),
+	     quote (file_name));
 
   acl_free (acl);
 }
@@ -249,7 +323,7 @@ xattrs__acls_set (struct tar_stat_info const *st,
 /* Cleanup textual representation of the ACL in VAL by eliminating tab
    characters and comments */
 static void
-xattrs_acls_cleanup (char *val, size_t *plen)
+xattrs_acls_cleanup (char *val, idx_t *plen)
 {
   char *p, *q;
 
@@ -271,21 +345,36 @@ xattrs_acls_cleanup (char *val, size_t *plen)
 }
 
 static void
-xattrs__acls_get_a (int parentfd, const char *file_name,
-                    struct tar_stat_info *st,
-                    char **ret_ptr, size_t * ret_len)
+acls_get_text (int parentfd, const char *file_name, acl_type_t type,
+	       char **ret_ptr, idx_t *ret_len)
 {
   char *val = NULL;
   acl_t acl;
 
-  if (!(acl = acl_get_file_at (parentfd, file_name, ACL_TYPE_ACCESS)))
+  if (!(acl = acl_get_file_at (parentfd, file_name, type)))
     {
       if (errno != ENOTSUP)
         call_arg_warn ("acl_get_file_at", file_name);
       return;
     }
 
-  val = acl_to_text (acl, NULL);
+  if (numeric_owner_option)
+    {
+#ifdef HAVE_ACL_LIBACL_H
+      val = acl_to_any_text (acl, NULL, '\n',
+			     TEXT_SOME_EFFECTIVE | TEXT_NUMERIC_IDS);
+#else
+      static bool warned;
+      if (!warned)
+	{
+	  warned = true;
+	  paxwarn (0, _("--numeric-owner is ignored for ACLs:"
+			" libacl is not available"));
+	}
+#endif
+    }
+  else
+    val = acl_to_text (acl, NULL);
   acl_free (acl);
 
   if (!val)
@@ -299,46 +388,31 @@ xattrs__acls_get_a (int parentfd, const char *file_name,
   acl_free (val);
 }
 
+static void
+xattrs__acls_get_a (int parentfd, const char *file_name,
+                    char **ret_ptr, idx_t *ret_len)
+{
+  acls_get_text (parentfd, file_name, ACL_TYPE_ACCESS, ret_ptr, ret_len);
+}
+
 /* "system.posix_acl_default" */
 static void
 xattrs__acls_get_d (int parentfd, char const *file_name,
-                    struct tar_stat_info *st,
-                    char **ret_ptr, size_t * ret_len)
+                    char **ret_ptr, idx_t *ret_len)
 {
-  char *val = NULL;
-  acl_t acl;
-
-  if (!(acl = acl_get_file_at (parentfd, file_name, ACL_TYPE_DEFAULT)))
-    {
-      if (errno != ENOTSUP)
-        call_arg_warn ("acl_get_file_at", file_name);
-      return;
-    }
-
-  val = acl_to_text (acl, NULL);
-  acl_free (acl);
-
-  if (!val)
-    {
-      call_arg_warn ("acl_to_text", file_name);
-      return;
-    }
-
-  *ret_ptr = xstrdup (val);
-  xattrs_acls_cleanup (*ret_ptr, ret_len);
-  acl_free (val);
+  acls_get_text (parentfd, file_name, ACL_TYPE_DEFAULT, ret_ptr, ret_len);
 }
 #endif /* HAVE_POSIX_ACLS */
 
 static void
 acls_one_line (const char *prefix, char delim,
-               const char *aclstring, size_t len)
+               const char *aclstring, idx_t len)
 {
   /* support both long and short text representation of posix acls */
   struct obstack stk;
-  int pref_len = strlen (prefix);
+  idx_t pref_len = strlen (prefix);
   const char *oldstring = aclstring;
-  int pos = 0;
+  idx_t pos = 0;
 
   if (!aclstring || !len)
     return;
@@ -346,7 +420,7 @@ acls_one_line (const char *prefix, char delim,
   obstack_init (&stk);
   while (pos <= len)
     {
-      int move = strcspn (aclstring, ",\n");
+      idx_t move = strcspn (aclstring, ",\n");
       if (!move)
         break;
 
@@ -362,58 +436,63 @@ acls_one_line (const char *prefix, char delim,
 
   obstack_1grow (&stk, '\0');
 
-  fprintf (stdlis, "%s", (char *) obstack_finish (&stk));
+  fputs (obstack_finish (&stk), stdlis);
 
   obstack_free (&stk, NULL);
 }
 
 void
-xattrs_acls_get (int parentfd, char const *file_name,
-                 struct tar_stat_info *st, int fd, int xisfile)
+xattrs_acls_get (MAYBE_UNUSED int parentfd, MAYBE_UNUSED char const *file_name,
+		 MAYBE_UNUSED struct tar_stat_info *st,
+		 MAYBE_UNUSED bool xisfile)
 {
   if (acls_option > 0)
     {
 #ifndef HAVE_POSIX_ACLS
-      static int done = 0;
+      static bool done;
       if (!done)
-        WARN ((0, 0, _("POSIX ACL support is not available")));
-      done = 1;
+	{
+	  done = true;
+	  paxwarn (0, _("POSIX ACL support is not available"));
+	}
 #else
       int err = file_has_acl_at (parentfd, file_name, &st->stat);
       if (err == 0)
         return;
-      if (err == -1)
+      if (err < 0)
         {
           call_arg_warn ("file_has_acl_at", file_name);
           return;
         }
 
-      xattrs__acls_get_a (parentfd, file_name, st,
+      xattrs__acls_get_a (parentfd, file_name,
                           &st->acls_a_ptr, &st->acls_a_len);
       if (!xisfile)
-        xattrs__acls_get_d (parentfd, file_name, st,
+	xattrs__acls_get_d (parentfd, file_name,
                             &st->acls_d_ptr, &st->acls_d_len);
 #endif
     }
 }
 
 void
-xattrs_acls_set (struct tar_stat_info const *st,
-                 char const *file_name, char typeflag)
+xattrs_acls_set (MAYBE_UNUSED struct tar_stat_info const *st,
+                 MAYBE_UNUSED char const *file_name, char typeflag)
 {
   if (acls_option > 0 && typeflag != SYMTYPE)
     {
 #ifndef HAVE_POSIX_ACLS
-      static int done = 0;
+      static bool done;
       if (!done)
-        WARN ((0, 0, _("POSIX ACL support is not available")));
-      done = 1;
+	{
+	  done = true;
+	  paxwarn (0, _("POSIX ACL support is not available"));
+	}
 #else
       xattrs__acls_set (st, file_name, ACL_TYPE_ACCESS,
-                        st->acls_a_ptr, st->acls_a_len, false);
+			st->acls_a_ptr, false);
       if (typeflag == DIRTYPE || typeflag == GNUTYPE_DUMPDIR)
         xattrs__acls_set (st, file_name, ACL_TYPE_DEFAULT,
-                          st->acls_d_ptr, st->acls_d_len, true);
+			  st->acls_d_ptr, true);
 #endif
     }
 }
@@ -422,11 +501,7 @@ static void
 mask_map_realloc (struct xattrs_mask_map *map)
 {
   if (map->used == map->size)
-    {
-      if (map->size == 0)
-	map->size = 4;
-      map->masks = x2nrealloc (map->masks, &map->size, sizeof (map->masks[0]));
-    }
+    map->masks = xpalloc (map->masks, &map->size, 1, -1, sizeof *map->masks);
 }
 
 void
@@ -441,20 +516,6 @@ xattrs_mask_add (const char *mask, bool incl)
   mask_map->masks[mask_map->used++] = mask;
 }
 
-static void
-clear_mask_map (struct xattrs_mask_map *mask_map)
-{
-  if (mask_map->size)
-    free (mask_map->masks);
-}
-
-void
-xattrs_clear_setup (void)
-{
-  clear_mask_map (&xattrs_setup.incl);
-  clear_mask_map (&xattrs_setup.excl);
-}
-
 static bool xattrs_masked_out (const char *kw, bool archiving);
 
 /* get xattrs from file given by FILE_NAME or FD (when non-zero)
@@ -462,59 +523,59 @@ static bool xattrs_masked_out (const char *kw, bool archiving);
    if no mask is given this includes all the user.*, security.*, system.*,
    etc. available domains */
 void
-xattrs_xattrs_get (int parentfd, char const *file_name,
-                   struct tar_stat_info *st, int fd)
+xattrs_xattrs_get (MAYBE_UNUSED int parentfd,
+		   MAYBE_UNUSED char const *file_name,
+		   MAYBE_UNUSED struct tar_stat_info *st, MAYBE_UNUSED int fd)
 {
-  if (xattrs_option > 0)
+  if (xattrs_option)
     {
 #ifndef HAVE_XATTRS
-      static int done = 0;
+      static bool done;
       if (!done)
-        WARN ((0, 0, _("XATTR support is not available")));
-      done = 1;
+	{
+	  done = true;
+	  paxwarn (0, _("XATTR support is not available"));
+	}
 #else
-      static size_t xsz = 1024;
+      static idx_t xsz = 1024 / 2 * 3;
       static char *xatrs = NULL;
-      ssize_t xret = -1;
+      ssize_t xret;
 
-      if (!xatrs)
-	xatrs = x2nrealloc (xatrs, &xsz, 1);
-
-      while (((fd == 0) ?
-              ((xret =
-                llistxattrat (parentfd, file_name, xatrs, xsz)) == -1) :
-	      ((xret = flistxattr (fd, xatrs, xsz)) == -1))
-             && (errno == ERANGE))
+      while (!xatrs
+	     || (((xret = (fd == 0
+			   ? listxattrat (parentfd, file_name, xatrs, xsz)
+			   : flistxattr (fd, xatrs, xsz)))
+		  < 0)
+		 && errno == ERANGE))
         {
-	  xatrs = x2nrealloc (xatrs, &xsz, 1);
+	  xatrs = xpalloc (xatrs, &xsz, 1, -1, sizeof *xatrs);
         }
 
-      if (xret == -1)
+      if (xret < 0)
         call_arg_warn ((fd == 0) ? "llistxattrat" : "flistxattr", file_name);
       else
         {
           const char *attr = xatrs;
-          static size_t asz = 1024;
+          static idx_t asz = 1024 / 2 * 3;
           static char *val = NULL;
 
-          if (!val)
-            val = x2nrealloc (val, &asz, 1);
-
           while (xret > 0)
             {
-              size_t len = strlen (attr);
+              idx_t len = strlen (attr);
               ssize_t aret = 0;
 
-              while (((fd == 0)
-                      ? ((aret = lgetxattrat (parentfd, file_name, attr,
-                                              val, asz)) == -1)
-                      : ((aret = fgetxattr (fd, attr, val, asz)) == -1))
-                     && (errno == ERANGE))
+	      while (!val
+		     || (((aret = (fd == 0
+				   ? lgetxattrat (parentfd, file_name, attr,
+						  val, asz)
+				   : fgetxattr (fd, attr, val, asz)))
+			  < 0)
+			 && errno == ERANGE))
                 {
-		  val = x2nrealloc (val, &asz, 1);
+		  val = xpalloc (val, &asz, 1, -1, sizeof *val);
                 }
 
-              if (aret != -1)
+              if (0 <= aret)
                 {
                   if (!xattrs_masked_out (attr, true))
                     xheader_xattr_add (st, attr, val, aret);
@@ -533,28 +594,29 @@ xattrs_xattrs_get (int parentfd, char const *file_name,
 
 #ifdef HAVE_XATTRS
 static void
-xattrs__fd_set (struct tar_stat_info const *st,
-                char const *file_name, char typeflag,
-                const char *attr, const char *ptr, size_t len)
+xattrs__fd_set (char const *file_name, char typeflag,
+                const char *attr, const char *ptr, idx_t len)
 {
   if (ptr)
     {
       const char *sysname = "setxattrat";
-      int ret = -1;
+      int ret;
+      struct fdbase f = fdbase (file_name);
 
-      if (typeflag != SYMTYPE)
-        ret = setxattrat (chdir_fd, file_name, attr, ptr, len, 0);
+      if (f.fd == BADFD)
+	ret = -1;
+      else if (typeflag != SYMTYPE)
+	ret = setxattrat (f.fd, f.base, attr, ptr, len, 0);
       else
         {
           sysname = "lsetxattr";
-          ret = lsetxattrat (chdir_fd, file_name, attr, ptr, len, 0);
+	  ret = lsetxattrat (f.fd, f.base, attr, ptr, len, 0);
         }
 
-      if (ret == -1)
-        WARNOPT (WARN_XATTR_WRITE,
-		 (0, errno,
-		  _("%s: Cannot set '%s' extended attribute for file '%s'"),
-		  sysname, attr, file_name));
+      if (ret < 0)
+	warnopt (WARN_XATTR_WRITE, errno,
+		 _("%s: Cannot set '%s' extended attribute for file '%s'"),
+		 sysname, attr, quote (file_name));
     }
 }
 #endif
@@ -562,38 +624,42 @@ xattrs__fd_set (struct tar_stat_info const *st,
 /* lgetfileconat is called against FILE_NAME iff the FD parameter is set to
    zero, otherwise the fgetfileconat is used against correct file descriptor */
 void
-xattrs_selinux_get (int parentfd, char const *file_name,
-                    struct tar_stat_info *st, int fd)
+xattrs_selinux_get (MAYBE_UNUSED int parentfd, MAYBE_UNUSED char const *file_name,
+                    MAYBE_UNUSED struct tar_stat_info *st, MAYBE_UNUSED int fd)
 {
   if (selinux_context_option > 0)
     {
 #if HAVE_SELINUX_SELINUX_H != 1
-      static int done = 0;
+      static bool done;
       if (!done)
-        WARN ((0, 0, _("SELinux support is not available")));
-      done = 1;
+	{
+	  done = true;
+	  paxwarn (0, _("SELinux support is not available"));
+	}
 #else
-      int result = fd ?
-	            fgetfilecon (fd, &st->cntx_name)
-                    : lgetfileconat (parentfd, file_name, &st->cntx_name);
+      int result = (fd
+		    ? fgetfilecon (fd, &st->cntx_name)
+		    : lgetfileconat (parentfd, file_name, &st->cntx_name));
 
-      if (result == -1 && errno != ENODATA && errno != ENOTSUP)
+      if (result < 0 && errno != ENODATA && errno != ENOTSUP)
         call_arg_warn (fd ? "fgetfilecon" : "lgetfileconat", file_name);
 #endif
     }
 }
 
 void
-xattrs_selinux_set (struct tar_stat_info const *st,
-                    char const *file_name, char typeflag)
+xattrs_selinux_set (MAYBE_UNUSED struct tar_stat_info const *st,
+                    MAYBE_UNUSED char const *file_name, MAYBE_UNUSED char typeflag)
 {
   if (selinux_context_option > 0)
     {
 #if HAVE_SELINUX_SELINUX_H != 1
-      static int done = 0;
+      static bool done;
       if (!done)
-        WARN ((0, 0, _("SELinux support is not available")));
-      done = 1;
+	{
+	  done = true;
+	  paxwarn (0, _("SELinux support is not available"));
+	}
 #else
       const char *sysname = "setfilecon";
       int ret;
@@ -601,22 +667,24 @@ xattrs_selinux_set (struct tar_stat_info const *st,
       if (!st->cntx_name)
         return;
 
-      if (typeflag != SYMTYPE)
+      struct fdbase f = fdbase (file_name);
+      if (f.fd == BADFD)
+	ret = -1;
+      else if (typeflag != SYMTYPE)
         {
-          ret = setfileconat (chdir_fd, file_name, st->cntx_name);
+	  ret = setfileconat (f.fd, f.base, st->cntx_name);
           sysname = "setfileconat";
         }
       else
         {
-          ret = lsetfileconat (chdir_fd, file_name, st->cntx_name);
+	  ret = lsetfileconat (f.fd, f.base, st->cntx_name);
           sysname = "lsetfileconat";
         }
 
-      if (ret == -1)
-        WARNOPT (WARN_XATTR_WRITE,
-		 (0, errno,
-		  _("%s: Cannot set SELinux context for file '%s'"),
-		  sysname, file_name));
+      if (ret < 0)
+	warnopt (WARN_XATTR_WRITE, errno,
+		 _("%s: Cannot set SELinux context for file '%s'"),
+		 sysname, quote (file_name));
 #endif
     }
 }
@@ -624,23 +692,20 @@ xattrs_selinux_set (struct tar_stat_info const *st,
 static bool
 xattrs_matches_mask (const char *kw, struct xattrs_mask_map *mm)
 {
-  int i;
-
   if (!mm->size)
     return false;
 
-  for (i = 0; i < mm->used; i++)
+  for (idx_t i = 0; i < mm->used; i++)
     if (fnmatch (mm->masks[i], kw, 0) == 0)
       return true;
 
   return false;
 }
 
-#define USER_DOT_PFX "user."
-
 static bool
 xattrs_kw_included (const char *kw, bool archiving)
 {
+  static char const USER_DOT_PFX[] = "user.";
   if (xattrs_setup.incl.size)
     return xattrs_matches_mask (kw, &xattrs_setup.incl);
   else if (archiving)
@@ -650,7 +715,7 @@ xattrs_kw_included (const char *kw, bool archiving)
 }
 
 static bool
-xattrs_kw_excluded (const char *kw, bool archiving)
+xattrs_kw_excluded (const char *kw)
 {
   return xattrs_setup.excl.size ?
     xattrs_matches_mask (kw, &xattrs_setup.excl) : false;
@@ -662,31 +727,30 @@ xattrs_kw_excluded (const char *kw, bool archiving)
 static bool
 xattrs_masked_out (const char *kw, bool archiving)
 {
-  return xattrs_kw_included (kw, archiving) ?
-    xattrs_kw_excluded (kw, archiving) : true;
+  return xattrs_kw_included (kw, archiving) ? xattrs_kw_excluded (kw) : true;
 }
 
 void
-xattrs_xattrs_set (struct tar_stat_info const *st,
-                   char const *file_name, char typeflag, int later_run)
+xattrs_xattrs_set (MAYBE_UNUSED struct tar_stat_info const *st,
+		   MAYBE_UNUSED char const *file_name,
+		   MAYBE_UNUSED char typeflag, MAYBE_UNUSED bool later_run)
 {
-  if (xattrs_option > 0)
+  if (xattrs_option)
     {
 #ifndef HAVE_XATTRS
-      static int done = 0;
+      static bool done;
       if (!done)
-        WARN ((0, 0, _("XATTR support is not available")));
-      done = 1;
+	{
+	  done = true;
+	  paxwarn (0, _("XATTR support is not available"));
+	}
 #else
-      size_t scan = 0;
-
-      if (!st->xattr_map_size)
+      if (!st->xattr_map.xm_size)
         return;
 
-      for (; scan < st->xattr_map_size; ++scan)
+      for (idx_t i = 0; i < st->xattr_map.xm_size; i++)
         {
-          char *keyword = st->xattr_map[scan].xkey;
-          keyword += strlen ("SCHILY.xattr.");
+          char *keyword = st->xattr_map.xm_map[i].xkey + XATTRS_PREFIX_LEN;
 
           /* TODO: this 'later_run' workaround is temporary solution -> once
              capabilities should become fully supported by it's API and there
@@ -695,16 +759,16 @@ xattrs_xattrs_set (struct tar_stat_info const *st,
              the first run except 'security.capability' which is restored in
              'later_run == 1'.  */
           if (typeflag == REGTYPE
-              && later_run == !!strcmp (keyword, "security.capability"))
+	      && streq (keyword, "security.capability") != later_run)
             continue;
 
           if (xattrs_masked_out (keyword, false /* extracting */ ))
             /* we don't want to restore this keyword */
             continue;
 
-          xattrs__fd_set (st, file_name, typeflag, keyword,
-                          st->xattr_map[scan].xval_ptr,
-                          st->xattr_map[scan].xval_len);
+	  xattrs__fd_set (file_name, typeflag, keyword,
+                          st->xattr_map.xm_map[i].xval_ptr,
+                          st->xattr_map.xm_map[i].xval_len);
         }
 #endif
     }
@@ -713,25 +777,23 @@ xattrs_xattrs_set (struct tar_stat_info const *st,
 void
 xattrs_print_char (struct tar_stat_info const *st, char *output)
 {
-  int i;
-
   if (verbose_option < 2)
     {
       *output = 0;
       return;
     }
 
-  if (xattrs_option > 0 || selinux_context_option > 0 || acls_option > 0)
+  if (xattrs_option || selinux_context_option > 0 || acls_option > 0)
     {
       /* placeholders */
       *output = ' ';
       output[1] = 0;
     }
 
-  if (xattrs_option > 0 && st->xattr_map_size)
-    for (i = 0; i < st->xattr_map_size; ++i)
+  if (xattrs_option && st->xattr_map.xm_size)
+    for (idx_t i = 0; i < st->xattr_map.xm_size; i++)
       {
-        char *keyword = st->xattr_map[i].xkey + strlen ("SCHILY.xattr.");
+        char *keyword = st->xattr_map.xm_map[i].xkey + XATTRS_PREFIX_LEN;
         if (!xattrs_masked_out (keyword, false /* like extracting */ ))
 	  {
 	    *output = '*';
@@ -768,16 +830,14 @@ xattrs_print (struct tar_stat_info const *st)
     }
 
   /* xattrs */
-  if (xattrs_option > 0 && st->xattr_map_size)
+  if (xattrs_option && st->xattr_map.xm_size)
     {
-      int i;
-
-      for (i = 0; i < st->xattr_map_size; ++i)
+      for (idx_t i = 0; i < st->xattr_map.xm_size; i++)
         {
-          char *keyword = st->xattr_map[i].xkey + strlen ("SCHILY.xattr.");
+          char *keyword = st->xattr_map.xm_map[i].xkey + XATTRS_PREFIX_LEN;
           if (!xattrs_masked_out (keyword, false /* like extracting */ ))
-	    fprintf (stdlis, "  x: %lu %s\n",
-		     (unsigned long) st->xattr_map[i].xval_len, keyword);
+	    fprintf (stdlis, "  x: %td %s\n",
+		     st->xattr_map.xm_map[i].xval_len, keyword);
         }
     }
 }
