@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2007 Free Software Foundation, Inc.
+   Copyright (C) 1999-2001, 2007, 2023-2026 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -12,8 +12,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public
-   License along with this library; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA  */
+   License along with this library.  If not, see <http://www.gnu.org/licenses/>.
+*/
 
 #ifndef _ARGCV_H
 #define _ARGCV_H 1
@@ -27,21 +27,11 @@
 extern "C" {
 #endif
 
-#ifndef __P
-# if defined PROTOTYPES || (defined __STDC__ && __STDC__)
-#  define __P(args) args
-# else
-#  define __P(args) ()
-# endif
-#endif /*__P */
-
-extern int argcv_get    __P ((const char *command, const char *delim,
-			      const char* cmnt,
-			      int *argc, char ***argv));
-extern int argcv_string __P ((int argc, char **argv, char **string));
-extern int argcv_free   __P ((int argc, char **argv));
-extern int argcv_unescape_char __P((int c));
-extern int argcv_escape_char __P((int c));
+extern int argcv_get (char const *command, char const *delim, char const* cmnt,
+		      int *argc, char ***argv);
+extern int argcv_string (int argc, char **argv, char **string);
+extern int argcv_unescape_char (int c);
+extern int argcv_escape_char (int c);
 
 #ifdef __cplusplus
 }
