@@ -1,6 +1,6 @@
 /* Multi-purpose tool for tar and cpio testsuite.
 
-   Copyright (C) 1995-1997, 2001-2018 Free Software Foundation, Inc.
+   Copyright (C) 1995-1997, 2001-2018, 2023-2026 Free Software Foundation, Inc.
 
    François Pinard <pinard@iro.umontreal.ca>, 1995.
    Sergey Poznyakoff <gray@gnu.org>, 2004-2018.
@@ -26,7 +26,6 @@
 #include <argp.h>
 #include <argcv.h>
 #include <parse-datetime.h>
-#include <inttostr.h>
 #include <fcntl.h>
 #include <sys/stat.h>
 #include <c-ctype.h>
@@ -34,12 +33,6 @@
 #define obstack_chunk_free free
 #include <obstack.h>
 
-#ifndef EXIT_SUCCESS
-# define EXIT_SUCCESS 0
-#endif
-#ifndef EXIT_FAILURE
-# define EXIT_FAILURE 1
-#endif
 #define EXIT_USAGE 2
 #define EXIT_UNAVAILABLE 3
 
@@ -71,14 +64,15 @@ static off_t seek_offset = 0;
 static enum pattern pattern = DEFAULT_PATTERN;
 
 /* Next checkpoint number */
-size_t checkpoint;
+intmax_t checkpoint;
 
 enum genfile_mode
   {
     mode_generate,
     mode_sparse,
     mode_stat,
-    mode_exec
+    mode_exec,
+    mode_set_times
   };
 
 enum genfile_mode mode = mode_generate;
@@ -90,22 +84,25 @@ enum genfile_mode mode = mode_generate;
 static char *stat_format = DEFAULT_STAT_FORMAT;
 
 /* Size of a block for sparse file */
-size_t block_size = 512;
+idx_t block_size = 512;
 
 /* Block buffer for sparse file */
 char *buffer;
 
-/* Checkpoint granularity for mode == mode_exec */
+/* Checkpoint granularity option for mode == mode_exec */
 char *checkpoint_granularity;
 
 /* Time for --touch option */
 struct timespec touch_time;
 
 /* Verbose mode */
-int verbose;
+bool verbose;
 
 /* Quiet mode */
-int quiet;
+bool quiet;
+
+/* Don't dereference symlinks (for --stat) */
+bool no_dereference_option;
 
 const char *argp_program_version = "genfile (" PACKAGE ") " VERSION;
 const char *argp_program_bug_address = "<" PACKAGE_BUGREPORT ">";
@@ -124,7 +121,7 @@ static char doc[] = N_("genfile manipulates data files for GNU paxutils test sui
 
 static struct argp_option options[] = {
 #define GRP 0
-  {NULL, 0, NULL, 0,
+  {nullptr, 0, nullptr, 0,
    N_("File creation options:"), GRP},
   {"length", 'l', N_("SIZE"), 0,
    N_("Create file of the given SIZE"), GRP+1 },
@@ -132,34 +129,42 @@ static struct argp_option options[] = {
    N_("Write to file NAME, instead of standard output"), GRP+1},
   {"files-from", 'T', N_("FILE"), 0,
    N_("Read file names from FILE"), GRP+1},
-  {"null", '0', NULL, 0,
+  {"null", '0', nullptr, 0,
    N_("-T reads null-terminated names"), GRP+1},
   {"pattern", 'p', N_("PATTERN"), 0,
    N_("Fill the file with the given PATTERN. PATTERN is 'default' or 'zeros'"),
    GRP+1 },
   {"block-size", 'b', N_("SIZE"), 0,
    N_("Size of a block for sparse file"), GRP+1},
-  {"sparse", 's', NULL, 0,
+  {"sparse", 's', nullptr, 0,
    N_("Generate sparse file. Rest of the command line gives the file map."),
    GRP+1 },
   {"seek", OPT_SEEK, N_("OFFSET"), 0,
    N_("Seek to the given offset before writing data"),
    GRP+1 },
-  {"quiet", 'q', NULL, 0,
+  {"quiet", 'q', nullptr, 0,
    N_("Suppress non-fatal diagnostic messages") },
 #undef GRP
 #define GRP 10
-  {NULL, 0, NULL, 0,
+  {nullptr, 0, nullptr, 0,
    N_("File statistics options:"), GRP},
 
   {"stat", 'S', N_("FORMAT"), OPTION_ARG_OPTIONAL,
    N_("Print contents of struct stat for each given file. Default FORMAT is: ")
    DEFAULT_STAT_FORMAT,
    GRP+1 },
+  {"no-dereference", 'h', nullptr, 0,
+   N_("stat symbolic links instead of referenced files"),
+   GRP+1 },
+
+  {"set-times", 't', nullptr, 0,
+   N_("Set access and modification times of the files to the time supplied"
+      " by --date option"),
+   GRP+1 },
 
 #undef GRP
 #define GRP 20
-  {NULL, 0, NULL, 0,
+  {nullptr, 0, nullptr, 0,
    N_("Synchronous execution options:"), GRP},
 
   {"run", 'r', N_("N"), OPTION_ARG_OPTIONAL,
@@ -171,18 +176,18 @@ static struct argp_option options[] = {
   {"date", OPT_DATE, N_("STRING"), 0,
    N_("Set date for next --touch option"),
    GRP+1 },
-  {"verbose", OPT_VERBOSE, NULL, 0,
+  {"verbose", OPT_VERBOSE, nullptr, 0,
    N_("Display executed checkpoints and exit status of COMMAND"),
    GRP+1 },
 #undef GRP
 #define GRP 30
-  {NULL, 0, NULL, 0,
+  {nullptr, 0, nullptr, 0,
    N_("Synchronous execution actions. These are executed when checkpoint number given by --checkpoint option is reached."), GRP},
 
   {"cut", OPT_TRUNCATE, N_("FILE"), 0,
    N_("Truncate FILE to the size specified by previous --length option (or 0, if it is not given)"),
    GRP+1 },
-  {"truncate", 0, NULL, OPTION_ALIAS, NULL, GRP+1 },
+  {"truncate", 0, nullptr, OPTION_ALIAS, nullptr, GRP+1 },
   {"append", OPT_APPEND, N_("FILE"), 0,
    N_("Append SIZE bytes to FILE. SIZE is given by previous --length option."),
    GRP+1 },
@@ -195,65 +200,51 @@ static struct argp_option options[] = {
   {"delete", OPT_DELETE, N_("FILE"), 0,
    N_("Delete FILE"),
    GRP+1 },
-  {"unlink", 0, 0, OPTION_ALIAS, NULL, GRP+1},
+  {"unlink", 0, 0, OPTION_ALIAS, nullptr, GRP+1},
 #undef GRP
-  { NULL, }
+  { nullptr, }
 };
 
-static char const * const pattern_args[] = { "default", "zeros", 0 };
+static char const * const pattern_args[] = { "default", "zeros", nullptr };
 static enum pattern const pattern_types[] = {DEFAULT_PATTERN, ZEROS_PATTERN};
 
+/* Multiply *VP by the base denoted by P.
+   Return -1 if invalid, 0 if OK, 1 if overflow.  */
 static int
 xlat_suffix (off_t *vp, const char *p)
 {
-  off_t val = *vp;
+  if (!p[0])
+    return 0;
 
-  if (p[1])
-    return 1;
-  switch (p[0])
-    {
-    case 'g':
-    case 'G':
-      *vp *= 1024;
+  if (!p[1])
+    switch (p[0])
+      {
+      case 'k': case 'K': return ckd_mul (vp, *vp, 1 << 10);
+      case 'm': case 'M': return ckd_mul (vp, *vp, 1 << 20);
+      case 'g': case 'G': return ckd_mul (vp, *vp, 1 << 30);
+      }
 
-    case 'm':
-    case 'M':
-      *vp *= 1024;
-
-    case 'k':
-    case 'K':
-      *vp *= 1024;
-      break;
-
-    default:
-      return 1;
-    }
-  return *vp <= val;
+  return -1;
 }
 
 static off_t
-get_size (const char *str, int allow_zero)
+get_size (const char *str)
 {
-  const char *p;
-  off_t v = 0;
-
-  for (p = str; *p; p++)
-    {
-      int digit = *p - '0';
-      off_t x = v * 10;
-      if (9 < (unsigned) digit)
-	{
-	  if (xlat_suffix (&v, p))
-	    error (EXIT_USAGE, 0, _("Invalid size: %s"), str);
-	  else
-	    break;
-	}
-      else if (x / 10 != v)
-	error (EXIT_USAGE, 0, _("Number out of allowed range: %s"), str);
-      v = x + digit;
-      if (v < 0)
-	error (EXIT_USAGE, 0, _("Negative size: %s"), str);
-    }
+  char *p;
+  errno = 0;
+  intmax_t s = strtoimax (str, &p, 10);
+  if (p == str || (errno && errno != ERANGE))
+    error (EXIT_USAGE, 0, _("Invalid size: %s"), str);
+  if (s < 0)
+    error (EXIT_USAGE, 0, _("Negative size: %s"), str);
+  off_t v;
+  if (errno || ckd_add (&v, s, 0))
+    error (EXIT_USAGE, 0, _("Number out of allowed range: %s"), str);
+  int bad_suffix = xlat_suffix (&v, p);
+  if (bad_suffix < 0)
+    error (EXIT_USAGE, 0, _("Invalid size: %s"), str);
+  if (bad_suffix)
+    error (EXIT_USAGE, 0, _("Number out of allowed range: %s"), str);
   return v;
 }
 
@@ -267,9 +258,12 @@ verify_file (char *file_name)
       if (stat (file_name, &st))
 	error (0, errno, _("stat(%s) failed"), file_name);
 
-      if (st.st_size != file_length + seek_offset)
-	error (EXIT_FAILURE, 0, _("requested file length %lu, actual %lu"),
-	       (unsigned long)st.st_size, (unsigned long)file_length);
+      if (st.st_size < 0 || st.st_size - seek_offset != file_length)
+	{
+	  intmax_t requested = st.st_size, actual = file_length;
+	  error (EXIT_FAILURE, 0, _("requested file length %jd, actual %jd"),
+		 requested, actual);
+	}
 
       if (!quiet && mode == mode_sparse && !ST_IS_SPARSE (st))
 	error (EXIT_UNAVAILABLE, 0, _("created file is not sparse"));
@@ -279,7 +273,7 @@ verify_file (char *file_name)
 struct action
 {
   struct action *next;
-  size_t checkpoint;
+  intmax_t checkpoint;
   int action;
   char *name;
   off_t size;
@@ -292,14 +286,14 @@ static struct action *action_head, *action_tail;
 void
 reg_action (int action, char *arg)
 {
-  struct action *act = xmalloc (sizeof (*act));
+  struct action *act = ximalloc (sizeof *act);
   act->checkpoint = checkpoint;
   act->action = action;
   act->pattern = pattern;
   act->ts = touch_time;
   act->size = file_length;
   act->name = arg;
-  act->next = NULL;
+  act->next = nullptr;
   if (action_tail)
     action_tail->next = act;
   else
@@ -321,7 +315,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
       break;
 
     case 'l':
-      file_length = get_size (arg, 1);
+      file_length = get_size (arg);
       break;
 
     case 'p':
@@ -329,13 +323,17 @@ parse_opt (int key, char *arg, struct argp_state *state)
       break;
 
     case 'b':
-      block_size = get_size (arg, 0);
+      {
+	off_t s = get_size (arg);
+	if (ckd_add (&block_size, s, 0))
+	  error (EXIT_USAGE, 0, _("Number out of allowed range: %s"), arg);
+      }
       break;
 
     case 'q':
-      quiet = 1;
+      quiet = true;
       break;
-      
+
     case 's':
       mode = mode_sparse;
       break;
@@ -346,9 +344,25 @@ parse_opt (int key, char *arg, struct argp_state *state)
 	stat_format = arg;
       break;
 
+    case 't':
+      mode = mode_set_times;
+      break;
+
+    case 'h':
+      no_dereference_option = true;
+      break;
+
     case 'r':
-      mode = mode_exec;
-      checkpoint_granularity = arg ? arg : "1";
+      {
+	mode = mode_exec;
+	if (!arg)
+	  arg = "1";
+	static char const opt[] = "--checkpoint=";
+	idx_t arglen = strlen (arg);
+	checkpoint_granularity = ximalloc (sizeof opt + arglen);
+	char *p = mempcpy (checkpoint_granularity, opt, sizeof opt - 1);
+	memcpy (p, arg, arglen + 1);
+      }
       break;
 
     case 'T':
@@ -356,21 +370,20 @@ parse_opt (int key, char *arg, struct argp_state *state)
       break;
 
     case OPT_SEEK:
-      seek_offset = get_size (arg, 0);
+      seek_offset = get_size (arg);
       break;
 
     case OPT_CHECKPOINT:
       {
 	char *p;
-
-	checkpoint = strtoul (arg, &p, 0);
-	if (*p)
-	  argp_error (state, _("Error parsing number near `%s'"), p);
+	checkpoint = strtoimax (arg, &p, 0);
+	if (checkpoint < 0 || *p)
+	  argp_error (state, _("Error parsing number near '%s'"), p);
       }
       break;
 
     case OPT_DATE:
-      if (! parse_datetime (&touch_time, arg, NULL))
+      if (! parse_datetime (&touch_time, arg, nullptr))
 	argp_error (state, _("Unknown date format"));
       break;
 
@@ -383,7 +396,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
       break;
 
     case OPT_VERBOSE:
-      verbose++;
+      verbose = true;
       break;
 
     default:
@@ -397,26 +410,24 @@ static struct argp argp = {
   parse_opt,
   N_("[ARGS...]"),
   doc,
-  NULL,
-  NULL,
-  NULL
+  nullptr,
+  nullptr,
+  nullptr
 };
 
 
 void
 fill (FILE *fp, off_t length, enum pattern pattern)
 {
-  off_t i;
-
   switch (pattern)
     {
     case DEFAULT_PATTERN:
-      for (i = 0; i < length; i++)
+      for (off_t i = 0; i < length; i++)
 	fputc (i & 255, fp);
       break;
 
     case ZEROS_PATTERN:
-      for (i = 0; i < length; i++)
+      for (off_t i = 0; i < length; i++)
 	fputc (0, fp);
       break;
     }
@@ -432,12 +443,12 @@ generate_simple_file (char *filename)
     {
       fp = fopen (filename, seek_offset ? "rb+" : "wb");
       if (!fp)
-	error (EXIT_FAILURE, errno, _("cannot open `%s'"), filename);
+	error (EXIT_FAILURE, errno, _("cannot open '%s'"), filename);
     }
   else
     fp = stdout;
 
-  if (fseeko (fp, seek_offset, 0))
+  if (fseeko (fp, seek_offset, SEEK_SET) < 0)
     error (EXIT_FAILURE, errno, "%s", _("cannot seek"));
 
   fill (fp, file_length, pattern);
@@ -446,14 +457,17 @@ generate_simple_file (char *filename)
 }
 
 /* A simplified version of the same function from tar */
-int
+static bool
 read_name_from_file (FILE *fp, struct obstack *stk)
 {
   int c;
-  size_t counter = 0;
+  idx_t counter = 0;
 
-  for (c = getc (fp); c != EOF && c != filename_terminator; c = getc (fp))
+  while (true)
     {
+      c = getc (fp);
+      if (c < 0 || c == filename_terminator)
+	break;
       if (c == 0)
 	error (EXIT_FAILURE, 0, _("file name contains null character"));
       obstack_1grow (stk, c);
@@ -462,7 +476,7 @@ read_name_from_file (FILE *fp, struct obstack *stk)
 
   obstack_1grow (stk, 0);
 
-  return (counter == 0 && c == EOF);
+  return c < 0 && counter == 0;
 }
 
 void
@@ -472,7 +486,7 @@ generate_files_from_list ()
   struct obstack stk;
 
   if (!fp)
-    error (EXIT_FAILURE, errno, _("cannot open `%s'"), files_from);
+    error (EXIT_FAILURE, errno, _("cannot open '%s'"), files_from);
 
   obstack_init (&stk);
   while (!read_name_from_file (fp, &stk))
@@ -483,7 +497,7 @@ generate_files_from_list ()
       obstack_free (&stk, name);
     }
   fclose (fp);
-  obstack_free (&stk, NULL);
+  obstack_free (&stk, nullptr);
 }
 
 
@@ -495,69 +509,76 @@ mkhole (int fd, off_t displ)
   off_t offset = lseek (fd, displ, SEEK_CUR);
   if (offset < 0)
     error (EXIT_FAILURE, errno, "lseek");
-  if (ftruncate (fd, offset) != 0)
+  if (ftruncate (fd, offset) < 0)
     error (EXIT_FAILURE, errno, "ftruncate");
 }
 
 static void
 mksparse (int fd, off_t displ, char *marks)
 {
-  if (lseek (fd, displ, SEEK_CUR) == -1)
+  if (lseek (fd, displ, SEEK_CUR) < 0)
     error (EXIT_FAILURE, errno, "lseek");
 
   for (; *marks; marks++)
     {
       memset (buffer, *marks, block_size);
-      if (write (fd, buffer, block_size) != block_size)
+      ssize_t written = write (fd, buffer, block_size);
+      if (written < 0)
 	error (EXIT_FAILURE, errno, "write");
+      if (written != block_size)
+	error (EXIT_FAILURE, 0, "write");
     }
 }
 
-static int
+static bool
 make_fragment (int fd, char *offstr, char *mapstr)
 {
-  int i;
-  off_t displ = get_size (offstr, 1);
+  off_t displ = get_size (offstr);
 
-  file_length += displ;
+  if (ckd_add (&file_length, file_length, displ))
+    error (EXIT_USAGE, 0, _("Number out of allowed range: %s"), offstr);
 
   if (!mapstr || !*mapstr)
     {
       mkhole (fd, displ);
-      return 1;
+      return true;
     }
   else if (*mapstr == '=')
     {
-      off_t n = get_size (mapstr + 1, 1);
+      off_t n = get_size (mapstr + 1);
 
       switch (pattern)
 	{
 	case DEFAULT_PATTERN:
-	  for (i = 0; i < block_size; i++)
+	  for (idx_t i = 0; i < block_size; i++)
 	    buffer[i] = i & 255;
 	  break;
-	  
+
 	case ZEROS_PATTERN:
 	  memset (buffer, 0, block_size);
 	  break;
 	}
 
-      if (lseek (fd, displ, SEEK_CUR) == -1)
+      if (lseek (fd, displ, SEEK_CUR) < 0)
 	error (EXIT_FAILURE, errno, "lseek");
-      
+
       for (; n; n--)
 	{
-	  if (write (fd, buffer, block_size) != block_size)
+	  ssize_t written = write (fd, buffer, block_size);
+	  if (written < 0)
 	    error (EXIT_FAILURE, errno, "write");
+	  if (written != block_size)
+	    error (EXIT_FAILURE, 0, "write");
 	  file_length += block_size;
 	}
     }
   else
     {
-      file_length += block_size * strlen (mapstr);
       mksparse (fd, displ, mapstr);
+      off_t mapstrlen = strlen (mapstr);
+      file_length += block_size * mapstrlen;
     }
-  return 0;
+  return false;
 }
 
 static void
@@ -573,24 +594,24 @@ generate_sparse_file (int argc, char **argv)
     flags |= O_TRUNC;
   fd = open (file_name, flags, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
   if (fd < 0)
-    error (EXIT_FAILURE, errno, _("cannot open `%s'"), file_name);
+    error (EXIT_FAILURE, errno, _("cannot open '%s'"), file_name);
 
-  buffer = xmalloc (block_size);
+  buffer = ximalloc (block_size);
 
   file_length = 0;
 
   while (argc)
     {
-      if (argv[0][0] == '-' && argv[0][1] == 0)
+      if (argv[0][0] == '-' && !argv[0][1])
 	{
 	  char buf[256];
 	  while (fgets (buf, sizeof (buf), stdin))
 	    {
-	      size_t n = strlen (buf);
+	      idx_t n = strlen (buf);
 
-	      while (n > 0 && c_isspace (buf[n-1]))
+	      while (n > 0 && c_isspace (buf[n - 1]))
 		buf[--n] = 0;
-	      
+
 	      n = strcspn (buf, " \t");
 	      buf[n++] = 0;
 	      while (buf[n] && c_isblank (buf[n]))
@@ -615,6 +636,22 @@ generate_sparse_file (int argc, char **argv)
 
 /* Status Mode */
 
+#define PRINT_INT(expr) \
+  do \
+    { \
+      if (EXPR_SIGNED (expr)) \
+	{ \
+	  intmax_t printval = expr; \
+	  printf ("%jd", printval); \
+	} \
+      else \
+	{ \
+	  uintmax_t printval = expr; \
+	  printf ("%ju", printval); \
+	} \
+    } \
+  while (false)
+
 void
 print_time (time_t t)
 {
@@ -626,71 +663,69 @@ print_time (time_t t)
 void
 print_stat (const char *name)
 {
-  char *fmt, *p;
   struct stat st;
-  char buf[UINTMAX_STRSIZE_BOUND];
 
-  if (stat (name, &st))
+  if ((no_dereference_option ? lstat : stat) (name, &st))
     {
       error (0, errno, _("stat(%s) failed"), name);
       return;
     }
 
-  fmt = strdup (stat_format);
-  for (p = strtok (fmt, ","); p; )
+  char *fmt = strdup (stat_format);
+  for (char *p = strtok (fmt, ","); p; )
     {
-      if (memcmp (p, "st_", 3) == 0)
+      if (strncmp (p, "st_", 3) == 0)
 	p += 3;
       if (strcmp (p, "name") == 0)
 	printf ("%s", name);
       else if (strcmp (p, "dev") == 0)
-	printf ("%lu", (unsigned long) st.st_dev);
+	PRINT_INT (st.st_dev);
       else if (strcmp (p, "ino") == 0)
-	printf ("%lu", (unsigned long) st.st_ino);
+	PRINT_INT (st.st_ino);
       else if (strncmp (p, "mode", 4) == 0)
 	{
-	  unsigned val = st.st_mode;
+	  uintmax_t val = st.st_mode;
 
-	  if (ispunct ((unsigned char) p[4]))
+	  if (c_ispunct (p[4]))
 	    {
 	      char *q;
 
-	      val &= strtoul (p + 5, &q, 8);
+	      val &= strtoumax (p + 5, &q, 8);
 	      if (*q)
 		{
 		  printf ("\n");
-		  error (EXIT_USAGE, 0, _("incorrect mask (near `%s')"), q);
+		  error (EXIT_USAGE, 0, _("incorrect mask (near '%s')"), q);
 		}
 	    }
 	  else if (p[4])
 	    {
 	      printf ("\n");
-	      error (EXIT_USAGE, 0, _("Unknown field `%s'"), p);
+	      error (EXIT_USAGE, 0, _("Unknown field '%s'"), p);
 	    }
-	  printf ("%0o", val);
+	  printf ("%0jo", val);
 	}
       else if (strcmp (p, "nlink") == 0)
-	printf ("%lu", (unsigned long) st.st_nlink);
+	PRINT_INT (st.st_nlink);
       else if (strcmp (p, "uid") == 0)
-	printf ("%ld", (long unsigned) st.st_uid);
+	PRINT_INT (st.st_uid);
       else if (strcmp (p, "gid") == 0)
-	printf ("%lu", (unsigned long) st.st_gid);
+	PRINT_INT (st.st_gid);
       else if (strcmp (p, "size") == 0)
-	printf ("%s", umaxtostr (st.st_size, buf));
+	PRINT_INT (st.st_size);
       else if (strcmp (p, "blksize") == 0)
-	printf ("%s", umaxtostr (st.st_blksize, buf));
+	PRINT_INT (st.st_blksize);
       else if (strcmp (p, "blocks") == 0)
-	printf ("%s", umaxtostr (st.st_blocks, buf));
+	PRINT_INT (st.st_blocks);
       else if (strcmp (p, "atime") == 0)
-	printf ("%lu", (unsigned long) st.st_atime);
+	PRINT_INT (st.st_atime);
       else if (strcmp (p, "atimeH") == 0)
 	print_time (st.st_atime);
       else if (strcmp (p, "mtime") == 0)
-	printf ("%lu", (unsigned long) st.st_mtime);
+	PRINT_INT (st.st_mtime);
       else if (strcmp (p, "mtimeH") == 0)
 	print_time (st.st_mtime);
       else if (strcmp (p, "ctime") == 0)
-	printf ("%lu", (unsigned long) st.st_ctime);
+	PRINT_INT (st.st_ctime);
       else if (strcmp (p, "ctimeH") == 0)
 	print_time (st.st_ctime);
       else if (strcmp (p, "sparse") == 0)
@@ -698,9 +733,9 @@ print_stat (const char *name)
       else
 	{
 	  printf ("\n");
-	  error (EXIT_USAGE, 0, _("Unknown field `%s'"), p);
+	  error (EXIT_USAGE, 0, _("Unknown field '%s'"), p);
 	}
-      p = strtok (NULL, ",");
+      p = strtok (nullptr, ",");
       if (p)
 	printf (" ");
     }
@@ -708,6 +743,19 @@ print_stat (const char *name)
   free (fmt);
 }
 
+void
+set_times (char const *name)
+{
+  struct timespec ts[2];
+
+  ts[0] = ts[1] = touch_time;
+  if (utimensat (AT_FDCWD, name, ts,
+		 no_dereference_option ? AT_SYMLINK_NOFOLLOW : 0)
+      < 0)
+    {
+      error (EXIT_FAILURE, errno, _("cannot set time on '%s'"), name);
+    }
+}
 
 /* Exec Mode */
 
@@ -715,7 +763,7 @@ void
 exec_checkpoint (struct action *p)
 {
   if (verbose)
-    printf ("processing checkpoint %lu\n", (unsigned long) p->checkpoint);
+    printf ("processing checkpoint %jd\n", p->checkpoint);
   switch (p->action)
     {
     case OPT_TOUCH:
@@ -723,9 +771,11 @@ exec_checkpoint (struct action *p)
 	struct timespec ts[2];
 
 	ts[0] = ts[1] = p->ts;
-	if (utimensat (AT_FDCWD, p->name, ts, 0) != 0)
+	if (utimensat (AT_FDCWD, p->name, ts,
+		       no_dereference_option ? AT_SYMLINK_NOFOLLOW : 0)
+	    < 0)
 	  {
-	    error (0, errno, _("cannot set time on `%s'"), p->name);
+	    error (0, errno, _("cannot set time on '%s'"), p->name);
 	    break;
 	  }
       }
@@ -736,7 +786,7 @@ exec_checkpoint (struct action *p)
 	FILE *fp = fopen (p->name, "ab");
 	if (!fp)
 	  {
-	    error (0, errno, _("cannot open `%s'"), p->name);
+	    error (0, errno, _("cannot open '%s'"), p->name);
 	    break;
 	  }
 
@@ -748,14 +798,14 @@ exec_checkpoint (struct action *p)
     case OPT_TRUNCATE:
       {
 	int fd = open (p->name, O_RDWR | O_BINARY);
-	if (fd == -1)
+	if (fd < 0)
 	  {
-	    error (0, errno, _("cannot open `%s'"), p->name);
+	    error (0, errno, _("cannot open '%s'"), p->name);
 	    break;
 	  }
-	if (ftruncate (fd, p->size) != 0)
+	if (ftruncate (fd, p->size) < 0)
 	  {
-	    error (0, errno, _("cannot truncate `%s'"), p->name);
+	    error (0, errno, _("cannot truncate '%s'"), p->name);
 	    break;
 	  }
 	close (fd);
@@ -771,14 +821,14 @@ exec_checkpoint (struct action *p)
       {
 	struct stat st;
 	if (stat (p->name, &st))
-	  error (0, errno, _("cannot stat `%s'"), p->name);
+	  error (0, errno, _("cannot stat '%s'"), p->name);
 	else if (S_ISDIR (st.st_mode))
 	  {
 	    if (rmdir (p->name))
-	      error (0, errno, _("cannot remove directory `%s'"), p->name);
+	      error (0, errno, _("cannot remove directory '%s'"), p->name);
 	  }
 	else if (unlink (p->name))
-	  error (0, errno, _("cannot unlink `%s'"), p->name);
+	  error (0, errno, _("cannot unlink '%s'"), p->name);
       }
       break;
 
@@ -788,11 +838,11 @@ exec_checkpoint (struct action *p)
 }
 
 void
-process_checkpoint (size_t n)
+process_checkpoint (intmax_t n)
 {
-  struct action *p, *prev = NULL;
+  struct action *prev = nullptr;
 
-  for (p = action_head; p; )
+  for (struct action *p = action_head; p; )
     {
       struct action *next = p->next;
 
@@ -804,7 +854,7 @@ process_checkpoint (size_t n)
 	    prev->next = next;
 	  else
 	    action_head = next;
-	  if (next == NULL)
+	  if (!next)
 	    action_tail = prev;
 	  free (p);
 	}
@@ -820,45 +870,31 @@ process_checkpoint (size_t n)
 void
 exec_command (int argc, char **argv)
 {
-  int status;
-  pid_t pid;
-  int fd[2];
-  char *p;
-  FILE *fp;
-  char buf[128];
-  int xargc;
-  char **xargv;
-  int i;
-  char checkpoint_option[80];
-  
   /* Insert --checkpoint option.
      FIXME: This assumes that argv does not use traditional tar options
      (without dash).
   */
-  xargc = argc + 5;
-  xargv = xcalloc (xargc + 1, sizeof (xargv[0]));
+  int xargc = argc + 5;
+  char **xargv = xinmalloc (xargc + 1, sizeof *xargv);
   xargv[0] = argv[0];
-  snprintf (checkpoint_option, sizeof (checkpoint_option),
-	    "--checkpoint=%s", checkpoint_granularity);
-  xargv[1] = checkpoint_option;
+  xargv[1] = checkpoint_granularity;
   xargv[2] = "--checkpoint-action";
   xargv[3] = "echo=" CHECKPOINT_TEXT " %u";
   xargv[4] = "--checkpoint-action";
   xargv[5] = "wait=SIGUSR1";
-
-  for (i = 1; i <= argc; i++)
-    xargv[i + 5] = argv[i];
+  memcpy (&xargv[6], &argv[1], argc * sizeof *xargv);
 
 #ifdef SIGCHLD
   /* System V fork+wait does not work if SIGCHLD is ignored.  */
   signal (SIGCHLD, SIG_DFL);
 #endif
 
-  if (pipe (fd) != 0)
+  int fd[2];
+  if (pipe (fd) < 0)
     error (EXIT_FAILURE, errno, "pipe");
 
-  pid = fork ();
-  if (pid == -1)
+  pid_t pid = fork ();
+  if (pid < 0)
     error (EXIT_FAILURE, errno, "fork");
 
   if (pid == 0)
@@ -866,8 +902,7 @@ exec_command (int argc, char **argv)
       /* Child */
 
       /* Pipe stderr */
-      if (fd[1] != 2)
-	dup2 (fd[1], 2);
+      dup2 (fd[1], STDERR_FILENO);
       close (fd[0]);
 
       /* Make sure POSIX locale is used */
@@ -879,26 +914,31 @@ exec_command (int argc, char **argv)
 
   /* Master */
   close (fd[1]);
-  fp = fdopen (fd[0], "rb");
-  if (fp == NULL)
+  FILE *fp = fdopen (fd[0], "rb");
+  if (!fp)
     error (EXIT_FAILURE, errno, "fdopen");
 
-  while ((p = fgets (buf, sizeof buf, fp)))
+  while (true)
     {
-      while (*p && !isspace ((unsigned char) *p) && *p != ':')
+      char buf[128];
+      char *p = fgets (buf, sizeof buf, fp);
+      if (!p)
+	break;
+      while (*p && !c_isspace (*p) && *p != ':')
 	p++;
 
       if (*p == ':')
 	{
-	  for (p++; *p && isspace ((unsigned char) *p); p++)
+	  for (p++; *p && c_isspace (*p); p++)
 	    ;
 
 	  if (*p
-	      && memcmp (p, CHECKPOINT_TEXT, sizeof CHECKPOINT_TEXT - 1) == 0)
+	      && strncmp (p, CHECKPOINT_TEXT, sizeof CHECKPOINT_TEXT - 1) == 0)
 	    {
 	      char *end;
-	      size_t n = strtoul (p + sizeof CHECKPOINT_TEXT - 1, &end, 10);
-	      if (!(*end && !isspace ((unsigned char) *end)))
+	      intmax_t n = strtoimax (p + sizeof CHECKPOINT_TEXT - 1,
+				      &end, 10);
+	      if (!(*end && !c_isspace (*end)))
 		{
 		  process_checkpoint (n);
 		  kill (pid, SIGUSR1);
@@ -910,6 +950,7 @@ exec_command (int argc, char **argv)
     }
 
   /* Collect exit status */
+  int status;
   waitpid (pid, &status, 0);
 
   if (verbose)
@@ -949,11 +990,11 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  parse_datetime (&touch_time, "now", NULL);
+  parse_datetime (&touch_time, "now", nullptr);
 
   /* Decode command options.  */
 
-  if (argp_parse (&argp, argc, argv, 0, &index, NULL))
+  if (argp_parse (&argp, argc, argv, 0, &index, nullptr))
     exit (EXIT_USAGE);
 
   argc -= index;
@@ -969,6 +1010,14 @@ main (int argc, char **argv)
 	print_stat (*argv++);
       break;
 
+    case mode_set_times:
+      if (argc == 0)
+	error (EXIT_USAGE, 0, _("--set-times requires file names"));
+
+      while (argc--)
+	set_times (*argv++);
+      break;
+
     case mode_sparse:
       generate_sparse_file (argc, argv);
       verify_file (file_name);
